//Feynmann scaling xsec fit for kaon+
//
//    This code is intended to generate weights for neutrinos originating from K+ decays
//    utilizing Feynman Scaling coming for world data on p+Be 
//    
//    A this code is adopted from the MiniBooNE flux paper and the MiniBooNE reweighting framework 
//    along with code written by Raquel Castillo, Zarko, MiniBooNE (Steve Brice and Mike S), and Athula 
//
//    Current person adding comments and functions is Joseph Zennamo (jaz8600@fnal.gov)
//

#include "../WeightCalcCreator.h"
#include "../WeightCalc.h"

#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/RandomNumberGenerator.h"
#include "CLHEP/Random/RandomEngine.h"
#include "CLHEP/Random/RandGaussQ.h"

#include "nusimdata/SimulationBase/MCFlux.h"
#include "nusimdata/SimulationBase/MCTruth.h"
#include "nusimdata/SimulationBase/GTruth.h"

#include <vector>
#include "TH1.h"
#include "TArrayD.h"
#include "TMatrixD.h"
#include "TRandom3.h"
#include "TLorentzVector.h"
#include <iostream>
#include "TFile.h"
#include "TTree.h"
#include <TROOT.h>
#include <TChain.h>

using namespace std;

namespace evwgh {
  class PrimaryHadronFeynmanScalingWeightCalc : public WeightCalc
  {
  public:
    PrimaryHadronFeynmanScalingWeightCalc();
    void Configure(fhicl::ParameterSet const& p);
    std::pair< bool, double > MiniBooNEWeightCalc(simb::MCFlux flux, std::vector<double> rand);
    std::pair< bool, double > MicroBooNEWeightCalc(simb::MCFlux flux, std::vector<double> rand);
    virtual std::vector<std::vector<double> > GetWeight(art::Event & e);
    std::vector< std::vector< double > > MiniBooNERandomNumbers(std::string);
    
  private:
    CLHEP::RandGaussQ *fGaussRandom;
    std::vector<double> ConvertToVector(TArrayD const* array);
    std::string fGenieModuleLabel;
    std::vector<std::string> fParameter_list;
    float fParameter_sigma;
    int fNmultisims;
    int fprimaryHad;
    std::string fWeightCalc;
    std::string ExternalDataInput;
    double fScaleFactor;
    TFile* file;
    std::vector< std::vector< double > > fWeightArray; 
    std::string fMode;
    std::vector<double> FSKPlusFitVal;
    TMatrixD* FSKPlusFitCov;    
    double fSeed;
    bool fUseMBRands;

     DECLARE_WEIGHTCALC(PrimaryHadronFeynmanScalingWeightCalc)
  };
  PrimaryHadronFeynmanScalingWeightCalc::PrimaryHadronFeynmanScalingWeightCalc()
  {
  }

  void PrimaryHadronFeynmanScalingWeightCalc::Configure(fhicl::ParameterSet const& p)
  {

    // Here we do all our fhicl file configureation
    fGenieModuleLabel= p.get< std::string > ("genie_module_label");
    fhicl::ParameterSet const &pset=p.get<fhicl::ParameterSet> (GetName());
    std::cout << pset.to_string() << std::endl;

    fParameter_list		=   pset.get<std::vector<std::string> >("parameter_list");
    fParameter_sigma		=   pset.get<float>("parameter_sigma");
    fNmultisims			=   pset.get<int>("number_of_multisims");
    fprimaryHad			=   pset.get<int>("PrimaryHadronGeantCode");
    std::string dataInput       =   pset.get< std::string >("ExternalData");
    fWeightCalc                 =   pset.get<std::string>("weight_calculator");
    fMode                       =   pset.get<std::string>("mode");
    fScaleFactor                =   pset.get<double>("scale_factor");
    fSeed 			=   pset.get<double>("random_seed");
    fUseMBRands                 =   pset.get<bool>("use_MiniBooNE_random_numbers");

    // Getting External Data:
    //   Now let's get the external data that we will need 
    //   to assess what the central value of the p+Be -> K+
    //   cross section is and then also what are reasonable 
    //   variations around that central value.
    //   First: Pull in the file
    cet::search_path sp("FW_SEARCH_PATH");
    std::string ExternalDataInput = sp.find_file(dataInput);
    file = new TFile(Form("%s",ExternalDataInput.c_str()));
    //  Second: Define what we want to gather from that file which is two fold
    std::vector< std::string > pname; // these are what we will extract from the file
    pname.resize(2); // there are two items

    // We want the parameterization of the Feynman scaling and the covariance matrix
    // which characterizes the uncertainties and how they correlate across the parameterization
    pname[0] = "FS/KPlus/FSKPlusFitVal";
    pname[1] = "FS/KPlus/FSKPlusFitCov";
    TArrayD* FSKPlusFitValArray = (TArrayD*) file->Get(pname[0].c_str());
    //TArrayD is the most annoying format I have ever experienced so let's convert it to a vector
    FSKPlusFitVal = PrimaryHadronFeynmanScalingWeightCalc::ConvertToVector(FSKPlusFitValArray);    
    FSKPlusFitCov = (TMatrixD*) file->Get(pname[1].c_str());
    *(FSKPlusFitCov) *= fScaleFactor*fScaleFactor;
    std::cout << "Scale Factor being applied : " << fScaleFactor << std::endl; 

    // This is done but it is important to note that the parameterization maps such that 
    // FSKPlusFitVal[0] corresponds to the diagonal element FSKPlusFitCov[0][0]
    
    //Prepare random generator
    art::ServiceHandle<art::RandomNumberGenerator> rng;
    fGaussRandom = new CLHEP::RandGaussQ(rng->getEngine(GetName()));
    
    //
    //  This part is very important. You will need more than a single random number
    //  per event. In fact you will need per universe as many random numbers as there
    //  are rows in the covariance matrix. 
    //
    //  To properly perform Cholesky decomposition you'll use a correlated set of random
    //  numbers when creating the smeared cross section methods. Please see documentation
    //  in WeightCalc to help with this.
    //
    if(fUseMBRands){//fParameter_list
      fWeightArray = PrimaryHadronFeynmanScalingWeightCalc::MiniBooNERandomNumbers(fParameter_list.at(0));
    }//Use MiniBooNE Randoms
    else{
      
      fWeightArray.resize(2*fNmultisims);
      
      for (unsigned int i=0;i<fWeightArray.size();i++) {
	fWeightArray[i].resize(FSKPlusFitCov->GetNcols());      
	if (fMode.find("multisim") != std::string::npos ){
	  for(unsigned int j = 0; j < fWeightArray[i].size(); j++){
	    fWeightArray[i][j]=fGaussRandom->shoot(&rng->getEngine(GetName()),0,1.);
	  }
	}
	else{
	  std::fill(fWeightArray[i].begin(), fWeightArray[i].end(), 1.);
	}
      }
    }
  
  }
  std::vector<std::vector<double> > PrimaryHadronFeynmanScalingWeightCalc::GetWeight(art::Event & e)
  {


    //Collect the event's Flux information
    //     This specifically deals with the neutrino type and parentage
    art::Handle< std::vector<simb::MCFlux> > mcFluxHandle;
    e.getByLabel(fGenieModuleLabel,mcFluxHandle);
    std::vector<simb::MCFlux> const& fluxlist = *mcFluxHandle;

    
    //Collect event's MC truth information
    //  This specifically deals with the neutrino energy and 
    //  counting how many interactions there are per event 
    //  (neutrino counting is CRITICALLY important for applying the 
    //   correct weights and not ending up with unphysical values)
    art::Handle< std::vector<simb::MCTruth> > mctruthHandle;
    e.getByLabel(fGenieModuleLabel,mctruthHandle);
    std::vector<simb::MCTruth> const& mclist = *mctruthHandle;
    
    //Create a vector of weights for each neutrino 
    std::vector< std::vector<double> > weight;
    weight.resize(mclist.size());
       
    //If there aren't any neutrinos in the event then just return an empty weight vector
    if(mclist.size() == 0 || fluxlist.size() == 0){return weight;}

    // Let's start by iterating through each of the neutrino interactions 
    for(unsigned int inu = 0; inu < fluxlist.size(); inu++){

      // First let's check that the parent of the neutrino we are looking for is 
      //  the particle we intended it to be, if not set all weights to 1
      if (fluxlist[inu].ftptype != fprimaryHad){
	weight[inu].resize(fNmultisims);
	std::fill(weight[inu].begin(), weight[inu].end(), 1);
	continue; //now move on to the next neutrino
      }// Hadronic parent check
          
      //Let's make a weights based on the calculator you have requested 
      if(fMode.find("multisim") != std::string::npos){       
	for (unsigned int i = 0; int(weight[inu].size()) < fNmultisims; i++) {
	  if(fWeightCalc.find("MicroBooNE") != std::string::npos){
	    
	    //
	    //This way we only have to call the WeightCalc once
	    // 
	    std::pair<bool, double> test_weight =
	      MicroBooNEWeightCalc(fluxlist[inu], fWeightArray[i]);
	    
	    if(test_weight.first){
	      weight[inu].push_back(test_weight.second);
	    }
	  }
	  if(fWeightCalc.find("MiniBooNE") != std::string::npos){

	    //
	    //This way we only have to call the WeightCalc once
	    // 
	    std::pair<bool, double> test_weight =
	      MiniBooNEWeightCalc(fluxlist[inu], fWeightArray[i]);
	    
	    if(test_weight.first){
	      weight[inu].push_back(test_weight.second);
	    }
	  }
	}//Iterate through the number of universes      
      } // make sure we are multisiming

        

    }//Iterating through each neutrino 

    

    return weight;
  }

  //////////////////////////////
  ////////       Auxilary Functions
  //////////////////////////////

  //// 
  //   Use the MiniBooNE Implementation to determine the weight 
  ////
  std::pair< bool, double> PrimaryHadronFeynmanScalingWeightCalc::MiniBooNEWeightCalc(simb::MCFlux flux, std::vector<double> rand){
    
    // We now know that we have the hadronic parent that we want so we need to
    // prepare the kinimatic information about the iteraction to go into the Feynman Scaling 
    // calculation
    
    // Computations are based on MiniBooNE code that can be found here:
    // 
    //   http://cdcvs0.fnal.gov/cgi-bin/public-cvs/cvsweb-public.cgi/~checkout~/...
    //                            miniboone/AnalysisFramework/Utilities/src/PhysicsFunctions.F
    //
    //  Attributed author in MiniBooNE code is Mike Shaevitz 
    //
    //   Useful references:
    //       C.Mariani, G.Cheng, J.Conrad, and M.Shaevitz, PhysRevD.84.114021
    //       MiniBooNE Collaboration, PhysRevD.79.072002
    //
    
    // We need to guard against unphysical parameters 
    bool parameters_pass = false;

    // Get Neutrino Parent Kinimatics       
    double HadronMass;
      
      if(fabs(fprimaryHad) == 321) HadronMass = 0.4937; //Charged Kaon
      else{ 
	throw art::Exception(art::errors::StdException)
	  << "Feynman Scaling is only configured for Charged Kaons ";
      }

      TLorentzVector HadronVec; 
      double HadronPx = flux.ftpx;
      double HadronPy = flux.ftpy;
      double HadronPz = flux.ftpz;
      double HadronE  = sqrt(HadronPx*HadronPx + 
			     HadronPy*HadronPy + 
			     HadronPz*HadronPz + 
			     HadronMass*HadronMass);
      HadronVec.SetPxPyPzE(HadronPx,HadronPy,HadronPz,HadronE);

      // Get Initial Proton Kinitmatics 
      //   CURRENTLY GSimple flux files drop information about 
      //   the initial state proton, but that this 
      TLorentzVector ProtonVec;
      double ProtonMass = 0.9382720;
      double ProtonPx = 0;
      double ProtonPy = 0;
      double ProtonPz = 8.89; //GeV
      double ProtonE  = sqrt(ProtonPx*ProtonPx +
                             ProtonPy*ProtonPy +
                             ProtonPz*ProtonPz +
                             ProtonMass*ProtonMass);
      ProtonVec.SetPxPyPzE(ProtonPx,ProtonPy,ProtonPz,ProtonE);
      

      double E_cm = sqrt(2*ProtonMass + 2*ProtonMass*ProtonVec.P()); //Bug inherited from MB code
      //
      //  From Mike S. :
      //  The fitting was done without this error so I think that you need to use:
      //           ecm = sqrt(2.*mp**2+2.*mp*eb)
      //
      //    This should be fixed once the comparisons between MB and MicroBooNE have converged
      //
      //

      double gamma = (ProtonE + ProtonMass) / E_cm; 
      double gammaBeta = ProtonVec.P()/E_cm;

      double MaxThreshold;
      if(fabs(fprimaryHad) == 321){
	MaxThreshold = 2.053; // Mike S. Comment: K+: lambda + proton
      }
      else{
	throw art::Exception(art::errors::StdException)
	  << "Feynman Scaling is only configured for Charged Kaons ";	
      }

      double E_cm_max = (E_cm*E_cm
			 +HadronMass*HadronMass
			 -MaxThreshold*MaxThreshold)/(2*E_cm);
      double p_cm_max = sqrt(E_cm_max*E_cm_max-HadronMass*HadronMass);

      double HadronPT = HadronVec.Pt();
      double HadronPparallel = HadronVec.P()*cos(HadronVec.Theta());
      double ParallelCMp = gamma*HadronPparallel - gammaBeta*HadronE;

      double xF;

      xF = ParallelCMp / p_cm_max; 
	
      // Define the scale central value cross section
      
      // Pull out the parameters for the Feynman Scaling 
      double c1 = FSKPlusFitVal.at(0);     
      double c2 = FSKPlusFitVal.at(1);     
      double c3 = FSKPlusFitVal.at(2);     
      double c4 = FSKPlusFitVal.at(3);     
      double c5 = FSKPlusFitVal.at(4);     
      double c6 = FSKPlusFitVal.at(5);     
      double c7 = FSKPlusFitVal.at(6);     
      
      double CV = c1*(HadronVec.P()*HadronVec.P()/HadronE)*exp(-1.*c3*pow(fabs(xF),c4) 
							       - c7*pow(fabs(HadronPT*xF),c6)
							       - c2*HadronPT
							       - c5*HadronPT*HadronPT);

      if(CV < 0) CV = 0;
      if(fabs(xF) > 1) CV = 0;
      // Define the variations around that cross section
      //    To do this we will call the a function from WeightCalc
      //    that will generate a set of smeared parameters based 
      //    on the covariance matrix that we imported
      std::vector< double > FSKPlusFitSmeared = WeightCalc::MultiGaussianSmearing(FSKPlusFitVal, FSKPlusFitCov, rand); 
      
// Pull out the parameters for the Feynman Scaling 
      double smeared_c1 = FSKPlusFitSmeared.at(0);     
      double smeared_c2 = FSKPlusFitSmeared.at(1);     
      double smeared_c3 = FSKPlusFitSmeared.at(2);     
      double smeared_c4 = FSKPlusFitSmeared.at(3);     
      double smeared_c5 = FSKPlusFitSmeared.at(4);     
      double smeared_c6 = FSKPlusFitSmeared.at(5);     
      double smeared_c7 = FSKPlusFitSmeared.at(6);     

      if(smeared_c1 > 0 && 
	 smeared_c2 > 0 && 
	 smeared_c3 > 0 && 
	 smeared_c5 > 0 && 
	 smeared_c7 > 0){
	parameters_pass = true;	
      }
        
      double RW = smeared_c1*(HadronVec.P()*HadronVec.P()/HadronE)*exp(-1.*smeared_c3*pow(fabs(xF),smeared_c4) 
								       - smeared_c7*pow(fabs(HadronPT*xF),smeared_c6)
								       - smeared_c2*HadronPT
								       - smeared_c5*HadronPT*HadronPT);

      if(RW < 0) RW = 0;
      if(fabs(xF) > 1) RW = 0;
      
      double weight = 1; 

      if(RW < 0 || CV < 0){
	weight = 1;
      }
      else if(CV < 1.e-12){
	weight = 1;
      }
      else{
	weight *= RW/CV;
      }

      if(weight < 0) weight = 0;
      if(weight >30) weight = 30;
      if(!(std::isfinite(weight))) weight = 30; //From MiniBooNE; In Fortran Nan > 30 

      std::pair<bool, double> output(parameters_pass, weight);

      

      return output; 

  }// Done with the MiniBooNE function


  //// 
  //   Use the MicroBooNE Implementation to determine the weight 
  ////
  std::pair<bool, double> PrimaryHadronFeynmanScalingWeightCalc::MicroBooNEWeightCalc(simb::MCFlux flux, std::vector<double> rand){

    // 
    //  Largely built off the MiniBooNE code 
    //  but this is intended to expand beyond it
    //
    //  Edits from MiniBooNE code:
    //
    //  JZ (6/2017) : Remove max weight, set to max_limit of a double
    //  JZ (6/2017) : Fixed typo in the E_cm calculation per Mike S. suggestions 
    //   
    
    // We need to guard against unphysical parameters 
    bool parameters_pass = false;

    // Get Neutrino Parent Kinimatics       
    double HadronMass;
      
      if(fabs(fprimaryHad) == 321) HadronMass = 0.4937; //Charged Kaon
      else{ 
	throw art::Exception(art::errors::StdException)
	  << "Feynman Scaling is only configured for Charged Kaons ";
      }

      TLorentzVector HadronVec; 
      double HadronPx = flux.ftpx;
      double HadronPy = flux.ftpy;
      double HadronPz = flux.ftpz;
      double HadronE  = sqrt(HadronPx*HadronPx + 
			     HadronPy*HadronPy + 
			     HadronPz*HadronPz + 
			     HadronMass*HadronMass);
      HadronVec.SetPxPyPzE(HadronPx,HadronPy,HadronPz,HadronE);

      // Get Initial Proton Kinitmatics 
      //   CURRENTLY GSimple flux files drop information about 
      //   the initial state proton, but that this 
      TLorentzVector ProtonVec;
      double ProtonMass = 0.9382720;
      double ProtonPx = 0;
      double ProtonPy = 0;
      double ProtonPz = 8.89; //GeV
      double ProtonE  = sqrt(ProtonPx*ProtonPx +
                             ProtonPy*ProtonPy +
                             ProtonPz*ProtonPz +
                             ProtonMass*ProtonMass);
      ProtonVec.SetPxPyPzE(ProtonPx,ProtonPy,ProtonPz,ProtonE);
      
      //
      //  From Mike S. :
      //  The fitting was done without this error so I think that you need to use:
      //           ecm = sqrt(2.*mp**2+2.*mp*eb)
      //
      double E_cm = sqrt(2*ProtonMass*ProtonMass + 2*ProtonMass*ProtonVec.E());

      double gamma = (ProtonE + ProtonMass) / E_cm; 
      double gammaBeta = ProtonVec.P()/E_cm;

      double MaxThreshold;
      if(fabs(fprimaryHad) == 321){
	MaxThreshold = 2.053; // Mike S. Comment: K+: lambda + proton
      }
      else{
	throw art::Exception(art::errors::StdException)
	  << "Feynman Scaling is only configured for Charged Kaons ";	
      }

      double E_cm_max = (E_cm*E_cm
			 +HadronMass*HadronMass
			 -MaxThreshold*MaxThreshold)/(2*E_cm);
      double p_cm_max = sqrt(E_cm_max*E_cm_max-HadronMass*HadronMass);

      double HadronPT = HadronVec.Pt();
      double HadronPparallel = HadronVec.P()*cos(HadronVec.Theta());
      double ParallelCMp = gamma*HadronPparallel - gammaBeta*HadronE;

      double xF;

      xF = ParallelCMp / p_cm_max; 
	
      // Define the scale central value cross section
      
      // Pull out the parameters for the Feynman Scaling 
      double c1 = FSKPlusFitVal.at(0);     
      double c2 = FSKPlusFitVal.at(1);     
      double c3 = FSKPlusFitVal.at(2);     
      double c4 = FSKPlusFitVal.at(3);     
      double c5 = FSKPlusFitVal.at(4);     
      double c6 = FSKPlusFitVal.at(5);     
      double c7 = FSKPlusFitVal.at(6);     
      
      double CV = c1*(HadronVec.P()*HadronVec.P()/HadronE)*exp(-1.*c3*pow(fabs(xF),c4) 
							       - c7*pow(fabs(HadronPT*xF),c6)
							       - c2*HadronPT
							       - c5*HadronPT*HadronPT);

      if(CV < 0) CV = 0;
      if(fabs(xF) > 1) CV = 0;
      // Define the variations around that cross section
      //    To do this we will call the a function from WeightCalc
      //    that will generate a set of smeared parameters based 
      //    on the covariance matrix that we imported
      std::vector< double > FSKPlusFitSmeared = WeightCalc::MultiGaussianSmearing(FSKPlusFitVal, FSKPlusFitCov, rand); 
      
// Pull out the parameters for the Feynman Scaling 
      double smeared_c1 = FSKPlusFitSmeared.at(0);     
      double smeared_c2 = FSKPlusFitSmeared.at(1);     
      double smeared_c3 = FSKPlusFitSmeared.at(2);     
      double smeared_c4 = FSKPlusFitSmeared.at(3);     
      double smeared_c5 = FSKPlusFitSmeared.at(4);     
      double smeared_c6 = FSKPlusFitSmeared.at(5);     
      double smeared_c7 = FSKPlusFitSmeared.at(6);     

      if(smeared_c1 > 0 && 
	 smeared_c2 > 0 && 
	 smeared_c3 > 0 && 
	 smeared_c5 > 0 && 
	 smeared_c7 > 0){
	parameters_pass = true;	
      }
        
      double RW = smeared_c1*(HadronVec.P()*HadronVec.P()/HadronE)*exp(-1.*smeared_c3*pow(fabs(xF),smeared_c4) 
								       - smeared_c7*pow(fabs(HadronPT*xF),smeared_c6)
								       - smeared_c2*HadronPT
								       - smeared_c5*HadronPT*HadronPT);

      if(RW < 0) RW = 0;
      if(fabs(xF) > 1) RW = 0;
      
      double weight = 1; 

      if(RW < 0 || CV < 0){
	weight = 1;
      }
      else if(CV < 1.e-12){
	weight = 1;
      }
      else{
	weight *= RW/CV;
      }

      if(weight < 0) weight = 0;
      if(weight > 30) weight = 30;
      if(!(std::isfinite(weight))){
	std::cout << "FS : Failed to get a finite weight" << std::endl; 
	weight = 30;}

      std::pair<bool, double> output(parameters_pass, weight);

      

      return output; 

  }// Done with the MicroBooNE function


  //// 
  //  This converts TArrayD to std::vector< double > 
  ///
  std::vector<double> PrimaryHadronFeynmanScalingWeightCalc::ConvertToVector(TArrayD const* array) {
    std::vector<double> v(array->GetSize());
    std::copy(array->GetArray(), array->GetArray() + array->GetSize(),
	      v.begin());
    return v;
  } // ConvertToVector()


  std::vector< std::vector< double > > PrimaryHadronFeynmanScalingWeightCalc::MiniBooNERandomNumbers(std::string pdg){
    
    if(pdg.compare("kplusOld") == 0){	 
      return {{0.987566, 0.590147, 0.458403, -0.122770, -1.367813, 1.949886, 2.023164},{0.539147, 0.758137, 1.196676, -0.584759, 0.179203, -0.480346, -1.054715},{-1.140844, -0.207580, 1.088544, 1.192162, -0.675312, -0.109232, -0.024246},{0.626242, -1.373842, 0.136977, -0.656965, -0.091256, 0.064668, 0.416249},{-0.724027, -1.072060, 1.121722, -0.561373, -0.761086, 0.183185, 1.723256},{0.607346, 0.681351, 0.912106, -0.048851, -0.624712, -0.731254, 1.259734},{0.376265, 1.987476, -0.070809, -0.297597, -0.979156, 0.056236, -0.674761},{-0.584635, -0.299805, 0.556268, -0.743457, -0.256697, -0.477591, -0.179009},{-0.400616, 0.825124, 0.038153, -1.011065, -0.688462, 0.304253, -0.682345},{1.222469, 1.069363, -0.288636, -0.779033, 0.178344, -0.995003, 0.501106},{-0.440267, 0.750927, -0.029705, 1.566773, 0.348531, -0.354376, 1.037448},{-1.897691, -0.008422, 1.910249, -0.253573, -1.313578, 1.363991, 0.439462},{0.698244, -0.670053, -1.241552, 1.249973, 0.204428, 0.330957, 0.871313},{-0.812217, 0.216954, 0.617920, -0.541960, -1.330447, -0.260093, 0.194204},{0.268647, -0.924625, -1.448855, 0.446158, 0.974278, -0.313940, 2.307845},{0.045024, 0.258856, 0.822321, -1.079189, -0.100214, -0.897349, -1.382724},{-0.621896, -0.667707, 0.661231, 1.232589, -0.777697, -0.335985, -0.298629},{-0.545555, 0.811748, 0.619237, 0.282474, -1.834809, 1.006987, 1.399189},{-1.441205, -0.181594, -0.091686, 0.399727, -1.307781, -1.782866, 0.681929},{0.667055, -1.170482, 0.455392, -0.407393, 0.543055, -1.064851, -1.034181},{-0.241101, 0.120451, -1.008629, 1.901972, 0.352733, -0.731724, -0.087648},{-0.544369, -0.977124, 1.166526, -0.619782, 1.754364, 0.237322, 1.607430},{0.179351, 1.976358, -1.250711, -1.034280, 0.606907, 0.183557, -0.131594},{1.144569, 0.156077, 0.666895, 0.834086, -1.005958, -0.054306, -0.814731},{-2.311010, -0.574249, 0.002755, -1.040539, -1.828027, 0.021301, 1.641081},{-1.520022, -0.326911, 1.936578, -0.956390, 0.368016, -0.112316, -0.341867},{-0.471542, 0.661799, 0.152018, -1.468440, -1.607912, -0.841030, -0.308894},{-0.149893, -0.958946, -1.195120, 0.324765, -0.143900, -1.197965, -0.309901},{-0.658717, 1.149493, 1.389772, 0.381987, -0.152526, 2.181308, -0.212018},{0.767291, 0.823090, 0.575474, -0.787939, 0.781887, -2.115347, 1.312135},{0.447331, -1.272043, 0.905651, -0.251088, -0.819299, -0.698765, -0.342840},{-0.971937, 0.438908, -0.598023, 0.968393, -1.436026, 0.059028, -0.267414},{-1.848993, -0.193153, 0.802172, -0.181578, -1.073178, -0.196675, 1.072166},{0.905213, 0.597347, -1.565406, -1.057459, 0.872671, 1.027683, 0.914627},{0.108589, 0.644082, 0.600828, -0.763597, -1.766559, 1.311954, -0.158515},{1.070219, -1.645363, 0.270964, 0.153934, -1.620207, -0.115930, -0.960926},{0.716433, 0.226697, -0.003128, 1.814947, -1.527354, -0.613644, -1.779851},{1.429107, 0.361562, 0.847099, -0.193465, -0.583809, 0.796497, 1.060541},{-1.535106, 0.648683, 0.954315, -1.233672, 0.561843, 0.962432, -0.101581},{-0.113205, 0.420161, -0.932145, 0.190818, 0.266188, -0.522332, 1.777947},{-0.613985, 1.424944, 0.558233, 0.961640, 0.561219, 0.295329, 2.148828},{-0.590766, 0.311581, -2.421137, 1.521962, 1.946332, 1.729404, 1.886342},{-1.397381, -0.359779, -2.083137, -0.425720, 1.731196, 0.231670, -1.043465},{1.028434, 0.205339, 0.144763, 0.653613, 0.145261, -0.468729, 0.648544},{-0.291219, 0.286708, -1.659819, 1.986640, -0.377493, 0.480341, 2.747777},{0.825223, 0.932854, -0.129643, -0.672552, -1.251400, 1.066675, -0.991367},{-2.369488, 1.222414, -1.986858, 0.991075, 0.981689, -1.303496, -1.300980},{-0.519004, -2.094331, 0.517905, 0.827368, 0.992231, -0.466653, 1.214431},{0.959192, -2.391499, 2.093422, -0.418450, -0.374235, -0.675370, 0.940353},{-0.410079, 1.119337, 0.358980, 0.844434, 1.152840, -0.135441, -0.274634},{0.262560, -1.439413, 0.798092, 2.477254, -0.296678, -0.642665, -0.874979},{-0.155769, 2.392772, -0.371881, 0.318066, -0.706122, -0.583314, 0.157781},{-2.192031, 0.144638, 0.083953, 1.503232, -2.133984, -1.266652, 0.441972},{0.640280, -0.485910, -1.756495, 0.073480, 0.814811, 0.110157, -0.039490},{-2.312853, -0.290227, -1.630635, 1.666717, 1.550232, 0.099041, -0.525953},{-1.779329, 0.477223, 0.170523, 0.749645, 0.239665, 0.833925, -0.115213},{-0.741578, -0.401856, -1.621058, -0.282865, -1.900711, 0.614528, 0.101124},{0.691067, 0.153188, -0.425875, -0.725144, -0.203046, -0.412803, -0.121265},{1.257548, -0.385161, 1.511449, -1.392009, 1.252869, 1.177145, 0.420760},{-0.439030, 1.177556, 0.383544, 0.442070, 1.154329, -0.183326, 0.005687},{-0.912443, 1.368476, -1.050008, 0.058536, 0.394202, 0.164380, 1.107255},{0.683407, -0.548884, -0.310040, 0.612932, -1.124034, -1.131498, -0.970562},{-0.044046, 1.025876, -1.258951, -0.057264, 0.901834, -1.769660, -1.608053},{-1.725599, -0.114166, -0.884059, -1.869534, 1.900306, 0.652599, 0.642498},{0.358140, -0.491707, -0.915839, -1.149057, -0.473809, 0.568296, 0.648020},{1.745424, -1.121469, -0.240656, -0.662892, 0.253769, -0.483687, -1.076843},{0.889069, -2.059444, 1.292080, 0.577106, 1.688457, -0.221378, -0.641502},{0.334921, 1.516564, -0.961590, 0.506769, 2.087729, -1.060068, 0.569560},{0.475541, -0.153244, 0.181889, 0.348256, 1.258619, -0.747172, -0.905298},{-0.481623, -1.126299, 0.174316, 1.477275, 0.723762, 0.255490, 1.019244},{-1.148142, 1.743187, 0.802170, -0.405180, 1.195575, -1.936020, -0.103404},{-0.164373, -0.721080, 0.165662, 2.260659, -0.014583, 0.914410, -1.327635},{-0.179198, -0.143391, 0.362583, 0.132691, 0.608638, -0.763702, 0.369684},{1.374352, 0.397866, 0.392494, 0.831848, -0.596462, 0.243157, -1.967196},{-0.234546, -1.074764, 0.190969, 0.847513, 0.156121, -1.477806, 0.821370},{0.660724, -1.843609, -0.596093, 1.052129, -1.009996, 0.575923, -1.371201},{-0.585032, 0.765901, 0.939232, 0.082643, -0.240123, 0.839862, 0.139080},{-1.667821, 1.145637, -0.677666, -1.404523, 1.503914, 0.431101, -0.001103},{-0.350383, 0.882801, 0.459725, 1.053988, 0.729148, -0.053007, 2.013875},{0.450593, -1.176154, -0.277810, 0.819075, 1.204701, -1.731610, 1.036752},{1.595623, -1.998478, 0.382954, -0.020644, 0.092662, 1.007416, 0.859721},{-1.178937, 0.625623, 1.331233, -2.064700, -0.603319, 0.032733, -1.710805},{-0.158590, 0.066204, -1.207755, -0.394973, 0.080982, 0.724663, -1.106958},{-1.345082, -0.735637, 0.177881, -0.035687, 1.650075, -0.200578, 0.764670},{-0.459446, 1.015325, -0.196229, 1.116246, 0.981574, 1.110039, 0.369887},{0.013893, -1.482541, -0.565858, 1.156010, 0.571803, -0.842280, 0.650344},{-0.727766, -0.319505, -1.488822, 0.300481, 1.699568, 0.131107, 0.997031},{0.418363, -1.103378, -0.170588, 0.026628, 1.214283, -0.156037, -1.027679},{1.170983, 1.426937, 0.207475, 0.194321, 1.004491, 1.566141, 0.110674},{-1.240379, -1.356160, 0.841013, 0.919542, 1.571214, -0.737325, 0.461503},{0.896231, -2.334933, 0.742667, 2.260547, -1.786749, 0.052101, 0.954759},{1.694942, 0.684905, 0.070856, -0.915424, 1.310825, -0.295141, -0.287149},{-0.310034, -0.658223, 1.491403, 1.352753, -0.733636, 0.297119, 0.661791},{0.877795, 0.236839, 0.495237, 1.160387, -1.112595, 0.179108, -0.824691},{0.596605, 0.051665, 0.637969, 0.231351, -0.800775, 0.907236, -0.082107},{-0.284185, 1.769118, 0.942419, -0.358655, 0.359480, -1.008070, -1.243408},{-0.441128, 0.370261, -0.480924, 1.418448, 0.502501, 0.158039, 1.229927},{-0.697179, 1.169014, -0.677076, -0.653265, 1.247390, 0.133338, 2.146465},{0.171130, -0.999078, 0.414958, -0.783751, -0.764720, -1.303999, -0.110760},{1.318226, 0.063897, -0.231674, 0.043894, -1.040422, -1.015446, 0.795568},{-2.013772, 0.745271, -0.470491, -1.677407, 1.124070, 0.505768, 0.733714},{0.172704, -0.991222, 0.082162, 0.095543, 2.573303, -0.335095, -0.080446},{0.652683, 2.064954, -0.448767, -1.481600, -1.231961, 0.475463, 0.634805},{1.991741, 0.104847, 1.062247, 1.515304, 1.267677, 1.120061, 0.470419},{0.457066, -1.591351, -0.059658, -0.537577, -0.944550, -0.592243, -1.289802},{2.627123, 0.605717, 0.079369, -0.177457, 1.454220, 0.540481, 0.419237},{-1.078775, 0.567086, -0.010431, 0.063787, -0.626162, 1.262453, -0.695746},{-1.571128, -0.132312, 0.151044, -0.789398, -0.550692, 1.306602, 0.153716},{-0.221647, 0.010432, -0.339434, -0.712337, -1.365557, -0.018685, 1.286168},{-0.650626, -0.441632, 1.191691, 0.682487, 0.000432, -1.379886, -0.280606},{-0.922982, 0.514888, -0.300369, -0.608309, 0.751441, 1.116133, -1.565230},{2.070771, 0.884214, 0.772386, 0.135275, 0.930447, 1.296854, 0.559127},{0.215233, 0.174287, 1.354968, -0.347821, -2.109350, -1.046985, 0.695767},{-0.116864, 1.630604, 1.567891, -0.037639, 1.237893, -0.106212, 0.641826},{-1.351391, -0.648300, 0.677396, 0.556840, 0.213294, 0.775235, -1.267129},{-0.907270, -0.165924, -0.732809, -0.124981, -0.300213, -0.933177, -0.475827},{-1.823163, 0.460852, 0.171415, -1.159655, -0.112603, 1.109503, 0.496348},{-1.478886, 1.249140, -0.672067, -0.281300, 0.365615, -1.470624, 0.334249},{-0.775635, -0.524436, -1.512947, -0.121192, -1.256112, -0.722740, -2.110209},{-0.406357, -0.623119, -0.689975, -0.832319, -0.690753, 1.581238, 0.131169},{-1.745654, 0.719678, 2.056678, 1.780356, 0.497503, 1.516364, -0.048374},{0.106471, -0.773457, 0.490553, 0.161006, 1.118300, -1.127222, -0.845426},{1.162704, -0.692165, 1.310347, 0.546723, 0.230749, 0.499493, 0.507224},{-0.368916, -0.771496, 0.660584, -1.340570, -0.968148, -0.168151, 1.198093},{-0.159978, -0.519051, -0.218055, -0.057162, 1.182909, -0.083753, -0.566497},{1.644934, 0.272128, 1.045612, 0.810220, -0.349049, -0.562061, 0.115042},{-1.283773, 0.564927, -1.925383, -1.602896, -0.294801, 0.830456, -0.648987},{0.537528, 1.165937, -1.828423, -0.051713, 2.477421, 0.916935, 0.451110},{1.098518, 1.120872, 2.385913, -0.622528, -1.253680, 0.623146, -0.568365},{0.836886, 0.160426, -1.476098, -0.216342, 0.380438, 0.145056, -1.224026},{0.984815, -0.052469, 1.137188, -0.660908, -0.330672, -0.713622, -1.516728},{0.370454, 0.361038, 1.135811, -0.063547, -1.187538, -1.562865, -0.694952},{-0.429713, 2.707496, 0.538798, -0.223975, -1.592489, -0.469184, 1.271266},{0.207002, -0.343845, 0.937440, 0.607674, -0.256318, 0.527442, -1.093916},{-1.258492, -0.303758, 1.594609, -0.451042, -0.304545, 2.255564, 0.196407},{0.035843, -0.121671, -0.175789, 0.721431, 0.810685, 0.145238, 0.708414},{0.668419, -0.570758, -0.790012, 0.436137, -1.681664, -0.038135, -0.792378},{-0.052244, 1.984220, -1.185776, 0.048164, 1.519224, -0.190204, -0.187655},{0.496671, -1.904743, 0.095582, -0.791579, -0.772473, 1.629920, 1.655294},{0.262551, -1.402570, 0.908902, 0.370387, -1.022575, 0.742212, 0.449157},{-0.065299, 1.945870, -0.828224, -0.134373, -0.140253, -0.638109, -0.167214},{-0.780552, 0.953268, 1.120994, -0.626493, -0.081714, 0.452160, 1.255117},{0.182480, -0.412507, -0.093917, -0.626541, -0.345197, 0.667863, 0.899302},{-0.669140, 0.215041, -0.108643, 0.910828, 0.589745, -0.793809, 0.540658},{0.186205, -1.443889, 0.131653, 0.332302, 0.256427, 1.845201, -1.300680},{-0.018924, -0.034385, -1.025080, -0.084907, -1.583787, -1.033863, -0.615050},{-1.822438, -2.314142, -0.042544, -1.586799, -1.514214, -0.330578, -0.597436},{0.325761, -0.027545, -0.179747, 0.886424, -1.128672, -0.089729, 0.325016},{0.520724, -0.613321, 1.808837, -0.347981, 1.418352, 1.129685, 0.557553},{-0.105172, -1.007796, 1.285585, -0.147819, -0.579910, 0.211751, -1.228843},{-1.070832, -0.356454, 1.190014, -0.106017, -2.268402, -0.807614, 1.123969},{0.680822, -1.791192, 0.214384, -0.154802, -0.813590, 0.339691, 1.641239},{0.275755, -1.752059, -2.779209, 1.976661, -0.315744, 1.019834, 0.270857},{0.430040, -1.681850, -0.812602, -1.022922, 0.671039, 1.572727, -0.988831},{0.156705, 1.496289, 0.876244, 0.225183, -0.258348, -0.483638, 0.466725},{1.502585, -0.316078, 0.219417, -2.251451, 0.607545, 0.825205, -0.009611},{0.106958, 0.199421, -0.478719, -0.126642, 0.002828, 0.960335, -0.646721},{-0.154881, -0.219166, -1.643821, 0.589645, 1.059551, 0.377243, 1.015106},{-0.293648, 2.012516, -0.009783, 1.062348, 1.107301, 0.184429, -0.127637},{-0.548454, -0.601426, 0.499240, 0.991327, 0.418651, 0.253193, -0.107517},{-0.714439, 1.359436, 0.944239, 0.759811, -0.425367, 2.081533, -0.789502},{-0.362270, -1.195911, 0.604053, -1.771744, 2.112737, -0.807933, 0.732799},{-0.982328, 0.148423, 1.458050, 0.319251, -0.738516, -1.263910, 1.686434},{-0.799495, -0.655133, 0.237856, -1.401343, 0.966510, 1.903182, -0.114758},{-0.347219, 0.749681, 0.713848, 1.287094, 1.924892, -0.453605, 0.315990},{-1.739684, 0.154400, -2.447857, 0.392489, -1.097630, -0.195769, 0.195824},{0.195806, -0.724168, 0.100227, -1.188346, 0.684901, -0.088319, -0.161340},{-0.090455, -0.440439, 0.401249, 0.378205, -2.010626, 0.550599, -0.068729},{-0.142807, -0.308110, -0.348779, 0.542822, -0.641606, 1.018645, 0.391289},{1.671360, -0.372483, 1.526818, 0.429404, -0.433686, -1.173547, 1.025825},{-0.025775, 1.010430, -0.509943, 0.492807, -0.526692, -0.327204, 2.301909},{-0.105312, 0.280747, 0.092182, 0.685023, 1.409522, 0.286837, -0.196865},{0.739660, 3.737917, 1.488467, -1.195451, 0.338756, 0.300484, 0.874028},{-1.499985, -0.812982, 1.038491, 0.470746, 1.107606, 2.296978, -0.989519},{-0.238803, -0.463303, 1.237911, 0.566354, -1.162288, -0.336847, -0.999219},{0.111127, 1.521193, -1.705186, 1.359646, -0.548171, -0.142765, 0.444188},{-0.699431, 0.406500, -1.770323, -0.020498, -1.148754, 1.202178, 0.930762},{1.494866, 0.809864, 0.059594, -1.236665, -0.806758, 0.238128, 0.789305},{1.261187, 2.526969, 0.153971, 0.316914, 0.290369, -0.260984, 1.041756},{-0.291752, 0.745475, -0.700315, -0.325944, 0.583735, 0.371964, 2.363027},{-1.445256, -0.658223, -0.951478, 1.710776, -0.961814, 0.549519, -0.370373},{0.913554, -0.361969, -0.268382, -0.851312, 2.059877, 0.911559, -0.701192},{-0.314220, 1.376409, 0.037590, -1.510405, 1.823345, -2.217447, 0.909880},{-1.461311, -2.364661, 1.487242, -0.710339, 0.208794, 0.885900, -0.729266},{-1.569258, -1.205786, -0.833666, 1.007340, -1.210826, -0.846285, 0.284148},{-0.170596, 1.316996, -0.650509, 1.355711, 0.160880, 0.604201, -0.515528},{-0.733463, 0.698632, -1.668289, 0.756107, -1.675099, 1.016934, 0.571903},{-0.531006, 1.520527, 0.099557, 0.220497, 0.007443, 0.460386, 0.500537},{-0.098119, -1.228252, -0.794437, 0.518959, 0.187433, -0.764905, -1.324016},{3.226697, -1.419957, 0.120332, -1.375437, -0.292468, -0.165306, 0.341392},{1.175343, -0.056189, 0.255147, -0.031817, 0.511516, -0.399063, -0.106516},{-1.046589, 0.995216, 0.511113, 1.107987, -1.338351, 1.950138, 0.772843},{-0.331593, -0.775631, -0.871665, -0.945177, 0.151126, -0.022745, -1.579811},{0.540497, 0.929121, -1.758172, 1.337824, 1.498203, -0.109656, -1.687890},{-1.997486, 1.870978, 0.059035, 0.016848, -1.201702, -1.109639, 1.032240},{1.017055, 0.103109, -1.254861, 1.767974, -2.244667, 0.445948, -0.589576},{0.306794, -1.866412, 0.486193, -1.672537, 1.187687, 0.443254, 1.767496},{-1.975252, 0.861455, -0.865982, 1.431688, 0.367798, 1.271098, 1.266318},{-0.177568, 0.517690, -1.389663, -0.435776, -0.455923, 1.630682, 0.314336},{-1.462716, -0.096765, -0.773052, -0.205330, 0.310692, 1.286700, 0.037777},{0.859334, -0.204027, -1.731793, -0.158785, -1.770551, -0.278439, 0.784962},{-1.511439, 3.425885, -2.308568, 0.470210, 2.543814, -0.475754, -0.123361},{-0.387419, -0.397819, 0.314572, 0.049639, -0.796375, -0.113282, -0.659663},{0.470716, -1.186307, 1.098412, -0.664017, -1.549576, -0.150165, -0.322284},{1.484456, -0.623928, 0.844807, 0.106067, 0.455583, 1.623806, 0.184204},{0.362871, -0.356932, 0.659696, -1.471247, -1.258891, -0.421465, 0.214738},{-1.496890, 0.061730, -0.203062, 1.030247, 1.247947, 0.323816, 0.388383},{-0.212523, -0.452507, 0.514302, -2.196751, -0.029399, -0.099594, 2.246914},{0.670909, -0.595648, -1.889996, 0.830232, 1.255243, -1.303455, 0.071786},{-0.080893, 0.519527, -2.500899, -0.163828, -0.269112, 1.657044, 1.038594},{1.256791, -1.315958, 0.868583, 0.576675, 0.362121, 0.735047, 0.745008},{-1.353501, -0.247096, 1.519948, 0.882877, -0.037273, 0.890989, 0.504725},{0.765931, -0.197398, -0.214861, -1.680420, 1.201258, 0.263001, -0.709555},{1.578805, 0.183662, -0.299338, -1.423890, -0.928225, 0.676357, 1.008425},{-1.249174, -0.236191, -1.335864, 1.357973, 2.284229, -0.988227, -0.193875},{-0.994510, 0.629917, 0.573929, 1.011974, 1.167690, -1.068560, 1.674399},{1.537007, 0.621022, 1.404330, -0.589717, -1.346231, -1.077187, 0.758084},{-0.711137, -0.579476, -0.206868, 1.030491, -0.150073, -0.351344, -0.909641},{-1.305369, 0.235420, -1.046254, -1.573489, 0.634870, 1.007632, 0.494437},{0.596953, -0.483630, -0.953271, -1.120691, -1.157593, -0.970792, 1.342888},{0.554507, 0.279003, -0.143638, -0.342005, 0.173400, -0.134642, -0.461301},{-0.254568, 0.391952, 0.043249, -1.182720, 1.517370, -1.140779, 1.610870},{-0.955436, -0.344051, 1.662211, 0.502465, 0.532495, -0.085866, 0.208687},{0.365759, 0.557320, 1.133614, 0.763081, 0.022655, -0.273849, 0.718262},{-1.067525, -0.750128, -0.540219, -1.144859, -0.945609, 0.670532, -1.265244},{0.042967, -1.430502, -0.457357, -1.010990, 1.534921, -0.310664, 0.321458},{-0.659211, 1.362966, 0.874695, 0.355890, 0.198982, 1.394696, 1.325735},{1.679861, -0.825314, 0.240005, 0.640526, 0.092863, 1.923671, 0.514489},{-0.144227, 2.111982, -0.714856, -0.801375, 0.074684, 1.269205, 0.433543},{0.278609, -0.488441, 0.502545, -0.760914, 0.315273, 1.108955, -0.150227},{0.503246, 1.013372, -0.836913, 0.809209, 1.339238, 1.002041, -0.244372},{1.458006, 0.064021, 1.257763, -0.127711, 0.077005, -1.674433, -1.579613},{-0.438802, -0.293864, -0.309402, 0.071699, 0.535491, -1.059206, 1.059044},{-0.689814, -0.260124, 0.078374, -0.393549, -0.911078, -0.976921, -1.326822},{0.078966, 1.132827, 1.808984, 0.352009, -1.913098, -0.117063, -0.159016},{0.699511, 0.330200, -1.605980, 0.590051, 2.779666, 0.230313, -1.278701},{0.540647, 0.004816, 0.318969, -0.877203, 0.849533, -1.047778, 1.886675},{1.101976, 1.794215, 0.153485, 2.007926, 0.383344, 0.447751, -0.023372},{0.671884, 0.137110, 0.516658, 1.026191, -0.647335, -0.052557, -0.008419},{0.127914, 0.410701, -0.455147, 1.187064, -0.987861, -0.981257, -1.001312},{-1.836980, 0.860464, 0.139009, -0.339169, 0.389513, -1.174609, -0.251393},{0.374220, -1.216579, 0.945710, -1.017989, 0.562833, 0.469543, 1.047480},{-0.422797, 0.455105, 2.110928, -1.110984, 0.164970, 1.061288, -0.548410},{-1.285348, 1.679245, -1.112164, 0.210521, -0.069650, 0.425783, 1.730841},{-1.858641, 0.426160, -1.329963, 0.261700, 0.747900, 0.265623, 2.129384},{0.167742, 0.251891, 0.644824, 1.563245, -0.132501, 1.410190, 0.885294},{0.252660, 0.733671, 1.324286, 0.219398, -0.531682, -0.578192, 0.486442},{0.962653, 0.705254, -1.457751, 0.252066, 1.811726, -0.734989, -0.247151},{-0.802072, -0.986747, 1.539938, -0.465972, 0.448448, 0.960120, -0.657179},{-0.014956, 2.897720, -1.132912, -1.428028, 3.167142, -0.399239, 1.843749},{-0.093004, 0.693019, 0.109587, -0.944294, -0.086289, 2.531008, 1.181620},{-2.504292, 0.868583, -0.662039, 2.730691, -0.765329, 0.582409, 1.019004},{-0.897932, 1.091561, 0.001794, 0.288569, 0.540828, -1.531330, 1.255413},{-1.405604, -0.120912, -0.390275, -0.769388, 0.599249, -0.265511, -1.329174},{-0.195745, -0.361764, 0.542223, -0.292734, -0.397351, 1.085668, -0.283986},{-0.025747, -0.773512, -0.059092, 0.679528, -0.838799, -1.204744, 0.608684},{-0.372741, -0.900781, -0.079310, -0.146898, 0.814496, -1.847625, 0.504824},{-0.539659, 1.943042, 0.209841, 2.485345, -0.075758, -1.022858, 0.090843},{-0.723378, 0.494207, 0.161715, 0.578361, -1.425149, 0.733860, -0.326898},{0.562336, 1.042796, -0.468065, 1.432209, 0.158661, -1.576081, 1.349984},{-1.578639, -1.920053, 0.260532, -0.298998, -0.781838, -0.913347, 0.672004},{-0.488429, -0.211963, 1.180777, 0.813001, -0.534110, -1.147567, 0.011684},{0.033817, 0.469805, 1.019146, -0.299215, -0.186256, 0.855646, -0.430952},{0.626004, 0.686660, 0.140365, -0.703758, -0.501333, 0.699701, -1.179831},{-1.225294, -0.316666, -0.401844, 0.421073, 0.059618, 0.486076, -0.377575},{-0.356128, -0.634033, -0.297715, 0.271903, 1.466068, -0.132636, -0.544956},{-0.362922, 0.601891, -2.826189, -0.587529, 0.852490, 0.223084, -0.506540},{-0.649043, -1.033998, -0.445299, 1.246552, 1.095944, 0.446121, -1.036947},{0.222386, 0.604726, 1.442143, 0.553415, -0.597809, 0.129413, -0.704852},{-0.649974, 0.097720, -0.901852, 0.999577, 2.313604, 1.708599, 0.256431},{-0.613793, -1.506175, -1.043126, -0.607055, -0.421631, -0.057646, -2.775742},{1.113088, 0.883452, 1.528824, 0.479807, 1.222608, -1.050528, -0.004671},{-0.190799, -1.056768, 1.286533, 0.514825, 1.591685, 0.730185, -0.493674},{-0.871354, 0.987529, 0.697790, 2.632049, 0.224371, 0.417127, -0.985333},{-0.856328, -0.164982, -1.923588, -0.106485, -0.916320, -0.863788, 1.938958},{-1.513395, -0.325840, -0.386241, 0.091603, -1.372935, 0.727862, 0.678048},{-0.110492, 1.318939, 0.429892, 1.633544, -0.562334, 0.359295, -0.706253},{0.784580, -0.244021, 0.496775, -0.914390, -0.843581, -0.129171, -1.167945},{0.718002, -0.162454, -0.592957, 0.947290, -0.049035, -0.640964, -0.160434},{-0.490040, 0.260909, -1.127846, -0.256342, -0.405375, -0.620358, -1.931227},{1.598250, -0.463649, -0.703460, -0.685239, -0.985365, -0.735958, -0.427023},{-0.789941, -0.583889, 1.357451, 1.675100, 0.014726, 0.121000, 0.055109},{-0.493583, -1.111959, -1.118415, 1.015206, 0.170325, -0.813726, -0.361539},{1.061134, 0.933324, 0.923047, 0.420452, -1.264901, 0.718701, -0.263710},{-2.399331, 0.325659, -1.063102, -1.066663, -0.830244, 0.713011, -0.020273},{-1.400856, 0.023729, -1.719881, -0.264266, -0.434888, 1.245988, -0.808568},{1.222224, 0.460709, -1.382345, -1.493025, 0.207823, -0.116907, -0.623962},{0.860033, 0.574352, 0.304327, -1.379313, -1.440501, 1.295846, -0.900064},{2.211847, -1.701886, 0.521990, 0.115596, -0.300330, 0.731062, -0.881469},{-0.022994, -1.397888, -0.434226, 0.736470, 0.191691, -0.112316, 1.098028},{0.044606, 0.046633, -0.389366, -0.452068, 0.842511, 0.912063, -0.019495},{-2.005528, 1.662364, -0.364564, -1.154008, -0.131214, -1.911225, 0.394132},{0.581468, -1.686960, -0.504944, 0.106661, 1.290826, 2.981258, -1.266516},{-1.130604, 0.652234, -2.147237, -0.931522, -2.008409, -0.711744, 1.264522},{0.098328, 0.015217, 0.495506, 1.477835, -0.392405, 0.908225, 1.387483},{0.359106, -0.373437, -0.658742, -0.051012, -0.988605, 0.478741, 0.739816},{-0.287963, 0.542081, -0.579087, -1.301615, 0.173119, 0.035659, 1.118528},{0.386952, 1.366274, 1.248697, -0.893405, 0.451531, 0.484759, 0.086962},{0.039900, 0.459154, 0.358596, 1.092043, 0.058023, -0.480477, -0.288844},{0.474066, -0.701636, 2.036203, -0.650024, -1.003029, 0.921420, -0.640723},{-0.937910, 1.815473, 0.659763, -1.023808, -1.488768, 0.962003, -1.252618},{1.599038, -0.525976, 2.191653, 0.493904, -0.078034, -0.403012, -2.075748},{1.435153, -0.294979, -1.296920, 0.216162, -0.921983, -1.396356, 0.770544},{0.736848, 0.701671, -1.058939, -0.626119, -0.655369, 1.893018, 0.309980},{0.181005, -1.180631, -0.349039, 1.032156, 0.052419, -0.648261, -2.055199},{-1.923629, -0.794624, 1.398806, -0.886746, 0.971320, 0.004773, -0.478741},{-1.461160, -0.567663, -0.832404, 0.170202, -0.422957, 1.090579, -1.693128},{0.723894, -0.472615, -0.767311, 0.118827, 0.805466, -0.553046, -0.171484},{0.077661, -1.341849, 0.738279, -0.303191, -0.357322, 1.361789, 0.864425},{1.473760, 1.707980, -1.329635, 0.081595, -0.474981, -0.632715, 1.471949},{-0.523617, -0.103181, 0.053177, 0.347285, -1.208578, -0.108332, 1.012540},{0.155126, -0.818568, -1.457909, -0.648343, 0.604239, -0.449050, 1.289403},{0.802331, -2.225819, -0.535148, 0.746443, -0.364574, 3.510071, 1.423758},{0.143706, 0.030658, 1.602318, 1.492474, 0.028683, 0.632314, 0.551744},{0.017552, 0.035297, 0.788257, 0.325389, 0.299926, -0.095345, -1.089239},{-0.471726, 0.071688, -0.054366, -0.630815, -2.719858, 0.224895, -0.819155},{1.429760, 0.408048, -0.408770, 0.680452, 0.523604, 1.583361, 2.225636},{2.339496, 0.389922, 0.132121, 0.130806, 1.686314, -0.346674, -0.415659},{1.333628, -1.187619, -0.114321, -0.150679, 1.669474, 0.054075, -0.976401},{1.917468, -0.462275, 0.950527, 0.933664, -0.021432, -1.731464, -0.797593},{2.366277, 0.126632, -0.148299, -0.620509, 0.231548, 0.600117, 0.418644},{0.928536, 0.990864, 1.619224, 1.876847, 1.444699, -0.935600, -2.089379},{-1.529289, -0.019855, 1.767870, -1.012517, 0.849101, -1.071718, -0.135399},{1.886165, -0.780640, -0.983804, 0.156097, -2.129428, -0.478902, -1.517826},{0.623523, 0.659839, -1.058427, -0.853844, 0.425675, -0.742869, 0.069555},{-0.388984, 1.393329, -0.603557, 0.545044, 1.554999, 0.155213, 0.736148},{0.604009, 1.984269, -0.444844, 0.156538, 1.141190, 1.872372, -0.238359},{0.463423, -0.102266, 0.026630, 1.771819, 0.510986, -1.197789, -0.226440},{-0.230775, -1.461847, 1.244950, 2.198602, 1.272295, -0.882755, 1.089183},{2.568325, 0.976254, -0.776853, 1.068797, -0.555135, 0.316278, -1.374590},{0.275816, 0.557834, -0.846624, 1.203818, -0.361034, 0.343956, 0.655517},{-0.205209, 1.460816, -0.601784, -0.660760, 0.734202, 0.371151, 2.200146},{-0.079205, -0.451646, 0.824464, 1.375771, -0.746441, 0.649791, -1.874775},{0.400320, 1.732331, -0.706162, 0.829080, 0.478262, -1.521817, -0.418684},{1.219180, -0.820053, -0.590840, -1.101892, 0.280547, 0.823794, 0.704814},{-0.071490, 0.261135, -1.005643, 0.643230, -0.442009, 0.435637, 1.700752},{0.855013, -0.071610, 1.150560, -1.206666, 1.521070, -0.765905, -0.151209},{-0.087036, -0.147290, 0.768744, -0.150590, 0.590573, 0.347717, 1.102826},{-0.260531, 0.090338, 0.465977, 1.310283, 0.066287, -0.638606, 2.553212},{0.944283, -1.579177, 0.078065, 0.406915, 1.864793, 0.553794, -1.498492},{-1.743748, -0.527654, -0.790443, 0.134772, -0.254456, 0.938389, -0.316357},{0.540279, -0.419921, 0.937961, 1.006033, -0.385264, -0.633390, 0.354756},{1.260616, -0.751943, -0.566955, 0.520306, -0.361526, 2.376514, 1.961396},{-1.005453, -0.250850, 0.541786, -0.507889, 1.590989, 0.504680, -0.734158},{0.125548, -0.897586, 0.003694, -1.064458, -0.597759, 1.916597, -0.259609},{1.852868, -1.530968, 1.571170, -1.012307, -1.933668, -0.999672, -1.240902},{-0.816790, -0.463458, 1.091273, -0.895273, 1.057480, -0.301578, 1.492671},{-2.195769, 0.683430, -0.175820, -1.701694, -2.041538, 0.078157, -0.025608},{-1.633354, 0.939373, -2.136360, -1.461396, 0.044205, 0.325939, -1.415377},{1.659444, -0.816725, 0.340734, 0.065312, -0.036103, -0.333001, -1.528283},{-0.976885, 0.121519, 0.185940, -0.055395, -0.869717, -0.783079, -1.128637},{-0.552813, -0.741295, -0.715611, 0.317225, 1.093487, 0.097047, 0.063166},{-1.887794, 1.706422, 0.101451, -0.509624, -1.385393, -0.935605, -0.247915},{0.294379, 1.767696, -0.460634, 0.658742, -0.467090, -1.017388, -0.529496},{0.004844, 0.979150, 1.979802, 0.173171, 0.738888, -0.289423, -1.916007},{0.440303, 0.949239, 0.083017, 1.401006, -0.145206, -0.162690, -0.701881},{1.651233, -0.079478, 0.046439, 0.753655, 0.522814, 0.235901, 0.646921},{-0.328625, -0.918156, 0.651743, 0.587345, 1.515576, -1.027462, 0.396006},{0.272323, -0.391607, 0.545049, -1.386877, -1.610882, 0.049845, 2.676634},{-0.196660, -2.320627, 0.393571, 0.366014, -0.090023, 0.107941, -0.872754},{-0.838403, 2.095972, -0.036284, -0.514963, -1.322178, 1.843984, 0.828389},{0.688084, 0.912507, 0.498290, -0.392908, 1.862091, 0.894802, 0.799333},{0.449148, -0.099146, 0.215289, 1.195059, -0.319513, -1.484598, 1.686496},{0.156146, -0.639883, 0.208810, 1.242869, 0.302934, -0.298256, 0.389762},{-1.160980, -0.329709, -0.150200, -1.874503, -0.111912, -1.012651, -0.352167},{-0.993107, 0.913313, -1.109127, 2.106486, 1.176778, -0.503507, 0.478732},{0.011921, 0.065373, 1.482591, 0.174096, 0.121081, -0.306923, 0.099060},{0.252067, -0.578217, 1.582312, 0.325913, -2.545859, -0.790695, -0.446196},{-0.134895, 0.148936, 0.783478, 0.049743, -1.321263, -1.144935, -0.638682},{-2.356383, 0.447850, -0.479947, -2.266700, -0.170322, 0.691127, -1.086646},{-1.290984, 0.390828, 0.933817, 0.085286, -0.884666, -0.051395, 1.159367},{0.144917, 0.170292, -0.981765, 0.001507, -0.074158, 0.549908, 1.237819},{-0.594210, -0.773635, -0.064529, 1.780218, 1.275236, 1.242064, 0.017725},{0.647193, -1.117697, -0.342581, -0.809927, -1.685181, 0.131338, 1.233573},{-0.783809, -0.976914, 0.026865, 0.695424, -0.356047, -0.006500, -1.659003},{-1.481295, 0.733782, -0.042271, 1.745956, -1.120502, -0.693116, 0.657306},{-0.714687, 0.766359, 0.160112, 1.037358, 0.133584, -0.136217, -0.125543},{0.366236, -0.164925, 0.409341, -0.897293, -0.138362, 0.197437, -0.338614},{-0.413607, 0.763857, 0.096189, 1.598626, 0.056830, 2.086921, -1.592401},{-0.428229, 0.480918, 1.055549, 1.325615, 0.843833, 0.923126, -0.384048},{1.572075, 0.353047, 0.893921, 0.093463, -0.719554, -0.225838, 0.012662},{-0.947471, 2.568060, -0.358874, 0.140897, 0.673453, 0.331575, -0.357591},{0.739156, -0.969089, -0.517895, 0.937776, -1.655825, 0.913012, 1.547511},{-1.149874, -0.262048, 0.838120, 0.222006, 0.560506, -0.990305, -1.592397},{0.125346, -0.750243, -0.295433, 0.314966, 0.002818, 0.219661, 0.641284},{-2.373867, -0.840042, -0.374141, 0.329599, -2.347408, 0.030948, 0.668475},{-2.722125, -1.779440, -0.270352, 1.156823, 0.531257, 0.374354, -0.302376},{2.599012, -1.204664, 1.669883, 1.686223, -0.389173, 0.898402, -0.078603},{-0.622568, -0.436549, 0.650545, -0.677776, 0.961231, 0.097834, 0.079367},{0.247423, -0.711720, 0.148401, -0.196230, -1.606994, -1.636457, 0.450761},{-0.910571, 0.848408, -0.379953, -0.806577, 0.418819, -1.533655, 1.026663},{0.417410, -0.314157, 0.242169, -0.330325, 0.960292, 0.262885, -1.298214},{0.430821, -0.499274, 0.833538, -1.234797, -0.472654, 0.192354, 0.202949},{-0.833969, -0.349698, -0.889692, 1.400488, 0.076413, 0.356009, 0.660167},{0.869635, -1.473666, 0.491642, -0.259324, -0.171282, -0.775467, -2.553835},{0.966348, 0.505510, 1.391385, 0.614410, 0.430354, -0.336178, -0.766763},{-0.665815, 0.170266, 0.154592, 0.968455, 2.236509, 1.156961, -0.303058},{0.334135, 0.402606, 2.408797, 0.837190, 0.225372, 2.385513, 0.364963},{-0.313021, -0.729428, -0.599060, 0.642403, -0.839383, 1.214397, 0.181147},{1.486532, -0.659308, -0.489405, -0.812271, 0.900241, -0.645651, 1.439764},{-0.565328, -0.098316, 0.136869, 1.067933, -0.590216, 2.098279, 0.076707},{0.301342, -0.659410, 0.261302, -1.804207, 1.495543, -0.201339, -0.514655},{-0.581834, -1.310191, 1.900218, -0.536388, 0.327947, 1.144470, -0.279026},{-0.031252, 0.935840, -0.749898, 0.753876, 1.108579, -0.596435, 0.175364},{0.505324, -0.181396, 0.913703, 1.115349, 0.732146, 0.468820, 0.717443},{0.508297, -0.246669, -0.790990, -1.139578, 0.490307, 1.314183, 0.451228},{1.582742, -0.367066, -0.886605, -2.122605, -1.527041, -0.513382, 0.855130},{-0.045886, -0.619012, 1.364512, -1.235426, -0.427484, 0.339898, 1.068303},{0.075325, -0.833418, -1.614671, 0.466686, -0.077467, -0.842454, 0.579489},{1.909352, 0.841099, -0.504528, 0.816101, 1.030680, 1.319795, -0.473648},{1.896089, 0.852468, -0.646187, 0.792239, -0.481251, 0.761973, -0.039724},{-1.126297, 1.051049, 1.047335, 1.199968, -0.932105, -0.928620, 0.017470},{0.335865, -0.250383, 0.709538, -1.806790, -0.334398, -0.691980, 0.653228},{1.592312, -1.272683, 0.226387, 0.391254, 0.082100, -1.070187, 1.090788},{-0.524333, 0.329089, -1.320911, 0.864324, -0.047126, 0.826128, -0.698246},{1.283037, -0.270708, -0.618773, 1.347108, 0.309351, -1.163061, -0.503010},{2.387042, -1.432546, 0.475512, 1.043241, -1.556086, -0.664262, -2.101372},{-0.125563, 0.174918, 0.668838, 1.674966, 0.126184, -0.492665, -0.209420},{-1.398418, -1.238796, 1.325366, -0.019846, -2.585681, 0.544317, -0.388759},{-0.590679, -0.378093, 0.227610, -0.106719, -1.108148, 1.351496, -0.011837},{-0.007455, 0.114209, -0.059510, 0.985654, -0.014450, 3.583395, -0.405261},{0.572973, 1.169492, -0.211583, 0.968670, -0.535682, 0.067161, -1.179047},{-0.443672, 0.542965, 0.766420, 0.643765, -0.170889, -0.161672, -0.530124},{-1.160527, 0.429655, -1.524649, 1.508697, -0.103504, 0.937647, 1.506842},{-0.067914, -0.629034, -0.471168, 0.111646, 0.178354, -0.264669, -0.612678},{-1.161563, 1.952873, 0.756087, 0.218417, -0.726234, 0.590113, 1.031893},{0.543820, -0.516284, -1.052750, -1.260110, -1.527920, 0.214170, -0.182338},{-0.121758, -0.815532, 0.314348, 1.379188, -0.873025, -0.578771, -0.901109},{-2.192180, 0.187018, -0.651078, 0.566965, -1.518475, 0.138137, -1.088604},{-0.841373, -0.060132, 2.139828, -2.100383, 0.486391, -0.406786, 0.246618},{-1.888973, 0.359273, 0.598334, -0.988010, 0.536376, 0.783524, -0.880969},{-0.384985, 1.144007, 0.006807, -1.052507, -1.230248, 1.306632, 0.474791},{-1.004748, 0.244856, -0.723266, -0.030241, 0.386077, -1.252433, 0.249113},{-0.201196, 1.619041, -1.429649, -1.860189, 0.937733, -0.385074, 0.240402},{1.057435, -0.878915, 0.497283, 0.772408, 0.991781, -0.247919, 0.011815},{-0.070489, 1.818262, 0.742590, -0.119238, 0.861835, -0.983072, -2.590870},{1.210837, -0.442291, 0.787380, 0.391089, -0.881173, -1.586614, 2.099773},{-1.216572, -0.207177, 1.660655, -0.068587, -0.087533, 0.807353, -0.176067},{0.082179, 1.358015, -0.540523, 0.510475, 0.352361, 1.139199, -0.362435},{-0.394291, 0.740089, 0.179741, 0.422384, 0.461070, -1.440486, -1.340335},{-0.098955, 0.060399, 0.341673, 1.483423, 0.596232, -0.065565, -1.690629},{0.528583, -1.373099, 1.269851, -0.345083, -1.785436, -1.571525, -0.043229},{-0.092001, 0.110141, -0.542484, 1.162877, -0.773155, 1.024540, -1.140805},{0.833878, -0.654159, 0.963547, 0.657100, -0.055674, -0.945653, 1.480198},{0.113786, 1.422485, 2.077114, 0.410791, 0.179718, -1.200369, -0.299042},{-1.099789, 0.134614, 0.854967, 1.237207, 1.393551, -0.018896, 0.906992},{-0.828720, 0.602878, 1.339032, 0.193226, 1.614078, 0.155280, 0.346504},{-0.974753, -1.359847, -0.334717, -1.202595, -0.047418, 0.205667, -1.643392},{0.436701, 1.473888, -0.063382, -0.164582, -0.331003, 1.052734, 1.137938},{-0.579169, -0.421115, -1.895438, 0.896208, 0.086384, 1.384463, 0.609785},{-1.171002, 0.223965, -0.420502, -1.125152, -1.207641, 0.725889, 0.452743},{0.609778, -0.846346, -0.044875, -1.000408, -0.875823, -0.541273, -1.695238},{-2.120376, 0.031340, -1.595520, -0.086556, 1.907109, -0.189592, -0.615180},{0.376487, 0.700573, -0.532395, -0.950324, -0.521511, 1.510889, -0.255839},{-0.782496, 0.787834, 0.666886, -0.510814, -0.046596, 1.463833, -0.341821},{0.639250, 0.364276, -1.090366, 0.326134, 1.301855, -0.157619, 0.639978},{-0.579713, -0.620085, 0.142464, -0.253642, -1.787603, -1.677742, -0.546659},{-0.549738, 0.112080, -0.438710, 0.786245, 0.677639, -0.186529, -0.020600},{-0.707305, 0.295040, -0.038845, -0.597371, 0.176653, -1.239234, -0.645602},{0.825973, 0.985560, 0.561160, -0.356871, -0.171271, 0.981152, -0.246067},{0.378022, 1.687600, 1.001096, 1.087936, 1.838274, 0.321584, -2.787924},{-0.691452, 0.971814, 0.124118, -0.076990, 1.047764, -0.014313, 0.495558},{0.826298, 0.610194, -1.659642, 0.742283, -0.890510, 0.887961, -1.198712},{0.630508, 0.954373, -0.681734, -0.341224, 0.082014, 1.573123, -0.778590},{-0.650865, -1.697061, 0.546636, 1.331861, 0.233593, -0.712997, 0.787000},{-0.292655, 0.869428, 0.422773, -0.160669, -0.161204, -0.369633, -0.526737},{-0.205869, 1.449137, 0.711800, -0.217418, -0.278979, -0.254499, -0.360744},{1.463542, -1.014050, 0.744538, -0.936142, -0.812183, 1.611570, -1.866311},{1.580078, -0.509763, -0.464897, -0.385135, -1.047948, 0.935622, 0.519500},{-0.242393, -0.282697, 0.742072, -2.389960, 0.875627, 0.095427, 1.507335},{-1.376617, -0.992109, 0.519893, 0.855874, 1.773230, 0.492410, -0.039355},{-1.616996, -0.782331, -0.725594, -0.709061, -1.486659, 1.528640, -1.110630},{-0.162309, -1.038395, -0.606613, -0.419562, -1.191241, 0.720380, 0.276000},{1.847847, -0.900666, -1.278623, 1.346086, 1.231562, -0.849895, 0.630857},{-0.926885, -1.314299, -1.370441, -0.347322, 1.114872, 0.778124, -1.676564},{0.369029, 1.113315, 0.355317, -1.027640, 0.199944, 1.564721, -0.464147},{-0.374247, -0.215787, -0.184416, -1.161292, -0.213150, -1.066694, -0.999130},{0.149204, 1.008430, -0.174250, -0.845292, -0.991938, -1.443897, -1.532516},{1.311644, -1.627291, -0.931038, 0.225779, -0.235379, 0.807175, 0.058336},{0.788077, 0.106345, 1.680260, 0.184680, 0.019493, -0.017912, -0.543211},{-1.052667, -0.984577, -1.417450, 1.377888, 1.127549, 0.342544, 0.268486},{-0.514629, 0.167328, -0.390171, 1.200297, 0.020480, 1.284708, -0.014686},{0.418827, 2.468934, -0.966880, 1.157148, 1.147669, -0.570076, 0.466701},{-0.445491, 1.113229, -1.637583, 1.439434, -0.826571, 1.350777, 0.393596},{0.277486, -1.524055, -1.954621, -0.870544, 0.259833, 1.446600, -0.367071},{-0.150547, 0.708721, 0.754515, -1.157172, -0.095736, -1.124476, 1.010216},{0.118588, 1.310647, 1.232678, 0.222572, -0.986542, 0.407983, -0.077158},{0.779715, 1.177703, -1.832292, 0.627285, 1.840677, -0.925107, 1.116560},{-0.269453, -1.952834, 0.335917, -0.044774, -0.532434, 2.503012, 0.986004},{0.496942, -0.210241, -0.672640, 0.629065, 1.446437, -0.249890, 0.375483},{0.571092, 1.168717, -0.017738, -0.624725, 0.600329, -0.997970, -1.069914},{-0.566594, -0.654693, 0.697337, 1.674824, -1.000014, 1.030198, -0.779925},{-0.046722, 0.202985, 0.291466, -0.365564, -2.431851, -1.824343, 0.402025},{-0.738728, -0.576233, 0.668386, 0.955759, -0.027783, 0.262575, 1.463633},{-1.016724, 0.530653, -0.044522, -0.221590, 0.017011, 0.435181, -0.513975},{-0.137611, 0.519489, -0.064418, -0.692607, 0.838056, -1.318082, 1.137365},{0.098038, -0.639585, 0.149548, 0.863750, 0.420348, 1.109694, -0.209790},{1.331477, 0.621337, 0.528091, -2.034079, -0.701169, 0.103004, -0.499510},{-0.132547, -0.075997, 0.687360, 0.881570, -0.447414, 0.406365, 2.420723},{-0.531586, 0.600516, -0.997105, -0.423448, -0.028437, -1.030394, 0.248768},{-1.390158, -0.020512, -1.578956, -0.776923, -1.513279, 0.793067, -1.583801},{-1.058356, 2.912935, 0.012597, -0.014807, 0.359948, 1.097666, -0.766377},{-0.228880, 0.132658, -1.232040, -0.204113, -1.792610, -1.473194, 2.886494},{-0.429964, -1.055285, 0.831695, 1.096219, -1.720957, 1.062396, -1.064951},{1.024588, -0.753693, -0.217846, 0.026543, 0.385261, 0.913343, -0.064793},{0.656781, -0.362226, -1.915231, -0.311897, 1.436219, 1.352195, -0.157180},{-0.194535, -0.757716, 0.094380, 0.720463, -0.731738, 0.618440, -0.118537},{1.111583, -1.096044, 0.320910, 1.199405, -1.110540, -1.036158, -1.101022},{-0.716035, 0.019436, 0.619120, -0.101260, -0.932622, -0.748216, -1.649944},{0.482363, 0.373587, 0.642945, -0.062450, -1.804080, -1.459024, -0.223075},{-1.579070, -0.193485, 0.314985, 1.291953, 1.518294, 0.649800, -0.390129},{1.042018, -0.117954, -0.002853, -0.283773, 0.554244, -0.143024, 0.302784},{-0.920683, 1.713459, -0.725559, -1.125786, 1.126364, -1.388778, 1.526253},{0.422813, -0.191137, 2.117993, 0.562575, -1.101117, 1.142492, -0.160442},{1.544251, 0.450856, -0.214425, 1.049592, -1.746009, 0.553591, -2.866272},{-0.081045, 0.870359, 1.675528, -0.196040, 1.907223, 0.649252, -1.385890},{0.964813, 0.560002, 0.011685, -0.692970, -0.267330, -0.563744, -0.209274},{1.451164, -0.809155, -0.746020, -0.224987, 0.434982, 1.297707, 1.769769},{-1.185500, 1.055720, 0.015189, 1.360245, 0.668878, -0.155481, -0.403502},{0.263225, 0.873541, -1.416755, -1.132063, -1.609901, -1.787090, 0.034396},{0.375960, 0.321398, 1.923676, -0.974203, -1.374189, -0.797288, 0.412115},{-0.452154, -1.285171, 0.880859, -1.179626, -0.780669, -0.134404, -0.045034},{0.736956, -0.030829, 2.204822, -0.842718, 0.881804, 0.116970, 0.024201},{1.879644, -0.317850, -1.744977, -0.730895, 0.567314, 0.700316, -1.761312},{-1.487213, 0.092474, -0.975087, 0.169426, 1.695369, -1.269037, 0.410440},{1.479576, 0.001627, -0.094069, 1.841274, 0.252786, 0.623843, 0.418894},{-0.322819, 1.027847, -0.747841, 0.409546, 2.082564, -0.541727, 0.284221},{-0.072633, -0.300699, 1.224934, -0.165675, -0.571469, -0.345711, 0.263493},{0.578659, -0.195247, 0.097176, -0.265968, 0.210839, 0.160243, -1.346968},{-0.710711, 0.346750, -0.944528, -0.665557, -1.872976, 0.954336, -0.509576},{-1.663545, 0.093317, 1.156692, -0.964948, 0.958201, -0.218665, 0.414624},{-0.179740, 1.293099, 1.140934, -0.057700, 0.550875, -0.394405, 0.118465},{-0.290535, -0.628065, -0.481454, 0.858114, -1.469583, 1.038828, 1.112766},{-1.660253, -0.250112, -0.700152, 0.514136, -0.352279, -0.806765, -0.820988},{0.655051, 0.800212, 0.799250, -0.703555, -1.207809, 1.930603, 1.140091},{1.727644, -0.525769, 0.480806, 0.253377, 0.193587, -0.417532, 0.739556},{0.462522, 0.710414, -1.206044, -1.217789, 1.684123, 0.141910, -0.237512},{-0.355743, -0.032735, -0.123892, -0.173773, 1.817428, -0.017333, -1.181918},{-0.500942, -1.238001, -1.127298, 0.702471, -0.098901, -1.217629, 3.105519},{-2.188273, 1.954288, -0.409291, -0.805162, 1.596809, 0.676137, 0.428647},{1.196926, -0.826175, 1.458455, 0.551911, -0.294652, 0.532300, 1.316814},{1.111311, -0.833057, -0.133232, 1.115789, 1.502341, -0.141074, -1.024917},{-1.364107, -0.289290, -0.697976, 0.246295, 1.408121, -0.776810, -0.638590},{-0.355755, -0.392857, -0.262324, 1.172829, 0.344726, 0.048093, 0.395932},{2.112790, -0.139280, 0.863001, 1.149583, 0.077433, 1.035234, -0.302485},{1.603379, 0.015865, -0.993136, -1.554852, -0.214737, 0.608332, 0.804528},{0.111464, 1.549761, -1.136447, 0.485348, 1.258404, 1.575091, -0.642022},{1.622131, -0.596027, 0.014602, -0.133079, 0.459914, 2.104326, 1.511360},{-0.231614, 0.613880, 1.236062, -1.788514, -0.114601, 0.015759, 1.328932},{1.646846, 0.215077, 1.003168, 0.418294, -0.789938, 1.448563, 0.381642},{1.118957, -0.498431, -0.760539, 0.753561, -0.108882, 2.056718, 0.533354},{1.364314, 0.081446, 1.290072, 0.230243, 0.810215, 0.010378, -1.493307},{-0.650793, -1.027351, 0.333580, 0.883410, 0.326150, 1.079764, -1.053928},{0.225145, 0.052773, 0.532669, -1.604536, -1.359763, 0.008561, 0.992941},{-0.311237, 0.890311, -1.331407, -0.255279, -0.168253, 0.019883, 0.502016},{0.618729, -1.187355, 0.351657, 0.184342, 0.963985, 0.602273, -0.900425},{1.252674, 0.329250, 2.379768, -0.677949, -0.065996, -1.343165, 0.842885},{0.604609, -0.883130, 2.259946, -0.170127, -0.155286, -0.249628, -0.875567},{1.003334, -0.251627, 0.779432, 0.212636, 1.602845, -1.580160, -0.023888},{-0.840977, -0.154366, 1.260342, -2.517648, 0.635044, -0.913030, 0.957728},{-0.631109, 1.341884, 2.363603, -0.530084, -0.266727, -0.999811, 1.360834},{1.446578, -0.335777, -1.103792, -0.216624, 0.092819, 0.081403, 0.645329},{1.208589, -1.949450, 0.629968, -1.637936, 1.376396, 0.401106, 1.058642},{1.041535, -0.757659, 0.221457, -0.198989, -0.274339, -1.358674, -0.133231},{0.402880, -0.828164, 1.448017, -0.898684, 2.477712, 0.721095, -0.697573},{1.918892, -0.287782, -0.924054, -1.637901, -0.520462, 0.868195, 0.623969},{0.740774, 1.052655, 0.194402, -0.178383, 0.610329, -0.879922, -0.305528},{0.464485, 0.486877, 0.324341, 1.550979, -1.032604, 0.124101, 0.419144},{0.711385, -0.833168, -2.169771, -0.471449, 0.156908, -0.441985, 0.147171},{0.161172, 1.391724, 0.711148, -0.464421, -1.283993, -0.696378, -0.954901},{0.272988, 0.644457, -0.972786, 0.628977, 0.123134, -0.340536, -0.815202},{-1.808599, 2.160339, 2.468143, 1.608286, 0.140299, -0.439443, -3.058830},{1.429060, 0.277695, 0.239631, 0.560248, -0.620557, -0.382259, -0.175579},{-0.680704, -0.645531, 0.382234, -0.660087, 1.710267, -1.225698, -0.022818},{0.641464, 0.819244, -1.674807, 0.645731, -0.043380, 1.414253, -0.944796},{1.417929, 0.760079, 1.002597, 0.427416, -1.688930, -1.332649, -0.939255},{0.666910, -1.882473, -0.932239, -0.268205, 1.497640, 0.642068, 0.165834},{-0.415266, 1.297354, 0.591316, -0.761113, -1.273574, -0.106520, -0.672929},{-1.217425, 0.168379, 1.084008, 0.191293, 0.570822, -0.032038, -1.474377},{0.823122, -0.377278, -0.197179, 0.693666, -0.621226, 0.550080, -0.429715},{2.154452, -1.991161, -0.066290, -0.097389, -0.755062, -0.778095, 0.461835},{-1.045198, -0.281974, 2.052294, -0.196310, -0.722158, 1.340051, -1.591257},{-0.646447, -0.329335, 0.945307, -0.701732, 0.909255, -0.137945, -0.788480},{0.701023, 2.242285, 1.084766, 0.071367, -0.767643, -0.804504, -1.870839},{-0.526825, 0.630142, -0.535266, 0.420900, -0.253450, -1.296852, -0.437947},{-0.242845, -1.626177, 1.160154, -0.920059, -1.948293, -1.593093, 0.280268},{-0.181057, -0.837898, -1.560244, 0.805115, 0.881685, 0.213731, 0.538298},{1.746315, 0.142857, 0.638878, 0.593883, -0.157773, -1.182078, -0.139232},{0.189605, -0.213329, 1.416746, -1.344616, 0.969581, 0.985398, -0.428308},{-0.668954, -0.935914, -2.383308, -1.376194, -1.519618, 0.836874, -0.390262},{1.274418, 0.009304, 0.930936, -1.273187, 1.604215, -0.609800, -1.194251},{0.409833, 2.185471, -1.221696, 0.156568, 0.032119, 0.049919, -0.469743},{-0.239922, 0.351387, -0.513157, -0.408124, -1.077284, 0.602938, -1.336373},{0.751827, 0.771264, 1.059860, -0.644082, 0.580808, 0.193658, 0.366777},{-1.031413, 0.066758, -0.594288, 0.010157, 0.042768, 1.227046, -0.008178},{0.669518, 1.739707, -2.179440, 0.402232, 0.279430, -0.578746, -1.291007},{-2.146904, -0.863288, -0.906310, 1.842107, -1.308415, -0.377511, -0.154392},{0.964048, 0.872244, -1.409876, -0.750878, 0.326109, 0.084809, 2.367475},{-1.598141, -0.040032, 0.205867, -0.001777, 0.970515, -0.247998, -0.534404},{1.502642, -2.093241, -0.506920, 2.065748, -1.358656, -0.523190, 0.229464},{0.229548, -0.488789, 1.671141, 0.316218, 0.692884, 0.041974, -1.770097},{0.468974, 0.199006, 0.699836, 0.664759, -1.687652, 0.463920, -0.142209},{0.396890, -0.748262, -2.142496, 0.693432, -0.485506, 0.250624, -1.821976},{0.861138, -0.591273, 0.309960, -0.890906, 1.088569, -0.430016, -1.404786},{-1.116173, 1.303696, 0.381518, -0.852147, -1.230014, -1.955840, -0.195110},{-0.813204, 0.946909, 0.025849, -2.667316, -0.178436, -0.133961, -0.024967},{0.622535, -0.144331, -0.386134, 0.682968, -0.945468, 0.356595, -2.005858},{0.900095, 1.348675, 1.420443, -0.692755, -0.351037, 1.037607, 0.131440},{-0.311115, -0.268125, -0.030584, -0.152322, 0.443975, 0.710729, -0.678558},{0.544037, 0.458128, 1.630484, -0.617879, 1.730788, 0.299600, 0.419872},{-0.125847, 0.036572, 0.959676, 0.924072, -0.895751, 1.540244, -0.368437},{1.202868, -0.882893, -3.104302, -0.624959, -1.670797, -1.484658, -1.267901},{0.923743, 0.081366, -1.206726, 0.273804, 0.337636, -0.551475, -0.987174},{-1.464900, -1.013000, -0.322419, 0.242652, 0.090435, 0.605928, -0.091525},{-0.408777, -0.235021, 1.650039, -0.455105, -0.784487, 0.202499, 0.920106},{-0.807496, 1.404284, 0.730570, -0.596847, 0.325632, -0.524987, 1.462103},{1.982006, -0.097317, -1.062919, -0.010124, 2.761115, 0.627402, 1.940401},{0.728811, -0.775483, 0.463613, -1.397040, 0.523859, -1.388243, 0.014706},{-0.194638, -2.130282, 1.421372, 0.023659, -0.067326, 1.253030, 0.565440},{0.483161, 0.535397, -0.960117, -1.477871, -1.309270, 1.442580, -0.489574},{-0.230321, -0.451830, 0.166908, -1.191910, 1.113002, 1.483536, -0.199709},{-0.079043, 0.722527, -1.618153, 0.121871, 0.379052, -1.639137, -0.401530},{2.655427, -1.032201, -0.185368, -0.110583, 0.048445, 2.966255, -0.314679},{-0.616558, -1.622212, 0.879255, -0.444415, 1.540326, 1.468482, 0.406067},{-0.070930, 1.737266, -1.658624, 0.366773, -0.098451, -0.382180, -0.162314},{-0.953588, -0.323903, -1.155324, 1.481265, 0.415621, -0.778504, 0.112099},{-1.456527, 0.005121, 0.495916, -0.217189, -0.724334, -0.397628, 0.192828},{-0.132245, -0.027368, -0.461927, -0.512487, 0.047721, 1.009462, 0.413118},{0.053327, -0.720514, 1.152624, -0.265299, -0.190927, -0.415434, 0.305600},{0.244064, 0.423652, 1.521946, 0.706732, 1.405587, -0.346817, 0.023163},{-1.638401, 0.943213, 0.096775, 1.518806, 0.853090, 0.013663, -0.315974},{0.318745, -0.266185, 0.062126, 0.452984, 1.473799, 1.796663, -0.098115},{0.262510, -0.869036, 0.511108, 0.989014, 0.568451, 0.806601, -0.458280},{0.355894, 0.182219, -0.063719, -0.632112, 0.557423, -2.322294, 0.001038},{1.965623, -0.794128, -0.079020, 0.181315, -0.697318, -0.637551, -1.044022},{-1.808945, 0.650200, 0.769303, -0.208905, 1.235302, 0.596830, 1.755683},{0.449290, 1.137421, 1.628057, -1.589554, 0.867161, 0.681358, -0.087417},{1.273686, 1.310590, -1.434337, -1.534512, -0.608355, -1.160559, -0.750622},{-0.109317, -0.803564, 1.611018, -0.106347, 1.000482, 0.279460, 0.726260},{-0.430243, -1.167938, -0.953408, -0.237114, 0.026685, -0.567874, -0.256457},{-2.211482, 1.501912, -1.418488, -1.151781, -0.677290, -0.379759, -0.456854},{-0.212572, -0.058903, -2.437380, 0.802419, -0.360705, -1.274453, -0.659841},{-0.324907, 0.055995, 0.031072, 0.199816, -1.420497, 0.692408, -0.548599},{0.185948, 0.678665, -0.300351, -0.403294, 1.436971, -0.217367, 0.863663},{0.580347, -0.622446, -1.263227, 0.015598, -0.219872, 0.994799, 0.837429},{1.519521, -1.100232, -0.207786, -0.068198, -3.486518, 0.574076, 0.602452},{-0.762680, -0.388887, 1.302593, -0.410562, -1.992299, -0.196844, -0.174896},{1.980966, -0.637087, -0.192104, 0.252567, -0.521861, 0.522455, -0.373255},{-0.103312, 2.031953, -1.090373, 0.219360, 0.673711, 0.355331, -1.022865},{-0.290675, -0.463248, -0.359603, -0.353734, -0.883181, 1.400651, -0.683986},{-0.136651, 1.849696, 0.137033, 0.729366, -0.914155, -0.259303, -0.695360},{1.032792, -3.297065, 0.755177, 0.783118, 0.698619, -0.260930, 0.190373},{0.177565, 0.549580, 0.080543, 1.122655, -0.510597, -0.357787, -0.000750},{-1.012818, -0.301801, 1.371779, -0.116363, 0.281219, -0.450259, 2.357900},{-0.845509, 0.049449, 0.536695, 0.271956, 0.621818, -1.312113, 0.228582},{0.004957, -0.989368, -0.093194, 0.079637, 0.551420, 0.582691, 0.957448},{-2.572688, -0.344547, -1.157908, 0.492034, 1.430376, 2.283480, 0.166735},{0.250511, -2.089498, 1.243672, -0.671428, -0.498308, 0.277437, -0.829303},{-0.437393, -0.308240, 0.060613, 0.084194, -2.117869, 1.244114, 3.522937},{-0.089713, -0.651591, 0.479609, -1.633692, 0.881872, -0.471621, -0.026320},{2.120167, -0.043103, -0.755564, 0.656677, 0.411827, 0.094625, -0.127209},{-0.034029, 1.038746, 0.884998, 0.321489, -0.585475, 1.469548, -0.872501},{-0.138013, 0.092759, -0.633958, 2.554908, 0.570775, -0.282479, 0.286460},{0.831761, 1.080078, 0.897961, 2.121403, -0.536227, -1.485155, 0.179191},{-0.002163, -0.704238, -0.885066, 0.339652, -0.321600, -1.012599, -1.903115},{0.273596, 0.547080, -0.124125, 1.257916, 1.186138, 0.024164, -0.530183},{0.747994, -0.371010, -1.089990, 0.166283, 1.148735, -0.265337, 0.714534},{0.968898, -2.620790, 0.990918, 0.488280, 0.298730, -0.095471, -1.273564},{1.317991, -1.125980, 0.410971, 0.218051, 1.254806, 0.646019, -0.195808},{0.501987, 1.409720, 1.520896, -1.639585, 0.410085, -0.517313, 0.576588},{-0.944038, -0.574277, 1.545844, 2.142778, -2.400988, -1.219028, 2.161184},{0.549218, -0.249417, -0.825887, 1.173776, -0.104732, 1.184138, -0.374188},{1.617310, 1.400450, -1.636609, 0.001784, -1.089178, -0.195771, -0.116709},{-0.628328, -0.709918, 0.653317, -0.514048, -0.045932, 1.764770, 0.097371},{0.106139, -2.157382, -0.630987, 0.228466, 0.392794, 0.264820, -0.832996},{-1.644920, -0.118993, 0.656958, -1.156168, -0.797612, 0.543524, -3.050241},{-1.525234, 0.227088, 0.699085, -0.167900, 1.839813, 0.977774, 0.069933},{-0.700030, 1.552301, 0.659485, 0.959916, -0.316185, 0.473113, -1.982778},{1.210889, 0.172932, -0.646273, 1.774618, 1.989075, 0.157853, 1.271159},{0.848524, -0.050672, -0.304015, 0.026580, 1.598687, 0.814053, -0.507787},{-0.909750, 0.607005, 0.064917, 0.641474, 0.535183, -0.378272, 0.230862},{1.262445, -1.279477, -1.113819, -0.603625, 0.398122, -0.640194, -1.690009},{-1.197427, 0.160110, 1.579369, -0.574266, 0.585603, -0.990262, 2.190489},{0.296852, -0.734583, 0.659567, -0.110017, 1.036683, -0.057346, 0.633890},{-0.625663, -1.082703, -0.379024, 1.757718, -0.372596, -0.000864, -1.696257},{-0.833145, 1.094707, 0.350528, -2.108336, -0.139568, -0.082590, -0.131268},{-1.243410, -0.228312, -0.354855, 1.169047, 0.040815, 1.075417, 0.252634},{0.711744, -0.269958, -1.341230, 1.298036, -0.506993, 0.474068, 1.852807},{0.763748, 0.457693, 1.456137, -1.311093, 1.591479, 2.542974, -0.425572},{-1.876611, -0.256806, 0.776593, 0.302883, 0.217222, -0.511255, 1.134839},{-1.679711, -0.397594, 2.919839, -0.994639, 1.294667, 0.815006, 1.377321},{-0.230504, -0.239938, -0.407244, 0.793983, 0.072762, -0.262639, -0.362485},{0.164850, -1.335255, -0.490813, 0.602254, -0.459242, -0.615918, -0.413959},{0.646839, -0.618109, 0.615614, 2.087588, -0.882477, -0.349085, -0.695778},{-0.418722, 0.007731, -0.738020, 0.939692, 0.585467, -0.267247, 0.986094},{-0.326396, -1.125841, 1.618832, -0.692319, -0.118391, 0.550429, 0.676194},{-0.351641, -0.573854, 0.070226, 0.770936, 1.099325, -0.705125, -0.040139},{1.330573, -1.966221, 1.767123, -0.625976, 0.973436, 1.309969, 0.444723},{-1.223942, -0.349585, 0.047026, -0.481430, -0.343779, 0.156128, -0.042409},{2.145686, 0.450828, 1.634487, 0.996033, 0.147413, -0.718434, 2.095250},{-0.708011, -0.412347, -0.629727, 2.286822, -1.342223, 0.268364, -0.054161},{-1.007592, -0.292188, -0.400617, 1.325966, 0.386858, -0.112386, -1.285440},{-0.497232, 0.677469, 0.195470, 1.172745, 0.554076, -1.310408, -0.508631},{-0.037245, 1.110089, 1.120915, -0.981752, 1.071985, -0.869591, 0.002866},{0.244980, 1.011419, 1.599265, 0.314984, -0.470700, 0.675060, -0.285732},{-0.172575, -0.677245, -0.253854, 0.617556, 0.088619, -0.924342, 0.488022},{-0.550000, -0.830264, -0.436860, -2.098804, 0.496448, 1.204359, -0.176149},{0.732329, -0.733209, -2.050964, -1.256133, 0.952807, 0.370377, 0.429491},{1.632812, -0.815162, -0.775753, 0.666297, 1.641503, 2.760329, 0.672022},{0.498033, -0.208130, -0.859736, 0.766657, -0.723880, 0.167811, -0.120710},{-0.327584, -1.321419, -0.669600, 0.086564, -0.214036, 0.996517, -0.709536},{-0.151817, 0.937303, -0.915504, 0.490153, 0.101280, -0.849901, 1.154564},{-2.623467, 0.672517, -0.033177, 2.030983, -0.106642, 1.041366, -2.434895},{0.803683, 0.303306, 1.209433, 0.031618, -1.807852, -1.076092, -0.654071},{0.386622, -0.550515, 0.491729, -0.645936, -0.173810, 0.859615, 0.327509},{2.315245, -0.989543, 0.638977, 0.153672, -0.014024, -1.263058, 1.481189},{1.608968, 1.940847, 0.473697, -0.203444, -0.408176, -0.803431, -0.229627},{-0.090187, -1.235336, -1.526706, 1.544295, 0.480799, -1.666599, 0.610516},{-0.724816, 0.059985, -0.699261, 0.129142, 0.599155, -0.019498, -0.696607},{0.202928, -0.620849, -0.116923, -1.279060, 0.862773, -0.097947, -0.397423},{1.628154, -0.428371, 2.704214, 0.103255, 2.557957, 1.168822, 0.296670},{0.334462, 0.307454, 1.100266, -1.094194, 0.537018, 0.953795, 2.379238},{0.776688, -0.234945, -1.562777, 0.617645, 0.529403, 0.672918, 0.922482},{0.639576, 2.456816, 0.419538, -0.362059, 1.008967, -0.061592, -1.023482},{-1.663417, 0.978396, -1.018166, 0.287622, -0.620214, 0.171886, 0.857802},{0.936275, 1.557288, -0.304449, -1.350482, -0.234428, -0.481491, 0.198655},{-1.401574, 1.359292, -0.230565, 1.616122, -2.127215, 0.662221, 0.236366},{-0.000295, -1.154265, -0.373931, -0.655950, 0.445096, -1.426195, -0.317015},{0.327564, -1.041144, 0.064480, -0.936754, 0.165013, -0.310450, 1.476986},{-0.604540, -0.850853, 1.024998, 0.511418, -0.041050, 2.339761, -0.324558},{-0.105576, -0.374638, -0.153336, 2.170394, -1.303452, -0.507706, 0.228363},{0.637303, -0.042965, 1.113996, 0.589854, -1.481784, 1.802025, -0.101330},{-0.222538, -0.647623, 0.412530, -1.595271, -0.876539, -0.854312, -1.857469},{-1.209518, 0.175161, 0.250434, -0.054931, 0.225970, 0.772673, -1.358723},{0.238236, 0.176916, 0.423947, 0.705530, -0.595106, 3.341096, 0.607562},{0.089341, 2.461885, -0.421162, 0.795105, -0.564819, -0.996437, 0.014777},{-0.564058, 0.286496, 0.981580, -2.647568, 1.197689, -0.257497, 0.787682},{-0.155199, -1.011660, 0.199453, 0.306568, -0.045005, -0.498773, 0.166719},{-0.990997, 0.968721, -0.039268, -1.381204, -0.283913, -0.137000, -0.727602},{-0.246728, -1.629786, -2.166220, 1.050683, -1.341543, 0.503821, -1.219156},{0.434543, -1.845173, 0.081784, 0.036511, 1.302888, 1.667147, 0.347156},{-1.109291, 2.222059, -0.869485, 0.172491, 0.509041, 0.537489, -0.682572},{-1.450390, -1.243324, -0.484581, 0.897104, -0.688653, -0.046026, -1.203532},{-1.165317, 0.518828, 0.235033, 1.018399, -0.870420, -0.214969, -0.179663},{0.406437, 1.931640, 0.067378, 0.426431, -1.188246, 1.929168, -0.706094},{1.269640, 1.190357, -1.613132, 0.113353, 0.999708, -0.757595, -0.585871},{-0.630865, -0.642443, 0.855455, 0.299976, 0.636186, -0.191205, 2.129640},{1.319297, 0.751836, -0.589098, -1.150419, -1.547793, 1.776266, -1.964097},{-0.168952, 0.311416, -1.647984, -0.636604, -0.832712, -1.220756, -0.409936},{0.080435, -0.128920, -1.396372, 0.206811, 0.086152, -0.241538, -1.150650},{-1.178737, 1.506580, -0.595188, 1.221092, 1.258743, -0.429268, 0.053779},{0.273511, -0.654855, 1.245844, 0.372089, 0.503802, 1.515187, 2.686513},{-0.837770, -0.972778, -0.472626, -1.096144, 0.122636, -0.661068, -0.060979},{0.613989, 1.363698, 1.271019, 1.434508, 1.527466, 1.236814, 1.346778},{0.688148, 0.596300, 0.716952, -1.138538, 0.583015, 0.434493, 0.088743},{0.869049, -0.171431, 1.322450, -1.127598, 0.286664, 0.072036, -1.832257},{-0.995229, -0.378164, -0.589126, 0.020919, -0.718455, -1.098523, -0.947705},{-0.793866, -1.026899, 0.123454, -1.411389, 0.018365, -1.746033, -0.250914},{0.807745, -0.209428, 1.898361, -1.407476, -0.435045, -0.451093, -0.403914},{-1.007544, 0.209705, 0.520470, -1.496128, -0.689501, 1.031110, -0.414603},{-0.974960, 0.057627, -0.586358, 0.154085, -0.818392, -0.358033, -0.663917},{-1.921465, 0.247975, -0.030612, -0.511644, -1.110008, 0.779104, 0.082677},{0.653630, 0.984338, 0.953753, -0.892119, -0.138116, -0.458362, 1.077861},{0.523819, -0.159189, -0.912319, 0.223901, -0.799386, -0.469272, -0.440075},{0.433130, 1.030500, -0.291904, 0.565403, 1.076008, -0.144995, 0.316642},{0.551969, 0.095398, 0.590136, -1.304245, 0.060690, 0.861695, -0.620339},{0.668655, -1.534065, -2.232836, -0.074845, 1.325236, 0.160026, 0.339189},{-0.471722, 0.558136, 0.116826, -1.782344, -0.658920, -0.018468, -1.095601},{0.705162, 0.007282, 0.342036, -0.820923, 0.230655, 2.412880, 0.479242},{1.103470, 0.228382, -0.316766, 0.998303, -1.152505, -0.165573, 1.730790},{-0.849328, 0.420376, -0.065304, 0.737170, 0.162807, 1.981356, 0.624084},{-1.456039, 0.911722, -0.656911, -1.119103, -0.567705, 0.150955, -2.644819},{1.489272, -0.230629, 0.554309, -0.636829, 0.671588, -0.812256, -0.671494},{0.529870, -0.164600, 0.321940, 1.854301, 0.243515, 2.402718, 1.685547},{0.024531, -0.094019, 0.945897, -0.328667, -1.983000, -0.304585, 2.898216},{0.363076, 0.474598, 0.412405, 0.351561, -1.262112, -0.652269, 0.940738},{-0.513880, 0.565651, 0.918500, 1.156608, -0.017333, 0.307517, 1.390159},{-0.673840, 1.177942, -1.346317, -2.595372, 0.840067, -1.111037, -1.239793},{-0.065834, -1.960473, 0.526003, -0.841406, 0.391345, 0.128174, 0.187019},{-0.424637, 2.278970, 1.064558, -0.257655, 2.150918, -1.155613, 0.639068},{0.391281, 1.908397, -0.379898, -0.582445, 1.791121, 0.307411, -1.133824},{0.932954, 0.407398, 0.508221, 0.238021, 0.554341, 0.023734, 0.740272},{0.070482, -1.209164, -1.124078, 2.636688, 0.105271, -0.149718, -0.280931},{-0.361295, 0.659608, -0.767610, -0.517581, -0.022428, -1.066500, 0.476766},{0.608631, 0.499798, -1.199405, -0.491608, -2.801515, -0.091211, 1.403754},{-1.250959, -1.750836, 1.205571, 0.184186, 1.368561, 1.418674, -0.650955},{1.048655, 0.517545, -0.849840, -0.798886, -0.150289, -1.858805, -2.125698},{-1.824630, -2.015988, 1.321562, 0.373175, -0.201845, 1.855356, -2.008699},{-1.161036, -1.567223, -1.217378, 0.812250, -1.499505, -0.585730, -1.044264},{-0.838068, -2.048246, 0.715320, 0.160575, -0.137511, -0.504610, 0.948873},{-0.022155, -1.312384, -0.151580, -0.333923, -0.104531, 1.235983, -0.144797},{-1.196579, 0.695330, -2.032011, -0.273108, -0.579306, 0.177668, 0.066453},{0.284374, -0.009049, -0.281533, 0.185627, 0.296692, -0.995303, 0.587951},{-0.777473, -0.383322, 1.296505, 0.176161, 0.781137, 1.425914, -2.221841},{-1.383292, -0.656848, 0.290081, -0.811488, -0.549140, -0.271199, 0.988992},{-1.763510, -0.639352, -1.137703, -0.147536, -1.584171, -0.377191, -0.441612},{0.069249, -0.996491, -0.281575, -0.798602, 0.776984, -0.189345, -1.585938},{0.388212, -2.175576, 1.136123, 0.119574, 1.747724, 0.066185, -0.937783},{0.053726, 0.026999, -2.329490, -2.334990, -1.148171, -0.831897, 0.914441},{1.174377, 1.259620, -0.739031, 0.173014, 1.719525, -0.366611, 0.020606},{-0.778054, 1.209767, 0.280074, 1.419449, 1.169696, 1.971283, -0.783461},{0.975621, -0.387526, 0.215951, -0.615024, -1.234943, -1.353268, 0.024192},{-1.808344, 0.983390, -1.016661, 1.197010, 0.366891, 0.747368, 0.974919},{-0.331394, 1.213154, 0.682529, 1.167970, -0.099003, -1.300499, -0.347412},{1.295360, 0.646820, -0.488244, 0.296383, 1.582941, -1.145345, 0.923999},{0.143349, 0.750964, 1.096306, -0.258312, 0.999825, -1.101682, -0.607383},{-0.330159, -0.476146, 0.495895, -0.342535, 1.517663, -0.461140, -0.689685},{1.177481, -0.646422, -0.239626, -0.163609, 0.116051, -1.518591, -0.197870},{0.078673, -0.441652, 1.498194, 1.683882, 0.673656, 1.084708, -1.576572},{0.508176, -0.244184, -0.732847, 0.925081, -0.369701, 0.393028, 0.099706},{-0.957409, 0.317056, -0.733569, 0.008481, -1.065488, -0.714415, 0.658037},{0.155243, 0.422142, 1.624698, -0.019442, 1.135317, -1.016559, -0.962783},{1.213531, 1.553459, 1.027735, 0.075564, -0.250710, -1.321818, 1.549021},{1.335805, -2.895008, 0.660921, 0.986778, -0.812686, 1.367250, -1.054978},{0.232249, 0.845724, 1.162484, -1.493697, 0.781999, 0.907512, -1.261096},{0.964240, -0.377411, -0.957386, -0.721901, -0.772649, 0.589671, 0.504299},{-0.331518, -0.040303, -1.084551, 0.369438, 1.737705, -1.167840, 0.739289},{0.666348, -0.881921, 0.358983, 0.732444, -0.726094, -0.006097, -0.800640},{0.616992, -0.377323, 1.152931, 0.631188, -0.080703, -0.000619, 0.248723},{0.221869, 0.702783, 0.569987, 1.845465, 0.153990, 0.813945, 0.532026},{1.601410, 2.168895, -1.393892, 0.073327, -0.010864, -1.875549, 0.993990},{-2.313931, -0.782474, 0.068443, -1.786493, -1.061550, -0.525723, -1.735659},{-1.548984, -1.102361, 2.226553, -1.030410, -2.589804, 0.221204, -0.287466},{0.060265, -0.659460, -1.842139, 1.705906, -1.120823, -1.324937, -0.221826},{-0.097766, -0.451966, -0.267778, -0.472667, -0.191788, -1.463389, 0.078156},{-0.528919, -1.742568, 1.944270, -0.386184, -1.144653, 1.204888, 0.429380},{0.090238, 0.966613, 0.499117, 0.034971, -1.377185, -0.897399, -0.090134},{0.160891, 0.361867, -0.638061, -0.977277, 1.468877, 0.626396, 0.433574},{0.688699, 0.149069, -1.082464, -0.036889, -0.978824, 0.409547, -0.142688},{0.426065, -2.239075, 0.214938, -1.879593, 0.027119, -0.228020, -0.995883},{-0.427248, -0.676331, 1.334590, 1.085060, 0.418341, 1.601153, 2.084364},{0.266908, -0.456621, 0.036080, 0.158585, -0.962852, -0.886030, 0.649038},{-1.103017, -0.875646, 0.846031, -0.272993, -1.262122, -1.286241, 1.347316},{-0.585079, 1.961065, 0.687739, -1.537685, -0.778102, -0.818263, -0.509101},{-0.503218, 2.604685, 0.157236, 0.515957, 0.672013, 0.331837, -0.434274},{-0.144853, 0.543562, 0.324192, 0.192052, -0.654863, -0.413982, 0.942477},{-0.103880, -1.324073, -0.623951, 0.649224, -0.166821, -0.461638, -0.329424},{1.951438, -1.021159, -0.984093, 0.576790, 0.903000, 0.335570, -0.562622},{0.401841, 0.349652, -0.659290, 0.384252, -0.172874, 0.733838, 0.786296},{-1.133395, 1.284090, 1.509229, -0.076673, -0.659604, 0.642959, 0.603055},{-0.208408, 1.018967, -1.451724, 1.438060, -0.367255, 1.414037, 1.660089},{-0.128894, 0.597899, -0.918390, -1.031743, -0.978479, -1.756471, 0.256949},{0.038281, 1.639354, 0.619955, -0.567933, 0.490861, -1.814254, -0.137714},{0.506304, -0.676962, 1.875471, 0.160408, -0.629998, 0.525377, 0.069023},{-0.464703, 0.311081, 0.003672, -0.457998, 1.040022, 1.994087, -0.358966},{-1.079792, -0.268297, -0.339435, 1.858903, 0.213169, 0.262988, -1.290490},{-2.008795, -0.364622, -0.181765, 2.103632, 0.948367, 0.053729, 0.488232},{1.931706, 0.951179, 0.100196, -0.401494, -1.129615, -0.130322, -0.761358},{1.366745, 1.428693, -1.461464, -0.570383, -1.664005, 0.202554, -1.220074},{-2.558177, 1.324187, 0.675157, 0.590083, 1.448332, -0.290499, 0.576900},{0.566225, -0.398343, 1.490554, 1.285596, -2.300022, -1.205364, 1.088725},{-0.363921, 0.012129, -0.856197, -0.917453, 1.000801, -0.619412, 0.011047},{-0.419045, 0.767774, 0.424019, 1.583920, 0.522707, -0.520758, -0.582986},{-0.767052, 0.473881, -1.411568, -1.476499, -0.922108, -0.642501, 0.316414},{0.490842, 1.552863, -0.165739, -1.282637, 1.249964, -0.434346, 0.718369},{-0.940686, -0.289872, 1.211261, 0.319433, -0.320493, 2.374084, 0.100607},{-0.207671, -0.189729, 0.345430, 0.311425, -0.792300, 0.821362, -0.347016},{0.443271, -0.533790, 0.930250, 0.659898, 0.280390, 0.995481, 0.440799},{1.020254, 0.000822, -0.477029, 0.357535, -0.545065, -0.763993, 0.193679},{-1.887605, -2.236458, -0.213170, -1.136760, 2.258200, -0.532609, 0.409079},{1.720381, -0.748792, 0.794794, -0.990428, 0.307420, 1.608057, 0.198807},{-1.018677, 0.566222, 0.225824, -1.845300, -0.673707, -0.513841, -0.084182},{0.764700, -0.063146, 0.730357, 1.998037, -0.963828, -0.513599, 0.006514},{-0.198746, 0.212496, -0.941919, 0.562570, -0.347121, 0.645512, -0.627036},{1.269159, -0.266855, -0.669735, -1.472496, -2.275723, -0.397848, -0.192164},{0.101789, 0.015880, 0.238920, -0.382796, 1.571123, 0.312089, -0.612071},{0.501103, -0.502522, 0.633881, -0.541404, -0.078474, 0.485944, 0.300221},{-1.077948, -1.014745, -1.226020, -1.101781, 0.233482, -0.859945, -0.242800},{-1.381470, 0.410031, -0.247284, 0.193237, 0.047811, -0.938739, 1.102601},{-0.202758, 1.747203, 1.352222, -2.417859, -0.678702, 1.299265, -1.016780},{0.796276, 1.195726, -0.689957, 0.190217, 0.240042, 0.337687, 0.024095},{0.406946, 0.269941, -1.271076, -0.343804, 1.049657, -1.180259, 0.064469},{-0.821901, -0.683705, 2.250961, -0.582472, -0.786091, 1.084955, 0.202031},{-1.179106, -0.371136, -0.203017, 1.585858, -0.338844, 0.878209, -0.095967},{-1.855041, -1.137725, 0.036872, 0.963269, 2.908330, 0.060208, 1.142179},{-1.470717, -2.647758, 0.047860, -0.317450, -0.703045, 0.709297, 0.847094},{-0.306256, 0.501104, -0.221519, 0.196217, -1.897952, 0.747493, 0.293507},{0.383064, -0.138951, -0.120200, 0.102221, -0.524612, 1.183451, 0.372946},{1.125406, 0.305277, -0.655922, -0.459194, 0.108993, 0.422208, 0.406221},{-0.548943, -0.716517, -0.450026, 1.821085, -0.003610, 0.522559, 0.071620},{-0.861934, 0.581077, -1.213219, 0.198703, 0.779304, -1.013930, 1.592736},{-0.540217, -0.210563, -0.496618, 1.460713, -0.989924, 2.171086, 0.136553},{0.572952, 0.247850, 0.362594, -0.968063, 2.085648, 0.369917, -1.744150},{-0.028656, 0.188899, -0.042633, -0.484837, -0.213911, 0.690248, -0.006958},{1.597923, 0.778940, -0.116285, -0.111849, -0.493162, 0.898490, -1.673509},{-0.541092, 0.317722, -0.814749, 0.155163, -0.872258, -2.273821, 0.288084},{0.159828, 0.504452, -2.526255, -1.074409, 0.982216, -0.426847, 0.617521},{-2.368590, 1.792034, -0.828588, 1.558975, -0.159227, -2.060896, -0.320412},{-1.248506, 0.317835, -1.377267, 1.628444, 1.087918, 0.294640, 1.183796},{-1.281652, 0.628733, -0.959189, 0.308459, 0.045063, -0.379480, -0.884272},{1.022398, 1.025020, -0.917816, -0.680136, -2.469826, -0.350456, -0.308494},{0.045110, 0.549076, -0.777928, 0.081234, 1.297754, -0.718672, 0.039550},{0.377584, 0.625691, -0.949935, -0.515948, -0.939958, 0.532052, 0.797551},{0.265356, 2.360389, -0.175470, 0.930370, 0.085377, -1.457643, -0.644480},{-0.457589, 0.259338, 0.308644, -0.223697, -0.375325, -0.386325, 0.086399},{-1.253978, 0.789378, 0.799372, -0.112183, 0.503951, 0.911454, -0.371705},{0.137362, 0.604111, -0.274632, 2.692091, 0.888789, 1.170380, 0.534405},{-0.549504, -0.443825, 0.370479, 0.648948, 0.448960, -0.817494, -0.161684},{-1.054122, 0.780451, -0.568324, 0.487724, -0.312124, 1.326133, -0.432028},{0.445576, -0.342692, 0.091962, 0.727603, -0.271014, -0.080836, 0.155904},{-0.174152, 1.698246, -0.550135, 0.133342, 0.015009, 0.258498, 0.425524},{2.468475, 0.318276, -0.716889, -0.443730, 1.503088, 0.607167, -0.242873},{-1.404217, 0.611115, 0.911413, -2.784576, -2.248960, -0.375252, 1.622402},{-0.100728, 0.011526, 0.405430, -0.505209, 1.406790, -0.257778, -0.829795},{-1.548201, 0.895785, -0.013299, -1.118055, 0.549122, -0.452268, -2.931083},{0.155137, 0.329249, -0.630949, 2.065924, 1.072339, -1.066682, 0.229066},{0.264281, -2.073254, 0.164902, -0.515166, -0.602789, 0.658269, -0.142314},{-2.304233, -0.478355, 0.388550, -0.097402, 0.124098, 0.397768, 0.823693},{-0.440567, 0.076947, 0.270741, -0.449433, -0.498749, 1.131907, 0.731758},{-0.865164, -1.379463, -0.490070, 1.350633, -0.057069, 0.295803, 0.903664},{0.618712, 1.198675, 0.678186, -0.552106, -1.011935, -0.007546, -0.088909},{0.863049, 0.294171, -1.437833, -1.452230, 1.031905, 0.580519, 0.272040},{-0.909487, 1.157666, -0.638967, -1.089717, 1.183980, 1.577345, -0.397360},{-0.955906, -0.088583, 0.544274, -1.204463, 0.048191, 1.643115, 2.149144},{-1.263262, -0.401276, 0.683418, 0.222625, -0.218938, -0.145336, 1.504174},{0.661540, 0.899346, -1.609462, -2.146488, -1.947262, -0.210766, 2.502861},{1.293427, -0.703974, 0.121971, -0.041583, -0.289968, -0.291499, 0.148551},{0.832896, 0.093855, -1.195660, -1.405706, 1.010496, 0.978784, 0.062399},{0.310637, 1.155759, -0.484297, 1.026847, 0.461807, -0.776422, 1.210608},{-0.049258, -0.835426, 2.837620, -0.775322, 0.829184, -0.210409, -1.223948},{-0.474031, -1.440389, 0.237795, 1.294168, 0.009231, -1.287002, 1.202420},{0.463271, -0.539596, 1.508058, -0.115509, 0.060158, 1.262474, 1.459955},{0.469324, 0.454620, -1.476089, 0.068687, 0.178268, 1.219376, 0.808819},{-1.376105, 0.097827, -0.818197, -0.165556, 0.098960, 1.549876, 0.628988},{1.058236, 0.635329, -0.107235, -0.326978, -2.335441, -2.206202, -1.130727},{-0.145746, -0.487994, -0.672095, -0.477841, -0.209927, -0.150380, -0.296597},{-1.312698, 1.013514, 0.202755, -0.243452, -0.280358, -0.298303, 0.120641},{0.207483, 1.609919, 1.078070, 0.601702, -1.030504, -0.333319, -1.677944},{-0.248520, 1.817343, -0.023447, 1.189190, -0.478113, -0.814836, 0.536688},{-0.098269, 1.460683, -0.076574, 1.045277, -0.307863, 1.449608, 0.446630},{0.997031, 0.264111, -1.081971, 1.964572, -0.583510, -0.166933, -0.049276},{1.476234, -0.400339, 0.405387, -0.390008, 0.430393, 0.409819, 0.204791},{-2.814436, 0.498429, 0.887459, 0.067380, -0.967613, 0.266982, -2.037815},{0.197645, -0.489972, -1.522733, 0.562553, -0.297902, -0.276362, -0.268736},{0.975261, -0.764815, 0.608246, 0.456181, 0.653350, 1.714302, -0.165807},{-1.019663, 1.166893, -1.706161, 0.424429, -0.332917, -1.712511, -0.421882},{1.919564, -0.072670, -0.530559, 1.045938, 1.368685, 0.059843, -1.060329},{0.918136, -1.174003, -0.398054, -0.080963, 0.033598, -0.024096, 1.326062},{1.927649, 0.229854, 0.648667, -1.026179, -0.180485, 0.740937, -0.024634},{-1.131938, -0.545770, -0.455364, 0.954044, -0.488131, -0.559444, -1.090974},{0.188432, 0.364719, 0.429139, 1.021053, -2.781907, -0.236627, 0.802281},{-0.971627, 0.597664, 0.620449, -0.004522, 0.593907, -1.001648, -0.613965},{0.351855, 1.183620, -0.029075, -0.149519, -0.502501, 1.269118, 0.794922},{0.180186, -0.329461, 0.830125, -1.204320, 0.112175, 1.082523, 0.354877},{0.624567, -0.122819, 0.111350, -0.661733, -0.392464, 0.571944, -1.034667},{1.709812, 0.497641, -0.082904, -0.154944, -1.092650, 0.263176, 0.529688},{0.719439, 1.216881, 0.963492, 0.160160, 0.279536, -1.020513, -2.179286},{-0.522034, -0.687427, -1.387313, 0.266412, 0.604206, 0.481321, -0.063494},{-0.487371, -1.279637, -0.145152, 1.062106, 0.913151, 0.409483, -0.358483},{-0.214987, 1.895348, -0.417666, -0.768466, 0.013085, -1.210523, -0.807510},{-1.863612, 0.348751, -1.702304, 0.426820, -0.837912, 0.678058, 0.999977},{0.399718, 0.736792, -0.016131, -0.075114, -0.073526, 2.556894, -0.065408},{1.890650, -0.359619, 0.235615, 0.276304, -0.750088, -0.514016, 0.041767},{0.852788, 0.000153, 0.012226, -0.327086, 1.663975, 0.835506, 0.391195},{0.785071, 1.070078, -1.002681, -0.569522, -1.416164, 0.187437, -2.412172},{2.053885, -0.595291, 0.438248, -0.614193, -0.614529, -0.599909, 0.463597},{-0.136333, 0.671714, 0.105807, -0.336981, -0.785185, 1.879065, -0.453540},{-0.452340, -0.134039, -1.125700, -1.244674, 1.478481, 0.368912, 0.535096},{0.542552, 0.556553, -0.094313, -1.031446, -1.131372, 0.246100, 0.775723},{-2.160718, -1.335950, 1.476196, -0.160644, 0.498840, 0.860401, -1.028860},{-0.790030, -0.630224, 0.505364, -0.379315, 0.801084, -0.174966, 0.068489},{1.645247, -0.477397, 2.515927, -0.419787, -0.060528, -0.161849, -1.859040},{-1.494901, 0.834750, -0.921785, -0.917207, -0.806551, 0.633330, -1.126296},{1.204480, 1.990483, -0.353056, 0.053747, 1.012510, 1.574940, -0.104345},{-0.371274, -0.090585, -0.728249, -0.519270, 0.142715, -0.342772, 3.146822},{1.480130, 0.435769, -0.177169, -0.006928, -1.401640, -0.490530, -1.018202},{0.661604, 0.051078, 0.836388, 1.126177, 1.835996, 0.938050, 0.015965},{-0.516325, 0.661262, -0.271399, 0.410158, 0.128326, -1.541329, -0.973640},{-0.358482, 0.311387, -0.747598, 0.078864, 0.032597, -0.791937, 0.371905},{-1.100187, -0.155943, -0.729592, -0.163874, 0.676163, 0.188790, -0.857994},{1.270031, 0.222137, -0.020667, 0.006941, 0.516339, -0.663487, 0.886240},{-2.074919, 0.752712, -0.492060, 0.904779, 0.253684, -0.652598, 0.457970},{0.031338, -0.249257, 1.497661, -0.077647, -1.545442, -0.380221, -1.849368},{0.355951, -1.409205, -2.368854, 2.072623, -0.122661, 1.032634, 1.760581},{0.670066, -1.253574, -0.421128, -0.979291, -0.804939, -0.869644, -0.062369},{0.748095, 0.853539, -1.673234, -0.364229, -0.039786, -0.851446, 1.574236},{-1.547390, -0.048027, 0.120894, 1.889464, -1.171798, 0.770636, -0.401802},{-0.353186, -0.366642, 0.468869, 0.994685, 0.631198, -0.383484, 1.201765},{-0.491280, 1.539733, 0.233434, 1.076664, -1.919779, -0.731316, 0.171582},{0.115099, -0.302965, 0.163792, 1.790480, -0.157862, -0.593447, 1.439132},{0.338635, -1.013790, 0.546498, 0.128967, 0.149517, -1.242087, 0.394211},{0.292292, -1.132211, 0.528674, -0.371258, 0.229505, -0.201778, 0.764397},{1.030646, 0.693020, -1.936326, -0.502709, 0.163877, -0.521795, 1.517562},{0.239205, 0.813885, -0.309862, -0.850242, 1.069130, 0.613708, -0.484740},{0.885988, -1.772845, 0.787652, 0.450655, -2.202718, 0.556812, 0.587906},{0.702830, -1.164513, 0.151033, 0.640510, -1.615753, -0.293892, -0.841031},{1.096342, 0.232653, -0.337739, -0.755975, 1.681527, -0.199852, 0.218784},{-1.388444, 2.321997, -2.040025, 0.155589, 0.912442, 0.756155, -0.671518},{-0.713826, -1.035185, -0.871040, 1.607946, 0.720177, 0.505040, 1.534111},{0.377558, -0.483770, -1.718216, -0.264397, -1.020381, -0.301005, -0.215613},{-0.136254, 0.024186, 1.504752, 0.371279, 0.884827, 1.476416, 0.493452},{-0.152330, -0.720727, -0.640898, 0.022295, -0.239261, -0.242175, -1.055437},{0.635958, -1.521684, -1.729573, -1.730051, -0.759899, 2.435567, -1.235021},{-0.595282, -1.168665, -1.860782, 0.899219, 2.142806, 1.470512, -0.356095},{-0.554979, 0.156656, -1.290511, -0.168249, -1.687016, -0.472202, 1.216604},{0.421609, -0.390090, -0.054723, -1.094338, -0.713606, -0.348135, -0.754037},{1.303529, 1.664906, 2.246689, 0.195910, -1.877296, -0.287216, -0.617786},{-1.317136, 0.414500, -1.188527, 1.062094, -1.200902, -0.578708, 0.847266},{2.148349, -0.211521, 0.742800, -0.070398, 1.976019, 0.070245, 0.609405},{-0.269646, 1.040683, 0.675166, -1.759514, 1.055452, -0.545371, 1.857998}};		
    }
    else if(pdg.compare("kplus") == 0){	 
      return { {1.367393, 1.292851, 0.121579, -1.411180, -1.095783, 0.144910, 1.155862}, {0.534574, -0.470033, 1.472608, 0.357328, 0.452747, 0.427289, -0.663398}, {-0.058443, -1.010801, -0.284830, 1.060527, 0.954475, -0.679052, -0.983169}, {-0.272804, 0.350220, 0.915789, -1.790793, -1.090060, 0.878522, 0.046376}, {-0.818568, -0.637788, 0.688878, 1.122198, -0.055504, 1.174830, -0.841907}, {-2.112710, 0.465744, -0.100384, 1.793911, 0.466980, 0.133821, 0.389187}, {-1.254074, 0.918904, 0.441761, 0.603208, 1.331181, -0.493696, -0.366663}, {-0.959399, -0.138858, 1.347242, -0.353400, 0.346127, -0.390838, -1.287324}, {1.076622, -0.264614, -0.374221, 1.253816, 0.419300, 0.809251, -0.462083}, {-0.951600, -0.348808, 0.481222, -1.140316, 0.416450, 2.188865, -0.906726}, {1.387130, 0.660727, 1.435455, -1.048722, 0.377444, 0.175381, -0.931098}, {0.822697, -0.674956, -0.606649, -0.843156, 0.095161, -0.338133, 0.649241}, {0.976130, -0.155503, -0.072318, -0.690634, 1.859409, 0.478516, 0.141487}, {0.403007, -0.036501, -0.634636, 0.365653, 1.571958, 0.326153, 0.011390}, {-2.239528, -1.665398, -0.527825, 0.199522, 1.794437, 1.135630, 0.076237}, {1.591812, -1.186962, 0.464429, -0.757568, 0.373626, 0.220932, 0.128204}, {-0.859189, -1.821905, -0.356017, -1.668283, -0.127355, -0.568570, 0.044279}, {-1.713520, -0.131788, -0.352018, 0.123967, 0.207797, 1.036016, -0.320258}, {0.390201, 0.058855, -0.762762, 0.904096, -2.071044, -2.653167, -1.922750}, {0.838542, 0.015263, 0.077651, 0.123674, -0.225388, -1.286382, 0.280462}, {-0.479608, -0.201553, -0.722731, 0.787112, 1.153146, -0.484563, 0.253358}, {-1.655522, 1.250833, -1.918763, -0.261135, 1.218867, -0.261632, 0.331215}, {-1.324356, 0.047344, -1.784084, -1.356787, 0.974882, -0.020912, 1.305153}, {0.451521, 0.739043, 0.186114, -0.053171, -0.678442, 0.952686, 0.041732}, {-0.673824, -0.184076, 0.860082, 0.621944, -0.795496, 0.505550, 0.886752}, {0.614312, 0.013892, -0.393245, 0.315558, 0.990688, 0.476907, 1.003443}, {-0.353014, -1.478345, 0.002287, -2.380836, -0.605650, 1.146255, -0.163743}, {0.570073, 1.728823, 0.120915, -0.078883, -0.191315, 0.309434, -0.362392}, {-0.437243, 0.275342, 0.539946, 0.341992, -0.697411, -0.567824, -1.365279}, {1.247100, -0.163064, 1.175782, -0.107684, 0.796579, -0.651988, -0.174295}, {1.399662, -1.129036, 0.334001, -1.616810, 0.474019, 0.234731, -0.630714}, {-0.876766, -0.192250, 0.007017, 0.166262, -0.181060, -0.788707, -0.554547}, {1.253438, -0.476422, -1.250669, -1.372617, 1.246160, 0.667025, -1.918340}, {-0.241910, -0.455894, -0.732315, -1.104191, -1.015629, -0.404273, -0.375398}, {-1.279045, 1.099559, 1.367319, 1.306720, -0.436777, -0.316746, -0.770576}, {-0.652935, 0.999871, -2.237195, -1.077424, 0.459849, -1.086575, -0.254053}, {-0.252841, -1.294561, -0.333224, -2.005728, -0.447528, -0.001064, -0.891914}, {-0.319927, -1.134769, 0.415169, -0.614512, -1.393372, 0.169070, 0.270368}, {-0.294245, -0.618711, 0.015857, -0.310293, -0.713962, 0.088625, -0.842542}, {2.527596, 0.039492, 2.669928, 0.366544, 0.279361, 0.762189, 1.828799}, {0.055032, -0.058872, -0.142719, 0.855246, 0.977136, 1.420616, 0.725833}, {-0.404217, -1.332883, 1.227831, 0.485221, 0.085397, -0.592295, 0.986754}, {-0.325498, -0.771263, 1.982212, -0.417529, 0.028417, -0.867805, -1.935174}, {-0.263318, -0.215111, 0.758501, -0.881838, 0.957105, 0.550695, -1.010453}, {1.852789, -1.478824, -0.455865, 1.357908, -0.005999, 0.309471, -0.330744}, {0.102872, 0.098354, 0.550731, 1.250726, 0.384669, -0.348222, 0.246726}, {-2.269240, 1.204862, -0.138336, 0.471483, -1.501316, -1.342081, 0.938320}, {1.109809, 0.296675, 0.393175, -2.026159, 0.244040, -0.080217, -1.574087}, {0.236029, 0.470966, 1.234077, -0.139533, 0.230512, 0.639314, 0.875788}, {-0.910047, 2.018447, 0.883528, -0.185594, 0.212502, 1.619980, -0.418186}, {0.798614, 1.924597, 0.692334, -0.140167, -1.100641, -0.152136, -1.070915}, {1.884273, -0.304710, -0.307352, -2.205734, 0.170142, 0.939021, 0.176899}, {0.675092, -0.905326, -0.238829, -0.105875, -0.287144, -1.345028, 1.441503}, {-1.432101, -0.887197, 0.260065, -0.079961, -0.397321, 0.718666, 0.438271}, {-0.469880, -1.561895, 0.067254, -0.316965, 0.479446, -1.364427, 0.419595}, {-1.039804, 0.485967, -0.894604, 0.730998, -1.399007, 0.246946, 0.595171}, {0.587768, -1.892871, -1.448513, 0.430821, 1.040498, 0.786355, 0.095247}, {-0.714729, -1.273467, -1.151599, 0.050034, -0.876063, 0.054160, 0.918962}, {0.173738, -0.447429, -1.314139, -1.830386, 0.513666, 1.084649, -2.057591}, {1.152085, -1.175941, -0.416018, 0.218836, 0.985711, 1.090137, 0.176001}, {2.412506, 0.062043, -0.508511, -2.057096, -0.001434, 0.303564, -0.522389}, {1.491138, -0.397101, -0.327974, 0.446851, -0.909835, 0.276997, 1.230421}, {0.219986, 0.382486, 0.863247, 0.178662, -1.130783, -0.917506, -0.490130}, {0.803039, -0.136966, -1.545492, 2.099304, 1.584225, -2.560939, -2.347047}, {-2.024322, 0.724917, 1.518803, 0.445846, -0.478013, 1.092029, -0.727423}, {-2.024081, -0.017266, 0.651234, 1.538047, 0.934897, 0.691084, 1.400253}, {0.385704, 0.185866, 0.936021, -0.653195, 0.674681, 0.371008, -0.323528}, {-0.121311, 1.164435, 0.932621, 1.277557, -1.492501, 0.112247, -0.213264}, {-0.797892, 0.117525, -0.653383, 0.290826, 0.187742, 0.858791, -1.304159}, {0.037741, 0.553286, 1.147591, 1.058379, 0.873605, -0.105446, 0.755101}, {-0.120708, 0.778904, 0.169742, 1.894190, 0.235840, -0.853419, -2.030334}, {0.507852, -1.729968, -0.721699, -1.086989, 0.238615, -0.682356, -1.900744}, {1.119774, -0.561489, 2.818501, -1.006487, 0.181647, 1.437963, -0.307794}, {1.568797, 0.198488, -0.463704, -1.147519, -0.065160, 1.157719, 0.317095}, {-0.774913, 0.349862, -0.762503, -0.973816, -0.230566, 1.306216, 1.862191}, {0.587131, 0.942492, -0.784700, -0.998814, 1.222798, 1.227133, -1.468333}, {-0.793610, -1.640108, 1.286355, 0.164381, 0.513376, -0.125712, -1.939704}, {2.129230, -0.619134, 0.526665, -1.026588, 0.070223, 1.681405, -0.453377}, {2.484923, 0.047514, 0.376944, -0.476603, 0.584461, 0.390282, 0.617555}, {-0.528916, 1.285034, -0.496738, -0.993776, -0.125106, 1.492855, 0.749256}, {-0.478815, -0.330134, -2.204542, -0.023525, -0.539820, 0.716771, 1.050324}, {0.083653, 0.896944, -1.463854, -1.268915, -1.987592, -0.396931, -0.473874}, {-1.163711, 0.173826, -0.407877, -0.739651, -0.241694, -2.455483, 1.996500}, {-1.090572, 1.567751, 0.338545, -0.329781, -1.800551, -0.205264, 0.406438}, {-0.043778, -0.559204, 1.440043, -1.180716, -1.432416, 0.333950, 0.305501}, {1.076003, -1.357743, -0.018667, -0.683852, -0.165188, 0.138934, 1.207832}, {-1.044973, 0.776156, -0.247934, -1.913294, -1.766180, 1.579774, 1.901647}, {0.054190, 2.427184, -0.043712, -1.569068, 0.764446, -0.535440, 0.934530}, {-0.047252, 0.328289, 0.793936, -0.414130, 0.664384, 0.206883, 0.211062}, {-0.007910, 1.526076, -0.366237, -2.249043, -1.317190, 1.993910, -1.854556}, {0.437817, 1.218331, 0.022893, -1.105741, 1.485830, -0.667264, -0.288585}, {0.578485, -0.251698, -1.406556, -0.307950, -1.081642, -1.292116, 0.797001}, {2.192826, -0.575589, 0.990992, 0.246482, 0.565263, -0.654641, -1.651744}, {2.930098, -0.825516, -1.046732, -0.269076, -0.366620, 0.016448, -1.618483}, {-1.293155, -0.451676, 1.272419, -1.745280, 1.309839, -0.741751, -1.792768}, {1.365330, -0.326242, -1.071577, 0.548796, 1.645021, -0.817249, 0.943298}, {-0.007465, -0.646315, 0.042122, 1.964342, 0.299860, 0.258665, 1.241399}, {-1.036765, -0.556057, 1.282265, 0.428306, -0.417417, 1.766718, -0.474185}, {-1.202560, 1.252272, -1.002922, -1.117897, -1.093327, 0.940489, 0.127428}, {1.243176, 0.909325, 1.331781, 0.962374, 0.323824, 1.235567, 0.358653}, {0.171306, 0.060433, 0.132325, 1.181687, 1.239200, -0.104157, -2.805955}, {-3.264112, -0.588143, -0.926106, -1.304915, 0.064297, 0.555950, 1.248894}, {1.108937, 1.427480, 0.503634, 0.125554, 1.054170, -1.227389, -0.380324}, {0.516670, 0.058359, -0.102649, -0.007599, 0.375774, -0.608257, -0.122211}, {0.651549, -0.323107, 1.462240, -0.952423, 0.397533, 1.075962, 0.480772}, {1.193869, 0.484202, 1.038267, -0.164330, 0.437212, 0.552845, -0.207019}, {-0.130309, 1.314106, -0.418370, -2.031417, -0.331402, 0.228869, -0.955918}, {-0.894349, 0.547245, 1.814219, 0.573127, 1.119512, 0.462631, -0.701033}, {-0.400268, -1.620739, -0.448540, -0.598693, -0.115677, -1.387428, 0.784156}, {-0.450371, 0.164485, 0.010101, -1.090221, 1.100077, -0.147733, 0.213443}, {0.211511, 0.133075, -0.849710, -0.056850, 1.419024, 0.830705, -0.288266}, {1.356759, -0.300316, -0.414269, -0.567364, 0.346235, 1.143420, -0.142693}, {0.971221, -1.575072, -0.058331, 0.313456, -0.320494, -0.073083, -0.704762}, {-0.303330, 0.694161, -1.569331, 1.822267, 0.447687, -1.718016, 0.306770}, {-0.504481, -0.376800, -0.944359, -4.595510, -0.004286, -0.500437, 1.275733}, {-1.733063, -1.943182, -0.335297, -0.423002, -0.555526, -1.002707, 1.500104}, {0.723323, 0.511216, -0.665850, -1.036413, 0.840418, -1.166265, -0.490069}, {1.130326, 1.479540, 0.178067, -1.073751, -1.768743, -0.468686, 1.383825}, {0.642909, -0.670751, 0.838026, -0.695331, -0.905427, 1.073269, -0.685286}, {0.461491, -1.243661, -0.698200, 0.133682, -1.453618, 1.485448, -0.728688}, {-1.238411, -1.274253, 1.074299, -0.447864, -0.675880, 0.355634, 0.093378}, {-1.306094, 0.044638, 1.692049, -0.209219, -1.482755, 0.633686, 0.355279}, {-0.551669, 1.968508, 1.320785, -0.609009, -0.140498, -0.411790, 0.000432}, {0.247394, 0.811598, 1.238368, -0.034039, 2.268867, -1.053564, -1.258794}, {-2.378108, 1.176066, -0.164216, -0.529159, -1.746556, 0.039905, -0.512412}, {-1.291623, -0.036329, -1.154858, -0.487026, 0.236572, -1.145062, -1.305656}, {-0.923975, 0.724277, 2.240334, -0.727561, -0.033077, 0.788743, 0.564252}, {-0.929007, 2.843882, 1.414580, 0.730268, 0.166862, -0.443642, -2.207899}, {-0.352043, 0.700081, -0.658357, -0.747351, -0.219043, -1.194075, 0.486510}, {1.281293, -0.653667, 1.144736, -1.878476, 1.264663, 0.497763, 0.172615}, {1.496026, 0.266491, 0.735618, -1.090113, -0.707901, 0.295544, -1.139583}, {0.277257, -1.658943, 0.455018, -0.668627, -1.209923, -0.355229, -1.224576}, {0.080042, 0.795693, 1.089746, -0.153890, 0.095769, 0.279119, -0.449085}, {0.941645, 1.023518, -0.139951, 0.747215, 1.201548, -0.856409, -0.552480}, {-0.656300, -0.383828, 0.817642, 1.748813, -0.020508, 1.005078, -0.395574}, {-0.924348, 0.366664, -0.108521, 1.937674, -1.152948, 0.530390, -0.470500}, {0.703354, -0.292489, 2.146928, -0.384284, 0.408556, 0.306452, 1.035938}, {-0.811361, -1.149397, 0.226082, -1.768511, -0.838028, 2.057089, 0.924221}, {1.682636, -0.078308, -1.114185, -0.131622, 0.017787, 1.082737, 0.277068}, {-0.471368, 0.783425, 0.621261, -0.398723, -0.805335, -0.036666, 1.285895}, {0.686133, 0.104625, 0.875176, 1.744015, 1.465335, 1.172941, -0.755119}, {0.656361, 0.997267, 0.070914, 1.496169, 0.524824, -0.021945, -1.553892}, {0.372251, -0.190503, -1.000878, -0.595715, -0.187036, -0.421981, 2.532178}, {0.559815, -0.449680, -2.101419, -0.624993, 1.433439, -0.670799, -0.828142}, {1.605747, -1.173496, -0.407000, -0.952267, 1.828775, -1.796203, 0.009007}, {-1.638720, -1.603119, -1.469830, 0.219467, 0.175744, -0.161647, -0.310799}, {0.186704, 0.618750, 1.695020, 0.079273, -0.037747, -0.688858, 0.274353}, {0.518032, -0.502755, 0.543360, 1.264776, 0.993161, -0.581863, -0.322524}, {-1.154364, 0.819427, 0.320268, 0.894066, -2.089308, -0.034651, -0.535512}, {0.573963, 1.718017, 0.378490, 0.489934, 0.921193, -0.453677, 0.093982}, {0.518235, 0.008725, -0.845939, -0.503828, -0.776669, 0.305255, -2.298262}, {0.286281, -0.061566, -0.248411, 0.378232, -1.601204, 0.161936, 1.704773}, {2.151842, -0.932847, -1.089765, -0.189848, -0.099307, 1.070347, -1.460052}, {1.277166, -1.408773, 1.239346, -1.097942, -0.506368, -0.182177, -0.066975}, {1.745260, -0.318140, -0.829921, 0.138967, -0.742236, -1.693001, -0.198429}, {0.817855, -0.808145, -0.643924, -0.222708, 0.538242, 1.653399, -0.755387}, {-0.252090, 1.251416, -0.498890, 0.583028, 0.631191, 2.244109, 1.351563}, {-1.272376, -0.501670, 0.821735, -0.234663, 0.013334, -0.673290, 0.282817}, {-0.263016, 0.068192, -0.410136, -0.597362, 1.004409, 0.780871, -1.296947}, {-0.461490, 0.780456, 1.712845, -0.171149, -0.637891, 0.773435, 2.126675}, {0.373580, -1.148450, 2.524954, -0.127923, 0.389533, 0.318897, 1.347223}, {-1.118665, 1.762065, -0.544924, -1.203573, 0.218797, -0.280895, -1.412702}, {-0.848276, 0.868560, 1.424999, 0.158772, -0.663257, -2.252339, -1.049719}, {-1.015560, -1.460783, 0.672623, -1.245867, 0.752254, -0.400706, 1.169296}, {-1.593419, -1.861650, -1.333832, 0.474836, -0.639325, -1.903870, 1.103539}, {-0.766455, 0.019457, 0.390962, 0.926757, 0.119451, -0.696442, 0.374439}, {-0.798476, -1.486756, 1.685077, 0.550925, -0.662016, -0.983609, 0.495739}, {-0.875754, -0.444696, 1.394639, -0.515775, 0.734618, 0.249356, -2.012718}, {0.690326, -0.768754, -0.040694, 0.659373, 0.156533, 1.528098, 0.342755}, {-0.677168, -0.447639, -1.620915, -1.905635, -2.700207, -0.698836, 0.768704}, {-0.051973, 0.186490, 0.945299, 0.281322, 0.533352, -0.675016, 0.306311}, {-0.681908, 0.167762, 0.047648, 0.162652, -0.040322, -0.510165, -0.639110}, {0.516809, 0.217484, -0.608501, 2.398010, 0.023374, -1.320762, -0.953877}, {-0.819292, -0.424132, 0.166868, 1.117352, -0.895062, 0.634209, 0.380213}, {0.008282, 0.954534, 0.841213, -0.259666, -0.187647, -0.399841, -0.319905}, {-0.292092, 1.080593, -0.958030, -0.756436, -1.334215, 0.647837, -0.611464}, {-0.801170, -1.210086, 1.601381, 1.696889, -0.429616, 1.951503, 0.308516}, {-0.604057, 0.922670, 0.364824, -0.603316, -0.987207, 0.161136, -0.189130}, {-3.115464, -1.071247, -1.176629, -0.885288, 0.943854, -0.079437, -0.897786}, {0.506357, -0.427465, 0.722417, 1.000937, -1.083111, 0.034819, 0.487900}, {0.530159, 1.074675, 0.662865, -1.142714, 1.645158, 0.025057, 1.104694}, {-0.221483, 1.183938, -0.256395, 1.080752, 1.628406, 0.262594, -0.907124}, {-0.096997, 0.184579, 0.671771, -1.936531, 0.413926, 0.102089, 1.669060}, {0.510408, -0.717742, 1.017587, -0.154325, 0.229847, 0.220618, -1.642520}, {-0.642584, 1.199838, 0.319366, 1.399155, 0.965100, 0.475482, -0.643657}, {-0.565678, -0.736135, -1.373354, 0.430104, 0.291612, 1.228546, 0.636742}, {1.605889, 1.922844, 0.112115, 1.658416, -0.861697, -1.502311, 0.931069}, {0.418723, -0.784212, -0.616234, -0.015679, 0.389441, 0.220890, 1.032574}, {0.991333, -0.587797, 0.080822, -1.875164, 0.311852, -0.852783, -0.908787}, {0.026222, -0.268749, 0.187778, 0.036714, 0.645356, -1.739338, -0.453860}, {1.357886, 0.747166, 0.411599, -0.899896, 1.558702, 0.479299, -0.820105}, {0.160341, -0.255689, 0.802928, -0.720936, 2.107416, -1.469754, -1.990128}, {-0.369195, 0.819973, -0.116987, -0.875811, 0.362258, -1.109409, 0.365193}, {1.218052, -0.441100, -0.286708, -0.681395, 0.371090, -1.495109, -0.552382}, {0.480229, 2.012447, 0.666974, 0.805913, 0.654723, -0.697137, -0.253412}, {-0.085379, -0.261567, 0.143828, -0.550467, -0.208909, 0.066020, -0.145170}, {-1.959076, -0.344976, -0.470223, 1.665392, 0.897741, 0.328515, -1.080489}, {0.546895, -0.007768, 0.348702, -0.416320, -0.157299, 1.767215, 0.660851}, {0.103611, -2.084327, -2.811583, 1.095475, -0.526916, -0.547668, -2.033243}, {-0.443968, 0.845751, -0.304341, 0.284698, 0.788852, -0.017907, 1.356245}, {-2.210817, -0.718888, -0.273852, 0.865456, -0.295171, -0.420573, -0.606293}, {0.333330, 0.898019, -1.291610, 0.389963, 0.138830, 1.071227, 0.697241}, {0.264228, 1.731671, 0.101122, 0.358413, 0.097155, 0.708639, -1.050797}, {0.108664, 0.626448, 2.047793, -0.035047, 0.375322, -0.344844, 0.500666}, {0.671016, 0.763748, -0.221545, 0.827504, 0.356383, 0.339128, -1.499322}, {-0.157362, -0.820891, -0.053302, 0.583154, -0.112519, -0.765332, -1.159287}, {1.749116, -1.951742, -0.724983, -2.701802, 0.130040, 0.091200, 0.878271}, {-0.988486, 0.902790, -0.920790, 0.620702, 0.308910, -0.370812, 1.493925}, {0.900060, -0.305593, 0.427565, -1.295820, 0.981730, -0.529203, -0.334994}, {-0.570472, 2.344493, -0.496063, 0.153940, -0.841490, -0.318728, -0.280252}, {-0.656479, -0.204981, 0.627117, 0.615112, -0.455393, -0.649320, -0.181668}, {1.303727, -0.356414, 0.566492, -1.088467, -1.269112, 0.613461, 0.465760}, {0.065777, -1.237912, -0.512823, 0.513747, -0.012137, 0.879229, 0.904166}, {-0.086172, 0.648828, -1.705379, -1.954128, 0.259875, -1.714363, 0.065534}, {0.712702, 0.252362, -1.444360, 1.402375, -0.083744, 0.287298, 1.262052}, {-0.145206, -0.995733, 0.535828, 0.749801, -1.122308, 0.090837, 1.101550}, {-0.279690, -0.471550, 1.409229, 1.179895, -0.399548, -1.679697, -0.014644}, {0.222359, -0.393775, -1.691170, 0.273100, 0.161846, -1.084473, -0.775651}, {0.762269, -1.233801, -0.347637, 0.593526, 0.824514, -0.821985, -0.183763}, {0.371891, 1.172071, -0.662261, -2.778579, 0.165519, -0.054903, 0.897983}, {-0.875868, 0.683910, -0.072956, 0.870428, 0.170199, -1.003876, -0.224833}, {1.397570, -1.159106, -1.927818, -0.806817, 0.032734, 0.155283, 0.245600}, {0.142736, 0.129211, -0.063779, 0.232723, 0.807273, 0.990292, -0.615939}, {-0.559299, 0.361469, 0.306688, -0.656385, -0.421816, -2.630677, 0.484330}, {-1.290123, -1.360252, 2.120205, 1.803964, -0.134177, 0.022221, -0.026959}, {0.161350, -1.243824, -1.114194, -0.233238, 0.617386, 0.276829, 1.811533}, {0.327081, 0.184619, -1.960805, 0.556340, -0.913914, -1.498119, 0.996423}, {-0.009022, 0.887457, -0.942244, -0.230040, 0.351004, 0.440716, -1.226256}, {0.976327, 0.150659, -0.221481, -0.155682, 1.048232, 0.015874, 0.527528}, {1.570075, 2.040219, 0.192533, -0.088838, -0.069708, -1.464450, 0.700815}, {-0.037904, 1.836559, 0.530495, 1.059529, 0.775036, -0.278261, 0.715370}, {0.631232, -0.364215, -1.299913, 1.184539, -0.894087, -0.067944, 0.835835}, {2.273437, 0.054534, 0.126472, -0.268782, -0.000299, 0.600976, -0.915070}, {-0.826026, -0.970442, -1.179925, -1.555971, 1.591109, 0.899381, -0.827173}, {0.112360, 0.065733, -1.617079, -0.441451, -0.041766, -0.335772, 1.545169}, {0.282583, -0.447642, 0.961917, 0.281734, 0.637853, -0.580209, 1.032078}, {-0.355677, 0.844725, -0.180979, -0.632134, -0.330045, -0.274461, -0.870968}, {1.383492, -1.028367, -0.740422, -0.882839, -1.220251, -0.568632, 0.624469}, {0.066301, -0.794800, -0.434406, -1.875648, 1.920318, -0.128366, -0.872067}, {0.495858, -1.247710, 0.766434, 0.102960, 0.080520, -0.318598, -0.346555}, {-1.566749, 0.373166, -0.687073, 0.017243, -2.068647, -1.519434, 1.885868}, {0.786502, 0.594050, 0.536875, 0.243136, 0.174392, 1.222291, -1.412980}, {-1.176589, -0.324809, -0.887729, 1.236608, -1.344350, 0.471987, 1.117147}, {-1.163936, -0.167203, 0.153672, 1.927978, -0.618017, 0.071907, 0.198031}, {0.046700, -0.947722, -0.845086, -0.380959, 1.915891, -0.853172, 0.384208}, {-0.306711, 0.915339, 0.381831, 0.159113, 0.122296, 0.021639, -1.769924}, {-1.703274, 0.246384, 0.301491, 0.493637, 0.300374, 0.459367, -1.504239}, {1.261713, 0.554462, 0.141613, 0.283045, -0.482012, -0.725806, -0.778279}, {-1.933474, -0.010310, -0.315612, 0.770951, -0.117141, -0.050130, -0.768054}, {-0.827104, 1.895813, 0.108923, 1.195037, -0.783149, -0.801726, -2.465575}, {0.296404, -0.625681, 0.600058, -0.120152, 0.396997, -0.472046, 0.215233}, {0.374433, -1.387535, 0.003504, -0.266504, 1.101441, -1.863083, -1.284127}, {0.439282, -0.000901, -0.844571, 1.061276, -0.173481, -0.187106, 0.442115}, {-1.049496, 0.161821, 1.586552, 0.816670, 0.132006, -0.760746, -1.987381}, {-0.128261, -0.487173, -1.908592, 2.082450, 0.585252, 1.339580, 0.180135}, {-1.018892, -1.031852, -0.804895, -0.399576, -1.854555, -1.437255, 1.045458}, {-0.862291, 0.639253, -1.558242, -0.730695, 1.558407, -0.009844, -0.492857}, {-0.513745, 0.251673, -0.036981, 1.896936, 1.361213, 0.010467, 0.823479}, {0.708414, -0.043409, -1.224397, -0.710891, -1.798674, -1.405463, -0.709619}, {0.232046, 0.580914, -0.827950, -0.014628, -0.995276, 0.067497, 0.394587}, {-0.345376, -0.731591, 1.130066, -0.205242, -0.391949, 0.750235, 1.554349}, {-0.539052, -0.558128, -0.289524, 0.359126, 0.713382, 0.985069, -0.941861}, {-0.594814, 0.979788, 1.863339, -0.981118, -0.309228, -0.056645, -1.411746}, {0.386474, -0.776481, -0.865477, 1.374274, 0.385023, -0.451608, 0.187306}, {-0.653230, -0.474595, 1.165723, 0.775566, -0.383052, -0.189398, 0.249502}, {0.738849, -1.287462, -0.562264, -1.238016, 1.885463, 1.182350, -0.900443}, {-0.108863, -0.067471, -1.150551, -0.406007, -0.558179, -0.959419, -0.158370}, {-1.972075, -0.672523, -0.438952, 0.808463, 1.260888, 0.211884, -0.635543}, {-0.550969, 0.532797, 0.382134, 0.683277, 0.900542, -0.234432, 0.081385}, {2.626596, 0.488253, 0.215853, -0.969500, 0.062074, 0.111133, 2.427027}, {1.056339, -0.807065, 0.000287, -1.942896, -0.718033, 1.712274, 1.230062}, {-1.678372, 0.118514, 1.245715, -0.117588, -0.002319, 2.596751, -0.704975}, {-0.979720, -0.432766, 0.420344, 0.010232, 0.488652, 0.897534, -0.640623}, {-1.553215, -1.763653, 0.680734, -0.393309, -0.119990, 0.151639, -1.857937}, {1.219699, 0.802162, 1.386149, 0.928920, 0.349350, 0.141040, 0.544299}, {0.617079, 1.989012, 0.385183, 0.582389, 0.589716, 0.105250, -0.319493}, {0.932496, -1.365287, 1.046578, 0.446944, -0.665848, 0.059561, -0.157678}, {1.038000, -0.411904, 0.297047, -1.263209, -1.222321, 1.166517, -0.983257}, {0.687177, -0.451556, 1.224129, -0.787336, 0.000820, -0.997504, -1.577491}, {0.290803, -0.555711, 0.403925, 0.635180, 1.475238, -0.574162, 0.618939}, {-0.214332, -0.568642, 1.676100, -0.087919, -0.830945, -0.745352, -0.633039}, {0.105168, 0.876507, -2.320168, -1.353495, 0.148990, -0.757444, 1.589108}, {-0.841402, 0.845016, 0.037418, -0.154461, -0.658870, 1.158291, 0.088971}, {-0.136712, -0.073427, 0.390756, -0.086323, 1.525113, -1.385603, 1.328146}, {0.832572, -0.774330, -0.381071, 0.334351, -0.555444, -1.493649, -0.766855}, {-0.264253, 1.624736, 0.209293, 0.715937, 0.625671, -0.726972, 0.265662}, {0.613317, 0.482082, -0.755253, 0.372313, -1.973633, -0.187457, -0.346930}, {-0.690068, 0.803888, 0.380257, -0.497035, -0.713006, 0.663746, 0.426944}, {1.379452, 2.362953, 0.122776, 0.214811, 0.171834, -1.283537, 0.537435}, {0.651746, -1.291162, 1.044319, 0.356418, -0.417875, 0.365258, -0.566336}, {-1.271616, 0.185476, -0.024559, -0.822330, 0.024370, 0.869304, 0.630038}, {0.517743, 0.587039, 0.340342, 0.028031, 0.968845, 0.888909, -0.026059}, {-0.616652, -0.328778, 0.409825, -0.859338, 0.351918, 0.611454, -1.320276}, {-0.825034, 0.342327, -0.053452, 1.586221, -0.869251, -2.263201, 1.436436}, {0.367839, -0.305996, -0.975122, 1.012007, -0.682963, -0.864928, 0.642931}, {-0.917521, 1.312233, 1.297228, -1.444908, -0.826066, 1.448843, 0.359439}, {-2.410693, -0.208134, -0.428586, 0.267069, -1.250155, -0.306793, -0.099958}, {-2.327913, 1.268839, 0.266776, 1.497480, -0.752299, -0.499548, -2.137461}, {-2.895071, -0.069257, -0.354296, -0.831330, -0.048385, -1.768508, -0.721324}, {-0.205087, -1.335720, -0.442353, -1.975409, 0.058027, 1.653962, 0.390633}, {0.526143, -0.533313, 1.482338, -1.606891, -1.419406, -0.570277, 0.431092}, {0.325393, 0.024580, 1.404170, -1.151899, -1.864493, 0.471147, -0.619332}, {0.799607, -0.529847, -0.618422, 0.876343, -0.177257, 0.747566, 0.170995}, {-0.755331, -0.205363, -0.945204, -0.428741, -0.842708, -0.757052, -1.575390}, {-0.885410, 0.804462, 0.101170, -0.015645, -0.252264, 1.480743, -0.728309}, {-0.158615, -0.104166, -0.209189, -0.422704, -1.833113, -0.241686, -2.128243}, {0.929851, 0.841463, 0.849667, -0.486051, 0.611360, 1.299512, -0.417349}, {1.693097, -0.135949, -0.984686, 0.516242, 0.283493, 1.774139, -0.607124}, {0.382773, -1.092116, -0.277571, 0.201827, 0.826478, -0.419664, 0.194628}, {0.378782, 0.819754, -0.979230, -1.587693, -0.147390, -1.549947, 0.020019}, {0.470986, 1.123615, -0.252995, 0.360123, -0.154954, -0.972308, -0.860390}, {-1.988485, 0.814867, 0.259291, 0.256940, -0.040026, 0.738134, -1.400198}, {1.390552, -0.396194, 0.398083, 0.643453, 0.201328, -0.113048, 1.130920}, {1.017942, -0.980723, -0.157912, 1.567849, 0.633224, 0.168062, -1.245850}, {0.310399, -0.432387, -1.149374, 0.450330, -0.894800, 0.340095, 0.460813}, {0.437202, 0.360703, -0.070089, 0.044121, 1.024531, -1.634638, 0.647008}, {0.574172, -1.226945, -0.973083, 0.170096, -0.221037, -1.369613, 0.382077}, {2.360887, 0.086447, 1.281891, -1.378718, -0.114549, 0.203390, 0.831419}, {-0.260050, -0.853225, -0.409273, -0.056988, 0.150596, 0.126622, -1.768691}, {0.876645, -0.658851, -0.956693, 0.224481, 0.842076, -0.459814, -2.628174}, {0.422417, -0.155943, 0.517883, 0.982272, -0.834974, 1.772637, 1.178132}, {0.074604, 0.659245, -0.705119, 0.223669, -0.485801, 1.033467, -0.715720}, {-0.136692, 0.137963, -1.168972, 0.644912, 1.061856, 0.513119, -1.560229}, {2.156336, 0.239713, -0.415305, 0.686935, 1.937618, -2.037241, -0.100820}, {0.020073, -0.373244, -0.647614, 0.150633, -0.449129, 0.430215, 0.277896}, {0.444533, -0.624721, -1.329337, 1.558808, 0.217422, 0.805949, 2.068083}, {-2.335518, -0.239333, -0.109405, 1.429009, -0.669401, -1.179957, -0.213814}, {-0.333921, -0.362677, -0.435474, 1.253892, 1.416072, -0.763690, 0.661967}, {-0.181115, 0.138493, 1.366809, 0.551946, -0.039155, 0.347498, 1.679509}, {1.198599, 0.075667, -0.054268, -1.279603, -1.373878, -0.703800, 0.677555}, {-0.167772, -0.732645, -1.515290, 0.856168, -0.494425, -1.681764, -0.163964}, {-0.149224, -1.848600, -1.263492, 0.993221, -0.381539, 0.734938, -0.156577}, {0.443171, -0.335002, 1.294719, 2.191558, 0.773367, -0.402849, -1.780908}, {-1.555923, -0.359046, 1.543773, 0.214130, -0.069633, 0.028185, 1.911095}, {0.429081, 0.735976, 0.101460, -0.524960, -0.967164, 1.054060, 0.470863}, {1.892414, 0.988418, 1.806905, -0.742363, 1.224471, 1.230520, 0.444990}, {-0.991714, 0.261782, -0.281851, -1.111159, 1.694775, 0.337672, 0.076663}, {0.763886, 1.026814, 1.219172, 0.250832, -0.729404, -0.055184, -0.158664}, {-0.695060, 0.238589, 0.865602, -1.057215, 2.463648, -0.774730, 0.358056}, {-2.705787, -0.078818, -0.055459, 0.540247, -1.080984, -0.053141, 0.817100}, {1.944077, -0.005803, -0.027517, 1.650756, 2.863406, 0.173228, 0.024172}, {0.172046, 0.912889, -0.508188, 1.183756, 1.891109, 0.921967, 0.476755}, {0.699844, -0.467966, -0.921602, 1.178662, 1.007938, 0.259489, 0.699302}, {-1.478899, 0.927853, -1.157417, 0.791290, -0.749162, 0.492176, -1.185043}, {-0.999042, -1.027887, -1.364015, -0.591327, -0.474763, -1.208045, 0.430538}, {-0.800343, 1.541555, 1.031772, -0.917328, 0.543998, -1.222769, 0.863224}, {-0.263042, 0.288231, 0.901492, 1.140097, -1.437371, 1.828394, -0.377533}, {0.163948, -0.273522, -0.129859, 1.628759, -1.639803, -1.794266, 1.457964}, {-1.750538, -0.582361, 0.330569, -1.283558, 0.682469, -0.615233, 0.429554}, {1.708116, -0.148980, -0.429488, 0.459757, 0.626010, 0.769382, -1.188111}, {-0.611271, -0.230935, -1.091861, -1.744548, -1.422058, 0.935411, 0.352200}, {-0.104508, -0.011096, 0.306886, -0.852831, -0.519469, -1.799441, -1.550412}, {-1.070489, -1.573226, 1.992178, -0.266655, -1.251889, -0.962516, 0.337915}, {0.363050, -2.380764, -0.304986, 0.425878, -0.732397, 0.260353, 0.395049}, {-0.000232, -0.128440, 1.520913, -0.560140, -0.037275, -0.470442, 0.668105}, {0.224975, 0.415550, -0.142758, -0.403193, 0.577823, 1.760222, -1.008182}, {1.679082, -0.522406, 1.252651, 0.726569, -1.138841, -0.134862, 0.864691}, {0.414512, -1.186584, -0.230021, -0.377805, -1.815897, -1.118923, -1.254323}, {-0.327371, -0.365776, 1.670324, -0.200645, -2.989900, 0.871709, -0.104388}, {-0.705777, 0.265598, -0.823743, -0.242920, -1.513456, -0.716423, 1.812060}, {1.291512, 1.478615, 0.184696, -0.396396, 0.854701, 1.134207, 0.529737}, {0.862489, -0.726241, 2.164594, 1.042785, -0.438666, 1.141170, -0.168175}, {1.061887, 0.895762, -0.052095, -1.416987, 0.095715, 0.691293, -0.918647}, {-1.000058, 0.547047, -0.546307, -0.084595, 0.621130, -0.302180, 0.469606}, {0.679107, -1.895292, 0.272574, 1.250030, 1.662812, -0.449364, -1.244615}, {0.156115, 0.125764, 1.057010, 1.011698, -1.499174, -1.133748, 1.094618}, {-0.281807, -0.119503, -0.800737, -0.349155, -0.076019, -0.402870, 0.029194}, {0.081589, -0.425287, 0.544395, 1.246330, 0.183875, 1.155470, -0.130784}, {0.888689, -1.712741, -1.523807, 0.195281, -2.089343, 1.418961, -0.198941}, {0.423233, -1.393455, 0.498937, -0.047019, 2.985555, 0.407483, -1.639636}, {0.421298, -1.948300, -1.615689, 0.767461, -2.109762, 1.120741, -0.428593}, {0.274957, -0.336195, 1.386681, 0.304736, 1.671508, 1.335241, -1.474302}, {1.436645, 0.498732, -1.458595, -0.268329, 0.469628, -0.190401, 1.138691}, {-0.838583, 0.327477, 0.736082, -0.183990, -0.159501, 0.333950, -2.606901}, {0.939967, -0.009277, -1.251972, -0.554533, -0.043566, -1.353669, -0.573020}, {-0.153817, 0.788379, 0.019268, 0.193870, -0.161068, -1.310599, -0.053685}, {0.604562, -0.749423, 1.293674, 0.217036, -1.265642, 1.580531, -0.006460}, {-0.486422, -0.219300, 0.911250, 1.490705, -0.548184, -0.601978, 0.761184}, {0.053844, -0.247163, -0.559801, 0.571919, 0.398882, 0.618781, 0.917986}, {-0.928002, -1.604622, -1.673003, -1.163895, 0.648231, -1.083364, -0.483824}, {0.389746, -0.112244, -1.156270, 0.432104, -1.815976, 0.174886, 0.501362}, {0.748531, 0.075194, -1.298366, -0.494940, -1.693832, 0.234850, -1.710935}, {1.054037, -0.733337, -1.560314, -0.704097, 1.727730, 0.513825, -0.606639}, {1.144406, 1.087550, -0.028094, 0.999519, -1.668986, 0.176972, -0.388373}, {-0.253142, 0.061833, -0.908363, -0.294242, -1.548744, 0.470493, 1.866765}, {1.348302, 0.743482, 0.362793, 0.527881, 0.387783, 1.496947, -0.722666}, {-0.278094, 0.078799, -1.308014, -1.004229, -1.313654, 0.198291, 0.235432}, {1.483193, 0.492566, -0.076458, -1.239170, -0.347386, 1.593749, -0.505469}, {-1.788257, -0.050411, 0.445664, 0.739832, 0.176453, 1.264318, 1.721851}, {0.424288, 0.494271, -0.577725, -0.204186, 0.103250, 0.570722, -0.977402}, {-0.508597, 0.616442, -1.095486, -1.204095, 0.163342, 2.341817, -0.720794}, {-0.112543, -1.351039, -0.506060, 0.951474, 1.842088, -2.075120, 1.893879}, {-0.763346, -1.037103, -1.618438, -0.283593, 0.858748, -0.342351, 0.046334}, {0.644740, 0.758130, 0.803799, -0.193059, 0.159468, 1.066760, -1.155951}, {-0.052770, 1.151688, 0.317891, -1.276412, 1.189200, -1.805785, 0.253575}, {-1.080002, 1.754104, 0.574145, 1.475415, -1.585669, -1.533567, -0.226906}, {2.050106, -0.722922, 1.473294, 0.798642, -0.765282, -0.044898, 0.724441}, {-1.509749, 0.525662, -0.535280, 0.543195, 0.064440, 1.395804, 1.440743}, {0.185261, 0.017359, -0.829525, -0.031584, -0.488548, -1.117409, -0.445351}, {-0.858505, 1.105898, 0.259082, 0.755942, 1.449547, 0.029463, -0.499907}, {-0.781235, -1.993569, 0.799914, -1.519198, -0.588485, -1.505965, -1.430879}, {0.645399, 0.488615, 0.287787, -1.419393, 0.299152, 0.347151, 1.258092}, {1.603764, 0.370864, -0.876371, -0.454501, -1.533583, -0.102865, 0.251980}, {-2.193332, -0.329708, 0.965220, -0.384782, 0.306656, -0.576190, 1.102584}, {0.969246, 0.238093, 0.693797, -0.829440, -2.235616, -1.269811, 1.355738}, {1.026065, 0.828108, 1.505872, -1.007762, -1.993225, 1.748945, -2.462584}, {0.981909, 0.786083, 0.756220, -1.480755, -1.531888, -0.232259, -1.595192}, {0.898055, -0.050623, 0.115459, -0.440205, 1.381456, -0.332516, 0.215156}, {1.342492, -1.696399, -1.359352, -0.761965, 0.572137, -0.528203, -2.159062}, {0.543757, 0.471798, 1.371531, -1.751162, 1.438243, -1.375432, -0.351719}, {-2.690765, -0.379073, -2.550643, 2.233613, -0.839781, -1.242048, -0.665363}, {0.140160, 2.484421, -0.188510, -0.971760, -0.439586, -0.980753, -1.819492}, {-0.836848, -0.793562, -1.405980, 1.030090, -0.104238, 0.906893, 0.746733}, {0.133552, 0.132739, 0.749369, 1.047591, 0.941972, 1.243876, 0.402260}, {1.469937, -1.205808, -0.509873, -0.671460, -0.009592, -0.399916, -0.361036}, {0.766383, -1.044789, -1.027987, -0.086007, 0.606486, -0.606362, -1.014915}, {-0.216639, 1.702560, 1.093835, 0.015748, 1.537193, 0.451429, -2.044684}, {1.236231, 0.193218, 1.082880, -0.141485, -0.395388, -0.568405, 0.465331}, {0.939213, 0.242122, 1.824631, 0.465393, -0.108150, 0.819084, -0.879828}, {0.330657, -1.406299, 0.662594, 0.017626, 1.146652, -0.711901, 0.269460}, {0.064573, -0.492460, 0.109162, -0.986099, -0.014957, 0.802984, 0.298934}, {-0.352672, -1.741359, -0.214951, -0.556506, 1.359173, 0.828460, 1.143465}, {-0.697815, -0.120267, 0.283855, 1.084394, -0.088466, 0.558256, 2.036899}, {0.242731, 0.278075, 1.213621, -0.723630, -0.005095, 0.307490, 0.357128}, {-1.007332, -1.051215, -0.125220, -0.556561, 1.534338, 0.990157, 0.444103}, {-0.522861, 1.150128, 0.074986, -1.379414, -0.532722, -0.944569, 0.902948}, {-0.425976, 0.553961, 0.327380, 1.500597, 1.278843, -1.177087, -0.838999}, {2.168246, 0.064886, -0.540491, -1.424466, -2.090398, 0.178045, 0.513616}, {-0.573694, 0.753284, 0.552182, 0.729499, 0.478713, -0.839644, -0.511373}, {-0.574466, -2.420368, -1.165116, -0.304274, -1.221357, 0.327235, -1.943678}, {-0.467212, -0.865116, -1.055721, -2.222356, -1.397684, -0.796103, 0.104394}, {0.280181, 0.385536, 0.063015, 1.762302, -0.106211, -1.845700, 1.337100}, {0.718680, 0.265581, 1.069172, -0.302968, 0.067504, -0.895981, 0.549785}, {0.378455, 0.745699, 1.136444, -0.001376, 0.929862, 0.652986, -0.843707}, {-0.574589, -0.159937, -0.707849, 1.234431, -2.086166, 0.858165, 0.379395}, {0.843912, -1.448519, 0.548376, -0.914162, -1.128419, -0.391916, 0.822381}, {-0.650671, 1.130470, 0.290640, 0.045476, -0.100821, 0.654088, 0.266244}, {-0.827412, 1.700369, 1.898151, 0.739782, -0.820580, 1.437900, 0.489482}, {0.124536, -0.475098, -1.241890, -0.819789, -1.855379, 0.710430, 1.190052}, {0.534321, 0.553301, 0.544184, -1.481023, 1.728111, 0.440311, -0.402655}, {1.722018, -0.316350, 0.640796, 0.930082, -0.488673, 0.405737, -0.225626}, {-0.155362, 1.881339, 0.199877, 0.347439, 0.865337, -0.053896, -0.652202}, {-1.805353, -0.142274, -0.029879, 1.977554, -0.220419, 1.429929, 0.535999}, {-1.058808, -0.191621, 0.053030, -0.976239, 1.450322, 1.497755, 0.357209}, {-1.721590, 0.810086, -1.098214, 0.309877, -0.918914, -1.991732, -0.566200}, {-0.254362, 2.061166, -0.131545, 0.031893, 0.117842, -0.729064, 0.844479}, {-0.673098, 0.751442, 1.762498, 0.317787, 0.935254, -0.431764, -0.915370}, {1.626083, 0.695033, 0.664834, 0.954930, 0.776319, -0.421994, 0.261857}, {-0.199195, -0.499351, -0.663040, 0.901247, -0.121757, 0.842862, -0.134692}, {-2.341785, -1.774343, 1.333935, -1.147909, 2.735073, 0.464495, 0.872819}, {-0.209604, 0.735154, 0.281346, -0.903590, -2.927820, -0.837440, -0.627100}, {-2.109795, -1.776224, 0.963115, 0.376069, -0.448406, 0.162079, 1.083669}, {1.307322, 1.834629, 0.171228, -0.120797, -0.944619, -0.247007, -1.472225}, {0.004194, 0.035183, 1.336903, 0.826501, 0.232748, 0.256298, -1.723673}, {0.033025, -0.201045, -0.490896, 1.384562, -2.089920, -1.310828, -0.885009}, {-0.070335, 2.067718, -1.267577, 0.354946, 0.850370, 2.045928, 0.164520}, {-0.995398, 0.697452, 0.980327, 0.026035, -0.047517, 1.422009, 0.101260}, {1.393381, 0.104683, -1.056751, -2.444540, 0.980447, -0.663547, 1.200030}, {1.240394, -0.645641, 1.260063, 0.846163, -1.298068, -0.546074, -1.026389}, {-1.527644, 0.208418, -0.579390, 0.521388, -0.595609, 1.025788, 1.122007}, {-0.110983, 1.021369, -0.630432, 0.900810, -0.128914, -0.117629, -1.431200}, {1.475664, 0.159046, -0.947344, -0.361665, -2.127504, 1.040952, -0.553049}, {-1.065357, 0.121993, -0.447551, 0.117675, -1.565283, -1.435383, -0.016059}, {0.537845, 2.430756, -0.269730, 2.629101, 2.330102, -1.267111, 0.698958}, {1.308903, -0.794684, -1.206443, 0.028939, -1.126543, 0.511488, -1.734527}, {0.856551, -0.216851, -0.572863, 0.161903, 1.101932, 0.214256, 0.775635}, {0.715279, 1.519330, 0.668956, -0.704693, -0.792006, 0.028706, -0.306443}, {1.140960, -0.082639, 0.732745, -0.823535, -0.192111, 1.490849, -0.855118}, {-0.404664, -0.207526, 0.578439, -0.619128, -1.121075, 0.813428, -0.863076}, {-0.710032, -0.036664, -1.956659, 1.078572, 0.556353, 0.378143, -1.657540}, {0.176620, -0.704009, -0.324513, 1.147560, -0.613513, -0.126924, -1.270730}, {-1.491357, 0.439628, 1.632689, 1.132555, -0.624201, -0.068683, -0.306465}, {-1.386708, 0.111036, 0.116283, 1.108908, 2.001926, 0.685580, -0.294138}, {-1.450397, -0.941282, -1.031969, -0.315377, -0.000317, 1.143028, 0.613340}, {0.349871, -1.352434, 1.025509, -1.367890, -0.094508, 0.516568, -1.821782}, {0.169347, -0.820089, 1.228509, -0.690141, -0.651776, 0.028256, -2.161114}, {0.849079, 0.533860, 1.251493, -1.055622, 0.214119, -0.881347, 0.074684}, {1.249189, -1.070243, -1.374940, 0.669398, -1.940428, 0.109882, -1.016083}, {-0.191908, 2.094440, 0.659095, -0.240652, 0.738487, -0.599779, -0.782658}, {0.708636, -0.409441, 0.435242, 1.163958, -0.436334, 1.297010, 0.229302}, {-0.620313, -1.300580, 1.987941, 1.018857, -1.217760, 1.753618, -1.069182}, {0.710488, -2.361732, -1.111964, 0.309194, 1.285574, 0.982525, 0.517224}, {-0.974145, 1.241151, -0.205081, -0.205147, 0.472725, 0.703682, -0.467798}, {1.020215, 1.016275, 0.899000, -0.688491, 0.886652, -1.162948, -0.957993}, {-0.304922, -0.725035, 0.543422, 1.481178, -1.164346, 0.744917, 1.196852}, {-1.646692, -0.923506, 0.330024, 1.412588, -1.379905, -0.229623, 1.986690}, {0.103967, -0.023628, -0.233622, 0.293328, 1.416605, -2.072649, -0.774643}, {1.687151, -0.631408, 0.385122, 1.446683, -0.585972, -2.403409, 0.614383}, {-2.504489, 0.270970, -0.191735, -0.059314, 0.613580, 0.063160, -0.948700}, {-1.565747, -0.690983, 1.032926, -0.113241, -0.490638, -0.272634, -0.136300}, {0.809185, -0.542141, 0.766953, 1.514788, 0.355193, -1.229567, -0.476010}, {0.794149, -1.319200, -0.658100, -1.204521, 0.577671, -0.606010, -0.167096}, {0.586281, 0.147649, 1.829041, -0.224892, -1.099504, -0.561805, 0.769496}, {-0.443344, -0.495639, 0.790493, 1.625606, -0.347512, 1.063451, -0.140390}, {-0.778686, 1.026932, -0.540004, 0.005751, 0.013379, -1.336466, -0.916806}, {-1.385688, 1.705243, -1.410701, -0.263344, 1.157063, -1.061039, 0.848959}, {0.060788, -0.113121, 1.424064, 0.170291, 1.246289, 0.916686, -0.331867}, {0.388211, 0.590356, -1.888345, -1.617196, -0.188628, 1.651168, 1.068662}, {0.870853, 2.052208, -0.950370, -0.714436, -1.190649, -1.306547, 0.190417}, {0.128200, -0.421287, -0.472580, -0.513686, 0.885454, -1.879903, -0.772448}, {-1.314694, -0.667809, -0.266896, 0.596300, -0.319814, 0.198982, -0.346163}, {-1.697728, 1.094654, -0.394325, 1.296079, 0.059745, -0.584435, 0.085642}, {-0.217153, 1.417331, 0.582242, 0.250364, -0.550717, -0.980701, 1.124882}, {0.299492, -0.450273, -1.058392, -1.575633, 0.882912, -0.370303, -0.729840}, {-0.071222, -0.339465, -0.816106, -1.448802, 0.508456, -0.461597, 1.079092}, {-1.456001, -1.121962, -0.679186, 0.086435, -0.139451, -1.173183, 0.075456}, {-0.139417, -0.987288, 1.215572, -1.907576, 0.782598, -0.653064, -1.034869}, {0.185559, -0.464468, 1.008524, 0.746082, -0.354635, 0.915755, 0.495952}, {-0.807069, 0.142774, -0.579362, 1.326638, 0.957820, 0.309561, -0.834743}, {-0.637368, 1.282063, -0.548289, -0.861888, -0.544529, -0.398007, 0.362147}, {-1.218120, 0.382936, 0.232775, -0.123555, -0.173388, -0.055352, -0.407116}, {-1.329970, -0.874650, -0.161990, -0.900299, -0.970200, 0.262206, 0.589357}, {-0.334452, 0.602618, -0.474235, -0.312540, -0.454567, 1.109282, 0.182758}, {0.794364, 0.488717, -0.365789, -0.202056, 0.847719, -0.099309, 1.646015}, {-0.278149, 1.164732, -1.451228, 1.736053, -0.491604, -1.325423, 0.432420}, {0.588059, 0.943728, -0.773525, -0.552799, 1.668467, -0.454548, 1.196114}, {0.928666, -0.168919, 1.449351, -0.838897, 0.419365, 1.484633, -0.532433}, {1.031364, 0.301993, -0.155017, -0.241466, 0.914034, -0.371891, -0.026825}, {-0.458315, 0.838749, -0.708707, 1.592318, -0.348008, 1.275376, 2.311813}, {-1.696464, 0.127808, -0.147908, -0.602094, 0.755603, -0.405628, 2.785633}, {-1.484661, -1.175168, 0.127759, 0.002326, 0.789803, 1.024954, 2.206698}, {0.848958, 0.249855, -0.046287, -1.217554, -0.256372, 0.836535, -0.534402}, {-0.795248, -1.101999, -0.668230, 0.955190, 1.034914, -0.153284, -1.209003}, {0.055632, 0.687008, -0.064635, -0.500997, -0.131865, -0.719676, 1.940682}, {0.140056, 0.990034, -0.313559, 0.240841, 0.026392, -0.394134, -0.652088}, {0.932033, -0.246426, -1.681496, -0.358824, 1.698621, 0.856195, 0.933850}, {-1.013349, 0.226633, 1.738019, 0.282761, 0.657410, 0.764595, 0.420551}, {0.216129, -1.874638, 0.105986, 0.112884, -1.795355, -0.243650, -0.483254}, {0.389845, -0.803005, 1.141344, 0.892194, 0.366850, 0.605004, 2.947323}, {0.754513, -0.152351, 0.904260, -0.844187, -0.746722, -0.112881, -0.079754}, {-0.104023, -0.556148, 0.955704, -0.538967, -1.228592, 0.885369, -0.264559}, {-0.055122, 2.197493, -0.001261, 0.897356, -0.651122, 0.026030, 0.261756}, {1.175068, 0.429401, -0.004905, 0.236582, 0.824499, 0.412780, -0.145065}, {-0.240078, -0.140963, -0.139243, 0.735519, 0.832915, -0.624603, 0.258648}, {-0.526684, 1.020440, 0.301098, 0.347675, -0.836435, -2.253123, -1.416922}, {-1.045768, -1.739485, -0.146873, -0.098475, 1.384575, 0.032904, 0.112782}, {0.231630, -0.404080, -0.350853, 0.143785, 1.408354, 1.773649, -0.998045}, {0.175660, -0.657093, 0.565556, 0.207900, -1.377063, 0.943428, 2.587887}, {-0.585274, 0.076816, -0.068205, -0.678534, -0.931584, -1.166510, -1.008846}, {-0.440184, 0.299828, -0.813679, -1.288249, -0.135907, -0.389672, 1.365928}, {0.138203, 1.263931, -2.012952, 0.501085, 0.372276, 0.819947, -0.219367}, {-0.492624, -0.253836, 0.375608, -1.614055, -0.574409, 0.680444, -1.699224}, {0.386536, 1.273978, 0.797841, -0.909902, -0.711780, 0.770082, 2.923317}, {-0.538138, 0.919321, -1.760819, 0.456131, 2.219289, -0.995029, 0.325506}, {0.910943, 0.639517, 0.472180, -0.648800, 0.812767, 0.725349, 1.161427}, {-1.021660, 1.093998, 0.649621, -0.520664, -0.572361, 0.222617, 0.023931}, {0.737298, 1.076790, 0.562945, 0.817660, -0.086542, 0.489743, -0.829748}, {-1.098515, -0.974715, -2.004971, 0.015051, -0.023912, 0.494556, -0.619373}, {0.380810, 1.071214, -1.597207, -0.395980, 0.131652, 0.563781, 2.179103}, {-1.508324, 0.332685, -1.432729, -2.279023, -0.720751, -0.112919, 0.159019}, {2.234890, -0.086654, 2.235577, -0.843684, -0.926850, -0.530383, -0.093459}, {-2.387376, 1.002488, -0.958377, 0.647899, 0.360085, -0.605239, -0.608577}, {-1.371006, 0.263484, -1.751364, -0.378199, 1.444411, -0.285884, -0.813939}, {-0.773161, -1.031148, -0.692105, -1.548615, -1.264169, -0.442366, -0.269334}, {1.284269, 0.451211, 1.759405, 1.121106, 0.176650, 0.281437, -0.047619}, {-1.702677, 1.159975, 0.827429, 0.417806, -1.180094, 0.690837, -0.196571}, {-1.296661, 0.887779, -0.628266, 0.428943, -0.795349, 0.788907, -0.940965}, {-2.207665, -0.246309, -0.438466, -0.093061, 1.479354, -0.470143, 2.363249}, {-0.160727, -1.285292, 0.231374, 0.136660, -0.018640, 0.610441, -0.886301}, {-0.358053, -0.469566, 0.733185, 0.262081, -0.386229, 0.152270, 1.331362}, {0.120915, -0.974180, -1.023874, 2.389655, 0.353017, 0.553500, 0.818873}, {0.154425, -0.374467, 0.632263, -0.231061, -0.315619, 1.105300, -0.456279}, {1.044090, 0.700787, 0.593102, 0.149732, -0.456184, -1.159135, -0.338937}, {0.965015, 0.083452, -1.266795, -1.926762, 2.216549, -0.236753, 1.289701}, {1.076939, -0.668390, -1.295918, -0.342966, 0.330159, -0.250453, -1.725529}, {0.502464, 0.510049, -1.063164, -1.311016, 1.110700, -1.373748, 0.805077}, {0.864617, 0.149471, -1.576784, -0.982653, 0.325026, 1.968739, 0.846863}, {-0.792146, -1.713199, -1.266703, -0.811383, 0.999921, -0.028773, 1.499848}, {0.626473, -0.405704, 0.084820, -1.108024, -1.495428, 0.812764, 0.241414}, {-0.234264, -2.559656, -1.265340, -0.480451, -0.542336, 0.421684, -1.491831}, {0.568948, -0.171436, 1.041529, 1.374401, 0.964391, -0.004261, 0.837077}, {-0.720025, 0.350989, 0.424013, -0.113058, -1.274315, -0.191567, 0.809307}, {1.820651, -1.097020, 0.228973, 0.008031, -2.278962, 2.733842, 0.601328}, {0.606183, 0.345723, 1.386245, -0.362229, -0.126057, 0.962151, 0.509763}, {1.158370, 0.069769, 0.794558, 0.930125, -0.807882, 1.121794, -0.163564}, {-0.080288, 0.072764, 0.156181, 1.190850, 1.028540, 0.207654, -0.326590}, {0.910024, -1.553642, -0.436976, 0.405369, 0.001199, -0.093161, -0.322829}, {1.036649, -0.281630, 0.538033, -0.453761, -0.492072, 1.654288, -0.497919}, {1.312004, 0.887968, -0.786384, -0.266101, -0.769453, 0.038264, -0.167714}, {0.341025, -0.061720, -1.819988, 0.165894, -0.399705, -1.502168, 0.000582}, {0.498297, -0.182240, 1.388778, -1.028521, 1.055056, 0.758442, 0.321348}, {0.585558, -1.032424, -0.659372, -0.604475, -1.399413, 0.837434, 0.674078}, {-1.197727, 0.785037, 0.817680, -0.554026, -0.426857, -0.697384, -1.593910}, {0.028073, -0.405201, -0.674933, -1.135414, -2.456697, -0.308847, -2.205059}, {1.763475, -0.272521, 1.650929, 1.125033, 0.928344, -0.058473, 1.817798}, {0.164859, 0.923594, -0.326754, 0.626440, 0.762630, -0.149993, 0.155407}, {0.791537, -1.313196, -0.955414, -0.176554, 0.524627, 0.318954, -0.361571}, {0.796277, 1.977801, 0.000732, 0.633084, -0.153876, 1.893749, -1.165565}, {0.304073, 0.438990, -2.375288, 0.131585, -0.751921, 0.448665, -0.172982}, {-1.061667, -0.695980, 1.055090, 0.459813, 0.435696, 0.479571, -0.105939}, {-0.413942, -1.518304, 0.962624, 0.309644, 1.548748, 0.248495, 0.306500}, {0.988570, 0.473306, -2.116423, 0.578215, -0.729631, 0.872724, -1.504895}, {0.471946, -0.131446, -0.245908, 0.001976, 0.852115, -0.752894, 0.533710}, {0.637505, -1.307210, -0.116857, -0.524877, -0.299164, -0.254088, 0.222315}, {-0.821244, -0.417748, -1.485281, -0.224061, -0.296139, -1.011433, -0.108758}, {-1.673193, 1.013387, -1.090323, 0.819671, -1.479576, 0.272410, 0.183399}, {1.085715, 0.013700, 1.173351, 0.259666, -0.619420, 0.720353, 0.045294}, {-1.189464, -0.595108, 0.926988, 0.676748, 0.221660, -1.009384, 0.192316}, {-2.293019, -0.588696, 0.601181, 0.724190, 0.063955, 0.070800, -0.313702}, {0.362201, 0.388412, 0.134280, 0.781923, -1.079580, 1.988211, -1.521848}, {0.993282, 0.718697, 0.013471, -1.144290, -1.439392, 0.247888, -0.290341}, {0.789427, 1.632838, 0.129057, 0.141953, -0.786594, 2.906119, -0.287394}, {1.832575, -1.339974, -0.374218, -0.059490, -0.716613, 1.533069, -1.559975}, {1.859441, 0.438989, -0.957978, 0.945378, 0.524435, 1.048707, -1.007226}, {-0.321436, -1.301612, 0.410043, 0.958277, 0.210781, -1.068966, 0.463035}, {0.934900, -0.395463, -2.646197, -0.299383, 2.104331, -0.377691, -0.115026}, {0.759788, 0.925262, -0.724657, 1.020230, -0.268928, -1.372091, 1.634653}, {1.784254, 0.995648, -1.631240, 0.298193, 0.583946, 0.561967, -0.839060}, {0.424816, -0.449142, 1.653732, 1.660718, -1.120627, -0.659651, 0.402288}, {0.214455, 1.368962, -0.702918, 0.039793, 0.855627, 0.413457, -1.079564}, {0.172208, 1.642594, -1.194962, -0.321636, 0.097694, 0.528139, -0.819461}, {-0.334117, -0.708552, 0.848390, -0.803910, 1.107010, -0.043898, -1.072062}, {0.440233, -0.095488, 0.860466, -1.657732, -0.123918, -0.102772, -0.381028}, {-0.956621, -0.243835, 0.684020, -0.948223, 0.889368, 0.265504, 0.425410}, {0.912726, -0.563666, 0.018831, 0.647837, -0.381323, -0.990349, 0.458437}, {0.770206, -0.070574, 0.009700, -0.063946, 0.853142, -0.985843, 0.090103}, {-1.477595, -0.295264, -0.649880, -0.114606, 0.980126, 0.510681, -1.637917}, {0.523418, 2.405528, -1.123686, -0.317072, 0.627951, -0.049108, 1.316867}, {-0.143291, 0.224014, -0.650541, 0.111443, -0.891877, -0.103505, -0.175623}, {-0.372424, -1.045561, -0.950021, -1.933128, -2.157691, 0.111224, 0.208013}, {-0.384911, 0.015543, 0.859340, 1.691500, -0.605339, -0.330656, 0.169301}, {-0.631831, -0.015524, 0.324716, 0.978409, -0.841173, -1.281133, 0.044665}, {-1.139274, -0.733096, -0.501593, 0.832262, -0.413065, -0.714639, 2.288466}, {-1.351503, 0.694370, 0.396826, -2.136583, -0.493341, 1.515917, 0.414049}, {1.004214, 0.146710, -0.777768, 0.239617, 0.097895, -0.703136, -0.022281}, {1.321678, -1.303958, -2.414288, -0.673055, 0.272396, -0.834920, -0.605274}, {0.246416, 0.282674, -0.808984, 0.427890, -0.224480, 0.239375, -0.676480}, {-2.049000, -1.133822, 1.027966, 0.770752, 0.412663, -0.560010, -2.033211}, {-0.880822, 1.346898, 0.571789, 0.472250, 0.011454, -0.510986, -0.460256}, {-0.705721, 0.075538, 1.189200, -0.117541, 0.881150, 1.680868, 0.246887}, {-1.850614, -1.237712, 1.144765, 0.628932, 0.319104, -0.538664, 2.123936}, {2.273178, -0.571699, -0.633609, 1.237220, 0.185485, -0.342544, -0.567902}, {-2.221677, -1.447404, -0.142187, -1.165208, 0.454832, -0.816312, 0.421933}, {-0.374106, -0.260450, -0.054814, -0.292649, -1.203924, -0.058692, 0.601391}, {1.912467, -0.233516, 0.646077, -1.913301, 0.660432, -2.682080, -0.321041}, {2.110535, 0.522387, -0.965496, 0.058939, 0.244477, -0.965246, 0.584958}, {0.078282, 1.809317, 0.521762, -1.193143, -0.249220, -1.131568, -1.280065}, {0.167213, -0.652448, 2.160539, -1.242445, -0.870984, -0.664023, -1.971796}, {0.287650, 0.449932, 0.468510, -1.313966, -1.592046, 0.365026, -1.326907}, {0.026079, -0.450736, -0.680118, -0.103548, 1.526023, -0.059752, -0.904730}, {-0.181320, -0.795575, 0.586679, 1.429102, 1.657079, -0.080260, 0.016473}, {0.349250, 0.377393, -0.919976, -0.621900, 1.515056, 2.269941, -0.710753}, {0.549709, 0.445624, 0.655874, 0.195474, 0.563713, 1.500631, -1.038561}, {0.202612, -0.284596, 1.510965, -0.235929, -1.032874, -0.031505, -0.285765}, {-0.409104, 0.427765, 1.388683, 1.605103, -0.155008, -1.567126, -0.934376}, {-1.060338, -0.945584, 0.174922, 0.468600, -1.930993, -1.341178, -1.355182}, {-0.620044, 0.080327, 0.971001, 0.044092, -0.361465, 0.690115, 0.661894}, {-0.615840, -0.541673, 0.368276, -0.043633, -0.039221, 0.526586, 0.463698}, {-0.197630, 0.590644, 1.754870, 0.480821, 0.161590, 0.760752, 0.044406}, {0.085727, -1.277469, -1.204071, -1.650892, -1.493575, -0.644772, 0.196575}, {-0.878529, 0.396970, -0.689568, 0.120623, 0.625886, 0.534299, 0.429441}, {0.285736, 0.687754, 1.126080, -0.394127, 0.126321, -0.523985, 0.489449}, {-0.486059, 1.533562, -0.890951, -1.200014, -0.366463, 0.770612, -1.319382}, {1.722713, 0.062125, -0.231415, 0.610589, 0.509245, -0.541925, 0.165558}, {-0.130205, -0.268022, 0.057131, -1.693884, -0.586481, -1.608942, -1.216073}, {-1.505120, -0.148775, -0.622224, 0.966304, -0.440621, -0.318404, -0.897527}, {-0.156700, 0.690792, -1.646728, -1.297075, -0.101440, 0.013257, -0.654132}, {-0.683585, 0.111085, 0.602438, -0.967855, 0.768949, -1.564880, -0.413881}, {-1.619628, 0.149076, 0.410138, 0.870283, 0.035537, -0.455880, 1.690361}, {-0.480361, -0.203939, -1.040302, 1.965703, -1.240347, -1.227220, 0.097510}, {-1.199162, -0.932904, 1.550829, 1.439179, -0.093107, -0.560801, 0.420389}, {0.501065, -0.465929, 0.352367, -0.907214, -0.768715, -0.330719, 1.388332}, {0.249946, 0.604382, 0.026765, 0.499627, -0.255924, 0.713578, -0.295038}, {0.590073, -1.044376, -1.299964, -0.005143, -0.205650, 0.537037, 1.541252}, {0.900795, -2.079538, 0.142867, -0.177031, -2.521735, 0.287368, 0.029437}, {0.009713, -0.917212, -0.236935, -0.587100, 0.480900, -1.862268, -1.665743}, {-0.936328, -0.221032, 0.313050, 0.898754, 0.854494, -1.032229, 2.832475}, {0.894701, -0.619096, 0.581343, -0.581397, -0.073143, -0.504056, 0.392045}, {-0.813489, -0.323072, 1.153873, -0.969649, -0.250335, 0.460943, -0.047531}, {-0.940062, 1.023808, -0.460067, -1.490924, -0.017497, 0.273336, -1.381088}, {0.414788, 0.690761, -0.304868, -0.872490, 0.034340, -0.837589, -1.231982}, {-2.284267, -1.914469, 0.576143, -0.828901, 0.031028, -2.297929, -0.783219}, {0.563678, -1.769648, -0.043613, -1.034238, -0.714883, 2.522755, -0.434110}, {0.037195, 0.862771, -0.919234, 0.085200, -0.790415, 1.307717, 0.751886}, {0.350614, -0.236360, 0.234814, -0.046440, -0.482830, -1.478974, -1.999957}, {1.026868, -0.346117, -0.297310, 0.079027, 1.386303, 1.399466, 0.352605}, {0.789704, -2.275653, -0.059052, 0.042086, -1.201241, -1.032599, -0.036411}, {0.256933, -1.635634, 1.632918, 0.738434, -0.981742, -0.179731, -2.554124}, {-1.304665, 1.410611, -0.795582, 0.201370, -1.408719, 0.908783, -0.384219}, {-0.448080, 2.513062, -0.484128, 0.100528, -0.255425, -0.292184, -0.376023}, {0.963615, 0.047359, -2.366836, 0.614421, -1.510530, -0.657384, 0.787332}, {0.622897, -0.349918, 0.041194, 0.284161, -0.731693, -1.440209, 0.390208}, {0.514076, 0.520830, -0.639314, 0.009141, -0.158139, 0.202823, 1.417163}, {-0.451953, 0.933121, 0.589510, 0.466943, 0.727406, 0.657263, -0.382734}, {0.114480, 0.587184, 2.170818, -0.941772, -0.664889, 0.641831, -0.571003}, {-0.344963, -0.599582, 1.238566, -1.077928, 1.803963, 0.852794, 0.324898}, {1.679473, 0.127020, 0.003283, -0.599126, 0.236611, -1.037926, -0.934600}, {-0.668191, -0.393220, -0.620071, -2.054776, -1.411215, -0.180612, -0.450051}, {0.304142, -0.794995, -1.619245, -0.310922, 1.257897, 0.315303, 0.113735}, {0.260607, 0.456464, 0.164832, -0.130918, 0.913486, 0.173962, 0.110185}, {1.993499, -0.752998, -0.044435, -0.036001, -1.245989, 1.020916, -0.328612}, {0.841581, -1.433094, -2.406729, -1.625424, -0.020767, 0.391722, -1.125255}, {0.247186, -0.386456, 0.151341, -0.642387, 3.109398, 0.159787, 0.472058}, {-1.962093, 1.001435, 1.097925, -1.801398, 0.125709, -1.415127, -0.289337}, {0.525535, -1.893592, -0.567805, 0.568444, -0.794669, -0.932222, -0.380393}, {1.503433, 0.502833, 0.304495, 1.388934, -0.929547, -0.429931, -1.531605}, {0.232012, 1.030456, -0.275748, -0.932265, 0.536162, -0.665278, -1.088788}, {-0.357166, -0.570030, 1.023096, 0.030803, -0.453003, -1.962444, -1.450872}, {0.941363, -0.419840, -1.319610, -0.832549, -1.632650, -2.022884, 1.557976}, {0.857577, -0.397711, -0.348045, 0.465581, -0.596469, 0.084675, -2.237782}, {1.126454, -0.522888, -0.952357, -0.757624, 2.633488, -0.179282, -1.612114}, {2.103862, -3.056547, 0.836959, 0.740499, -0.817763, 0.126751, -0.228426}, {0.567129, -0.418811, -0.996708, -0.083529, -0.272893, -1.973238, -0.996501}, {0.084300, 1.332270, 0.014912, 0.601808, 0.524118, 0.387050, -0.269004}, {-0.141732, -0.208751, 0.219178, 1.148916, -0.218763, 0.204544, 0.747312}, {0.500220, 0.818361, 0.012424, 0.083361, -1.311254, -0.607254, 0.594103}, {1.477521, 1.526504, -0.155475, -1.528140, 0.341483, -0.538182, -0.712630}, {0.024711, 1.240859, -0.929679, 0.740447, 0.962622, 1.095394, -0.521867}, {-0.683034, -1.046842, -0.601905, 1.362008, -1.473906, 0.034883, 0.292153}, {0.972336, -1.589494, 0.620629, -0.188608, 0.446685, 0.119791, -1.317976}, {0.193438, 0.092501, 0.118184, -0.332968, 0.498864, 1.691725, 0.822124}, {0.652973, -0.616013, -2.468577, 0.530063, -1.772129, -1.528796, -0.317076}, {0.640658, 0.435571, -1.380704, -0.289741, 0.532431, -0.665856, 1.104393}, {1.391229, -1.295727, -0.584583, -0.255654, -1.227334, -0.318360, -0.347679}, {0.626043, 0.451829, -0.694094, 1.381478, 1.802846, -0.668006, 0.679442}, {0.007392, 1.749495, -1.873895, -0.372441, 0.202955, 0.296948, -0.529400}, {1.140691, -1.136595, -0.453597, -0.150014, -0.621034, 0.349101, -0.226424}, {-0.898817, -0.440015, 0.816469, 0.404689, -0.871208, -0.234488, -0.003685}, {1.291388, 1.679722, -0.273798, 0.416549, 0.692261, 0.643943, -1.493176}, {-0.917927, 0.377109, -0.119745, 1.306845, -0.700674, -0.758321, -0.280135}, {-0.013789, -0.057656, 0.045405, -1.042437, 0.347101, 0.484056, -0.353559}, {-1.737075, 0.117514, 0.050999, -0.429552, -0.313298, -0.409963, -0.550239}, {-0.244103, -0.405335, -0.770432, 0.717393, 0.397430, -1.053373, 1.883385}, {-0.705243, 0.680155, -1.764936, 2.054156, 0.698147, -1.260638, 0.474326}, {-0.606643, 0.941444, 0.589597, -1.025951, 1.895964, 0.141145, 0.130247}, {-1.217005, 0.613750, 0.094717, -1.293187, 0.335481, 2.115223, -0.224903}, {0.149973, 1.100592, 0.278615, 0.654351, 0.373159, -0.937382, -0.377601}, {-1.838855, 1.320461, 0.157081, -1.708270, -0.295089, 0.025272, 0.392339}, {0.327103, 0.277205, -1.040636, 0.386477, 0.814508, -0.137790, -0.461415}, {-0.044005, -0.223334, -1.456722, -1.343840, 0.651528, 0.028582, 0.367417}, {-1.228544, -0.722516, 0.258582, -0.374139, -0.312150, -1.988744, -2.122348}, {-0.766870, 0.065087, -0.712206, 0.607495, 0.524200, -1.222981, -0.498995}, {1.331420, -0.344784, 0.581728, 0.328789, 0.226018, -0.630868, 0.853180}, {0.926163, 1.354888, 2.207795, -0.542477, 0.159988, -0.419095, -0.523508}, {-0.615034, -0.158456, 0.783548, -2.253297, 0.853628, 0.798805, 1.611182}, {0.405043, -1.134122, 0.443819, -1.123335, -1.663445, -0.238260, -0.245738}, {-0.338119, 1.126165, 0.758929, 1.762113, -0.852597, -0.683673, -1.105512}, {0.789363, 0.887237, 0.183715, -0.205711, 0.885339, -0.135511, 1.829929}, {1.437758, -1.407469, 0.037315, -0.524658, -0.912583, -1.248907, -1.420971}, {-1.638385, -0.651506, 1.732063, -0.120014, -1.080648, 0.435695, -0.306428}, {-2.531944, 1.069275, 1.529995, -1.504426, 0.298989, 1.606777, 0.405308}, {-0.878052, -0.801724, 1.507284, 1.493489, 0.461408, 1.104592, 1.282009}, {-0.301480, -0.461647, -1.555266, 0.541093, -1.632795, 0.492918, -0.763241}, {0.651875, -0.086134, 1.317502, 0.520760, 0.663380, -0.651072, -1.540595}, {-0.498507, -0.555812, 0.176321, 1.168060, -0.140163, -0.372175, -0.336852}, {1.937597, -0.649668, 0.147594, 1.100091, 0.140252, 0.230038, -0.041066}, {0.036026, -0.752531, -0.797455, 0.472821, -0.661831, 0.349013, 0.581051}, {-0.711459, 0.576905, -0.307944, 1.345109, -1.653191, 0.786776, -0.192724}, {-0.296901, -1.149438, 0.479911, -0.221608, -0.891582, -0.042378, -1.784598}, {0.379147, 0.623050, -0.353962, 2.891661, -0.204526, -0.238885, -2.526278}, {0.465237, 0.930167, 0.481739, -1.528223, 0.770846, -0.313312, 0.372662}, {0.706643, 0.288839, 0.691567, -0.987476, 0.798044, -1.427046, -0.096795}, {-0.023885, -0.663183, 0.762280, -0.350342, 1.637208, 2.011845, 0.410385}, {-0.376118, -0.904158, 0.188624, 1.564896, 0.253367, -1.052413, 0.896197}, {-0.989908, -1.151683, 2.225896, -1.088252, 0.521631, -1.302126, -1.406643}, {1.423899, 1.248999, -0.676942, 0.870329, 1.661335, -0.309805, -0.134420}, {-0.821129, 0.383002, -0.134883, -0.330527, -0.764706, 1.025174, 0.865924}, {-0.835902, -0.675111, 0.834143, 0.057808, 0.688831, 0.992333, 0.894187}, {0.880211, -0.115602, -0.135913, 0.907723, 0.585054, -0.113029, 1.100147}, {0.555928, 0.994050, -1.833750, -1.747518, -0.999348, -1.898860, 1.485551}, {-1.188225, -0.213719, -0.940605, 1.719402, -1.037876, -0.558164, 0.824853}, {0.729725, 1.507961, 0.722925, 0.194073, 0.074570, 1.165318, 0.590950}, {0.554244, 0.331723, -0.859304, 2.947245, 1.135388, -0.173964, -0.115934}, {-0.003610, -1.176757, -0.209049, 0.497781, -1.581173, 0.764063, 0.368826}, {-0.986796, 1.471169, 0.923091, -0.272620, 1.418782, 0.416485, 1.895751}, {-1.185790, 3.046993, -0.056301, 2.114152, 0.570414, -0.750507, -2.612489}, {-0.720436, -0.836171, -0.655760, -0.721453, 0.650939, -1.545510, -2.475172}, {-0.972655, -0.131792, 2.572473, 1.165564, -0.860865, -0.045600, -0.856052}, {-0.555219, -1.721927, -0.961283, 0.474036, -1.255175, 0.901884, -2.145342}, {-0.648288, 2.544053, -1.020584, 0.732852, 0.218533, 0.835013, 1.677240}, {-0.104055, 1.449715, -0.500222, 0.643679, -0.889953, 0.885053, -0.584798}, {0.536265, -1.303450, 0.057385, -0.930930, 0.179426, -0.160285, 1.185504}, {-0.361538, -0.621828, 1.657280, 1.459064, -1.663440, 0.928572, -1.153851}, {0.655740, 1.127465, -0.094599, 0.431102, 0.906256, 1.246297, 0.915368}, {0.246760, -0.599710, 0.791852, 1.307363, -0.163174, -0.419561, -1.110618}, {1.368734, -0.246270, 0.115218, 0.307573, 0.390157, -0.093264, 0.553336}, {-1.817590, 1.212554, 1.110898, 1.254714, 2.373303, -0.105443, 0.527913}, {-0.151182, 0.898776, -0.198134, -0.133090, 0.750785, -0.111658, -0.081137}, {-0.329553, -0.679677, -1.496706, -0.327138, 0.212125, -1.018408, -1.331650}, {2.090477, 1.059842, 0.797457, 0.353581, -0.470253, -0.434295, 0.729926}, {-1.295978, 0.804371, 0.765577, -1.353896, -1.217295, 0.021318, 0.422214}, {-0.904534, -0.518731, 0.740555, -0.688165, 0.377971, -1.046986, -0.966467}, {1.894096, -1.359721, 1.122836, 1.443767, 0.107584, 1.940343, -0.064409}, {0.659161, -1.857683, 1.274660, 0.253785, -1.653762, -1.120347, 0.734150}, {-0.532793, 0.438959, -0.148543, -0.437250, -1.188545, -1.606313, 1.868265}, {0.049481, 0.047026, -0.443974, -0.260439, -0.001234, 1.021873, -0.614201}, {-0.271174, -0.425814, -0.314505, -0.635194, -1.348818, 1.134873, 0.888128}, {-0.751989, 1.377187, -0.167753, -0.527288, -0.570145, 0.048528, -0.536485}, {0.084501, -0.941567, 0.624254, -0.690099, -2.128767, -0.480184, -0.236286}, {-0.557814, 0.025291, 1.576000, 1.147629, -1.508760, -0.564076, 1.148636}, {-0.784258, 1.968054, 1.305918, 0.452868, -0.405109, 0.937945, 1.437706}, {0.330600, -0.366477, 0.376820, 1.532702, -0.450893, -2.357305, 0.169478}, {-1.248719, 1.281648, -0.251801, 0.928360, -1.130744, -0.086237, -0.208130}, {-1.644638, -1.040806, -1.357800, -0.631705, -0.934628, -0.292432, 0.638597}, {-1.771907, 1.214901, -0.431488, 1.383539, -0.347536, 0.561601, 0.619678}, {0.296923, -0.179967, -1.094055, -1.521950, 0.015397, -0.749784, 0.288224}, {-2.631542, -0.990094, 1.130217, 0.876536, -0.039506, -1.798061, -0.968046}, {-1.494388, -0.897878, -1.240891, 0.161008, 0.372336, -0.216436, 1.108248}, {-1.372758, 0.587058, 0.012788, 0.497540, 0.184767, -1.298405, 1.023742}, {-0.810579, 0.009537, -1.228814, 1.875837, -0.762093, 0.102902, -0.160565}, {0.794064, 0.121298, 0.624067, -0.673633, 2.128769, 0.756397, 2.302002}, {0.716022, 0.883732, 0.999924, 0.460740, -1.983769, -0.572930, -1.739895}, {0.587196, 0.249389, 0.381084, -0.445856, -1.715069, 0.132932, -0.005581}, {-0.493578, 0.044248, 1.227825, -0.164291, 0.544746, 2.315311, -0.190447}, {0.773807, 1.151169, -0.036155, 0.762301, -1.205489, -0.906797, -0.525466}, {-0.029225, -1.127693, -0.050942, -0.177535, -0.409819, -0.090096, -0.558048}, {-0.476597, -0.250861, -0.633359, 0.854235, -0.707889, -0.929471, 1.628549}, {0.876626, 1.206475, -0.181032, -0.719900, -0.455572, -0.220954, 1.903518}, {0.773673, 1.487990, -1.438013, -1.523500, 0.369856, 0.817499, -0.261725}, {1.023268, 1.313006, -0.038294, 0.366648, -0.630238, -0.507827, 1.862207}, {-0.977727, 0.306340, -0.545541, 1.622414, 0.075125, -2.331230, -1.437390}, {-0.322842, -1.672787, -0.420805, 0.483383, -1.251742, -0.126733, 1.068153}, {-0.288752, -0.416336, 0.666534, 0.543414, 0.126732, -0.052730, 0.830196}, {0.817745, -0.367167, 0.659677, 0.688855, 0.160913, 0.012704, -0.998625}, {0.054255, -0.011014, 0.378950, -0.346900, 0.263701, -1.811237, 0.271455}, {-0.054998, 0.350160, 0.157146, -0.223026, -0.443746, -0.700677, -0.622348}, {0.713207, -0.228222, -0.022214, 0.391873, -0.954266, -1.608299, 0.983168}, {-0.521847, -0.159477, 0.627857, -0.622727, 0.193596, 0.188650, 0.568792}, {-1.158739, -0.124456, -0.264725, -0.731661, -1.359865, -0.815648, -0.563964}, {-0.521326, 0.910636, 2.074595, -0.764166, 2.563886, -0.245220, 1.262529}, {-2.167355, -0.827073, 0.425753, -0.284508, 0.100964, -0.311740, -0.539446}, {-0.007913, 0.723281, -0.077219, -0.092305, 1.633310, -0.382763, -0.708487}, {-0.322645, 0.283537, -0.001553, -0.142012, 2.508863, 0.889467, -0.536864}, {0.573779, 0.831677, 1.822706, -0.761531, -0.181691, -0.710539, 0.102680}, {1.055023, -0.274555, 1.257804, 0.104938, -0.302132, 0.328194, -0.746671}, {0.872409, 1.198798, 0.641151, 1.331615, -0.522204, -0.006662, -0.118979}, {-1.682087, 0.003525, -1.177142, -0.606235, -0.992898, -0.708305, -0.966526}, {-0.392630, -0.174445, 0.953302, 1.336785, 1.122855, -1.355065, 0.954136}, {-1.757205, 0.577486, -0.540632, -1.030686, 0.374531, 1.325568, 0.179320}, {-0.408357, 0.416282, 1.600561, -1.394662, -0.601462, -0.378027, 0.687370}, {1.266074, 2.013237, 2.187580, 0.972633, 1.578154, 0.335066, -1.236748}, {-0.021387, 0.308866, -1.698581, 1.734218, -0.822539, 0.413273, 0.539095}, {0.494829, 0.192712, -0.549170, 0.477660, -0.103955, -1.253195, 1.550705}, {-0.455398, -0.072911, 0.630930, 0.337719, 0.946035, -0.139479, -0.072280}, {0.755391, 0.011193, -0.347107, 1.497738, -0.535791, -0.392912, 0.022512}, {0.927008, 1.130803, -2.613123, -1.024495, 1.114678, -0.271390, -0.931756}, {-2.033414, -0.160403, -1.014771, -0.843037, 0.786553, 0.369766, -0.620414}, {0.254299, -0.416201, -1.354612, 0.437092, 0.182342, -0.108313, -0.427610}, {-1.147111, 1.048526, 1.189641, 1.312739, 0.012272, -0.767098, 0.112266}, {1.381763, 1.144672, 1.644198, -0.716620, 0.431239, -0.469271, -0.663428}, {-0.799351, 0.945270, -0.450341, -0.314165, -1.110798, 1.266789, 0.947114}, {0.527016, 0.029085, 0.323646, 0.927999, 0.251778, -2.141325, 0.525589}, {-0.098425, 1.046122, -0.544815, -1.875435, 1.076873, -0.241634, -0.229483}, {-0.955482, -0.621567, -0.238230, -0.730770, -0.916576, 0.208944, 2.126328}, {0.330295, -0.527221, -0.602040, -1.522662, -1.058972, 0.609020, 2.083899}, {-0.139851, 0.601366, 0.003333, 0.359494, 0.270175, -0.057281, 0.489689}, {-0.727197, 1.470613, -0.323397, -1.090109, -0.332717, 0.827950, -0.597445}, {0.070838, 0.911204, 0.041000, -0.163160, -0.201929, -0.524702, -1.703640}, {0.537457, 0.412973, -0.892908, 0.144123, 2.413496, 0.598494, -2.083276}, {0.382162, 0.280156, 1.160373, 0.115920, -0.093699, 0.190090, 0.111211}, {-2.370534, -1.919573, -0.294458, 1.503610, 0.834008, 2.533596, -0.280920}, {-0.880665, -0.759384, -0.253680, -1.857860, 0.183148, 0.779603, -2.111708}, {-0.149102, -0.790166, 0.357215, 0.899742, -0.957919, -1.467269, -0.194352}, {-1.682501, -0.504246, 1.080800, -0.271392, 0.534855, 1.974253, -0.277454}, {0.894072, -0.580306, 0.576136, 0.825187, 1.224605, 0.017453, 0.658880}, {1.125659, 0.064842, -0.325383, -0.359887, -0.110578, -0.610508, -1.199665}, {-2.370680, -1.350287, -0.154496, 0.734764, 0.368286, -0.206855, 0.439744}, {-2.196751, -1.377250, 0.157212, -1.469687, 0.535740, -1.102298, -0.029323}, {0.441192, -1.259770, 0.004365, 0.203175, -0.353053, 0.487604, 1.266825}, {0.156687, 0.983366, 2.124747, 0.457097, 0.478290, 1.074510, 1.702383}, {-0.081406, 0.653101, 0.026522, -0.058267, -2.561610, 0.180430, -0.327008}, {0.127843, 0.060723, 1.152524, 0.787462, -0.553017, -0.915264, -0.916079}, {-0.540393, 2.299439, -0.882622, -2.018815, -2.414161, -0.215655, -0.012021}, {-0.276439, -1.696114, 1.066675, 0.102804, 0.747931, -0.997156, 0.650623}, {-0.158684, -0.501369, -0.190596, -0.624901, 1.398399, -2.162299, -0.237880}, {-0.936564, -1.790218, -0.425909, -0.814441, -1.270964, 0.034761, -1.582903}, {0.616466, 0.093897, -0.841682, 0.754563, 0.583410, 1.396155, -0.597503}, {0.173145, 2.070756, 0.539334, -1.511140, -2.974784, 0.683364, 0.949812}, {1.107109, 0.180776, 0.962043, 0.858002, 0.111384, -0.851834, 0.792491}, {-0.979293, 1.266662, -1.315667, -0.252086, -0.086262, 0.980501, -1.328897}, {0.122673, 1.514364, -0.322397, -1.330540, -0.981831, -0.025605, 1.682700}, {-1.249083, 3.631692, -1.435516, 0.205633, 0.590720, 0.878163, 1.244944}, {-0.946469, -0.721779, -1.584730, 1.561548, -0.423234, -1.078360, 2.458404}, {-0.510769, -0.419712, -0.895006, -1.424243, -0.849491, 0.446026, -1.471639}, {0.925562, 1.344485, 0.019709, 1.076863, -0.176246, -2.496274, -1.042111}, {1.678325, -1.492802, 0.236961, -0.196951, -0.607950, 0.029592, 0.044702}, {1.612326, -0.722979, -0.590016, 0.290270, -1.845148, -1.475409, -0.931161}, {-0.850885, 0.710202, -0.525294, 1.015225, 0.393863, -0.548868, -1.297050}, {1.192412, -1.817209, 1.203575, -1.031844, -1.325262, -0.757338, -0.786668}, {-0.795825, 0.790257, 0.253259, 1.345093, -1.726948, -0.152056, -0.455732}, {0.036932, -0.014621, 0.442830, -2.010368, -1.129328, -0.308233, 0.700509}, {-1.551423, 1.785109, 0.243297, 1.629381, 0.555006, 0.813651, 1.565859}, {0.003656, -0.237959, -1.544701, 2.103238, -1.246837, -0.925836, 0.384896}, {0.875305, 0.207904, -1.698263, 1.099212, -0.439450, 0.659978, 0.973552}, {-0.495833, -0.494948, -0.439154, 0.291859, 0.885437, -0.829974, 1.219212}, {-0.275464, -0.079543, -1.500611, 1.441402, 0.251700, 0.730580, 0.293869}, {1.723451, 0.365651, 0.052490, -0.632276, -0.823859, 0.412312, -0.782115}, {0.990571, -0.521673, -0.877992, 0.390429, 0.009672, -0.478859, -1.024799}, {-0.727257, -0.407345, 2.377596, -0.320370, 0.119516, 0.158291, 0.082696}, {0.416775, 1.860321, 2.644242, -0.520820, -1.168277, 0.123046, -0.060878}, {0.114738, -1.285403, -0.318716, 0.697424, 0.015087, -1.944640, -0.701772}, {0.188261, 0.355705, -3.078341, -1.430381, 1.154537, 1.326982, 0.246997}, {0.674308, 0.526969, -0.846005, -0.302878, 1.067277, -0.562291, -0.524793}, {-0.258098, 0.244979, 1.186652, 1.407913, -1.616099, -0.113655, 1.017360}, {-1.026947, -0.464805, 0.944617, -1.220266, 1.404198, 0.885597, -0.015811}, {0.353422, 1.058273, -0.272143, -1.351380, 1.755563, 0.005156, -0.850584}, {0.018010, 0.604133, -1.571532, -0.462147, -0.576427, 0.064359, 0.236105}, {1.789178, 0.146618, 1.095836, 0.065349, -1.860818, -0.497292, 0.677763}, {-1.738913, 1.000360, 0.434407, -0.262263, 1.839007, 0.490402, -0.432746}, {-1.007457, 0.381658, 0.431630, 0.367608, -1.829154, 1.246204, -0.501653}, {-0.077066, -0.122626, -0.330751, -0.388571, -2.105696, 1.970481, 1.024379}, {0.045789, -0.482165, -1.097683, -0.587480, 1.754863, -0.723961, 0.049125}, {2.398289, -0.014247, -0.010517, -1.048863, 1.793415, 0.724414, 0.345597}, {-0.387480, 0.247423, -0.767035, -0.296047, -0.781397, -0.811697, -0.961941}, {2.056451, 0.650939, -0.334725, 0.300514, -0.736718, -0.748084, -1.415819}, {0.304750, -0.461351, 1.207222, 0.688591, -0.626836, 0.841907, -1.164326}, {0.181588, 0.460078, 0.533605, -0.758522, -2.185635, 1.099960, -1.089442}, {-0.348393, 1.615009, 0.831526, -0.845718, -1.097685, -0.675965, -0.100090}, {0.133409, -2.532090, -0.484735, 1.227685, 0.343454, 0.019992, 0.350228}, {-0.287083, 0.582524, -1.091683, 0.305858, 0.094354, 0.331354, -0.301957}, {0.650260, -0.062333, 0.213851, 0.208812, -0.409438, -1.101232, 0.080191}, {1.380896, -1.761835, 0.239195, -1.025903, 2.097023, -0.891387, -1.001102}, {0.607572, -0.992269, 2.123532, -1.103205, 0.623911, 1.556516, -0.018458}, {-0.020051, -0.605311, -0.893203, 1.220786, 0.403786, -0.196448, 0.893111}, {-1.102315, 0.432933, -0.855875, 1.397229, -1.258067, -2.077415, 0.091388}, {0.089151, 1.173030, -0.733883, 1.074375, -1.661406, 0.273466, -0.490401}, {0.665783, -0.655792, 0.309226, 0.580976, -0.327878, 1.814138, 0.764631}, {-1.278265, -0.617510, -0.981969, 0.635523, -0.737211, -0.001204, 2.081229}, {0.000546, -2.010122, -0.450215, -0.725896, -1.107731, 0.753141, -0.825538}, {-0.285064, 2.425245, 0.307542, -0.827546, -0.438320, -1.023097, -0.485819}, {0.451618, 0.176695, -0.438968, -0.380940, -0.056898, -1.626765, 0.358338}, {0.490108, -1.369832, -0.027519, -1.335622, 0.846508, 0.522289, 0.725681}, {1.548137, -0.016396, 0.839309, 2.153360, -0.773789, 0.511515, -0.214529}, {0.367955, 1.798677, -2.045437, -0.132491, -3.005441, 0.711181, -0.788438}, {-0.998753, -0.794558, 0.301405, 2.975637, 0.395634, -2.380460, 0.403648}, {0.243842, -1.201189, 0.128868, -1.231570, 1.319226, 1.069984, -0.534684}, {0.161099, 0.459025, -0.823799, 0.878943, 0.187488, 0.742251, 1.468353}, {0.029333, 0.896327, 0.184125, 0.013839, -0.584877, -1.242871, -0.538538}, {1.989153, 2.345865, 0.742668, -0.433982, 1.061705, 0.280603, 1.032346}, {-0.138076, 1.258558, 0.333153, -0.532453, 0.349123, 0.624820, 1.099784}, {-0.508865, -0.599014, -1.115013, 0.695952, 1.996056, 0.948906, 0.752990}, {-0.515089, 2.433534, 0.061520, 0.687747, -0.735501, 0.047111, 0.731445}, {-0.764102, -0.049405, -0.570211, -1.156736, 1.151111, 1.166849, -0.559156}, {-1.214655, 0.531277, -0.749718, 0.001714, -0.908939, -0.795358, -0.860563}, {-1.356771, 1.290532, -0.694592, 1.586272, 0.414753, 1.056531, -1.325587}, {-0.049520, 0.039713, 0.090488, 0.459068, 1.722823, -0.582978, 0.226075}, {-1.056377, -0.569695, -0.568812, 1.229393, 0.330043, 0.922518, 0.938401}, {-0.146189, -0.415615, -0.345777, -2.253835, -1.348542, -1.302687, -0.961243}, {-0.571848, 0.871991, -0.162783, 1.871673, -0.354754, 0.760256, -0.656748}, {0.399739, -0.158923, -0.430210, 0.776031, 0.195993, 0.773286, -2.754941}, {1.619081, 0.534640, 0.016871, -0.783172, 0.819716, -0.951173, 0.522147}, {-1.414255, -0.958190, -2.473473, 0.616754, 0.320105, -0.378605, -0.466613}, {0.510833, -0.291936, 2.238774, 1.112824, 1.264775, -1.933618, -0.196907}, {0.794348, 1.942942, 0.260459, 0.350966, 1.770447, -0.095186, -0.042499}, {-0.610639, 0.606154, -0.471901, 0.797665, 0.440699, 0.740675, -0.081243}, {-0.465655, 0.176802, -1.873604, -0.039789, 0.332668, -0.572674, 1.183735}, {0.111099, -0.195607, 0.785269, 1.461691, 0.011706, 0.668153, -1.245505}, {0.472664, 0.022479, -0.443963, 0.736387, -1.841162, -0.277862, -0.188264}, {-0.981828, -1.441825, -1.880463, 2.262637, 0.557903, -0.462703, 2.742091}, {0.419926, -0.234010, 1.396111, -1.534789, -0.610757, -0.118252, -0.494976}, {0.905606, -0.772716, -1.182609, -1.885146, 0.561858, 0.057854, 0.317396}, {-0.938665, -0.055798, 0.251827, -2.790857, 0.810214, 0.089854, -0.751666}, {0.739937, -0.608413, -2.186872, 0.375318, 0.373012, 0.597464, -0.577956}, {0.939925, 1.635401, -0.557236, -0.159409, 0.049681, -0.257990, 0.314559}, {0.028938, -0.343351, -0.765230, 0.234780, -2.063201, 0.513337, 0.170518}, {0.022820, 0.375655, -0.345589, -2.391409, -0.618376, -1.644823, -0.436247}, {1.265240, 1.917425, -1.562015, 1.475056, 0.711231, -0.696575, -3.000501}, {-1.909431, -0.422452, 0.942052, 0.136108, -1.058239, -0.027052, -0.267861}, {0.436659, -0.165707, 0.148909, -0.693782, -0.936383, 0.649276, -0.994255}, {2.220199, -1.347508, 1.383118, -0.424420, 0.085799, -1.781701, 0.191859}, {-1.114305, -0.741197, 0.712654, -0.483241, -0.893474, 0.604628, 0.518635}, {-0.842414, 1.560152, -0.544739, -0.264723, -0.078805, 0.256358, -0.790219}, {-2.271390, 1.547173, 1.122401, 0.610481, 0.642029, 0.652040, 1.254825}, {1.523089, 0.775700, -1.909716, -1.002979, -1.802290, 0.513716, 1.288523}, {-0.312692, -0.783777, 2.641662, -0.090605, 1.716928, 0.308853, -0.838124}, {0.434590, 0.197825, -0.083546, -0.567031, 0.674897, -0.741491, -2.113316}, {1.161506, 0.692684, 0.803065, 2.058100, -0.031628, 0.928551, 0.890256}, {-0.418115, -0.886216, -0.429419, -0.228491, 0.135690, 0.909045, 1.398511}, {1.928719, -0.369214, 0.936557, 0.056074, 0.272401, 0.237059, -1.232510}, {-0.344817, 0.977045, 0.674258, -0.702958, -0.611400, -1.088770, 0.078649}, {-0.562473, 1.566720, -0.368806, 0.819902, -0.189307, 2.179437, -0.897562}, {-0.022760, 0.126629, 0.827750, 1.301873, 1.563749, -0.152952, 2.217130}, {-0.839688, -1.068477, 0.537277, -1.144707, 0.787460, 0.871021, -0.556820}, {-0.340391, -1.123461, 0.241991, 0.543957, -0.206570, -0.082934, -0.711977}, {-1.183122, -1.838593, -0.471401, -0.303723, -0.406972, -1.137663, -0.244321}, {-0.442694, 1.154965, 0.197743, -0.464117, -1.590452, -0.250015, 0.632427}, {0.308510, 1.498557, 1.280704, 0.430438, -0.845515, 0.508266, 0.229633}, {-0.452780, -0.832601, 0.385452, 0.894225, 0.821074, 0.958300, -0.059476}, {-1.499613, 0.689722, 0.109255, -0.487166, 0.816692, -0.226638, 0.638032}, {-0.434180, 0.251153, 0.671547, 1.735936, -1.018805, 0.417416, -0.122175}, {0.965841, 1.029623, 1.816699, -1.067706, 0.804623, 0.332524, 0.092694}, {-0.664528, -0.446576, -0.157270, 0.953403, 2.280762, 0.826639, -0.077365}, {-1.325961, -1.312978, 1.652204, -0.990007, -0.810417, -0.098851, -0.361116}, {-1.496660, 1.041467, 0.540462, -0.365665, 0.535440, 2.345705, -1.104377}, {-0.722418, -0.967779, 0.049856, -1.264095, -0.099990, -1.108017, 0.039034}, {-0.309721, 1.460855, 0.688303, -0.527677, -1.322783, -0.439276, 0.981775}, {1.692926, 0.966025, -0.596386, 1.627384, -1.057129, -1.478737, 0.151794}, {0.359125, 1.094137, 1.426654, -1.116938, -1.186179, -0.185226, 0.584080}, {1.205736, -0.512740, -0.078743, 1.302787, 0.196303, 1.400361, 0.135881}, {0.511349, 0.906437, 0.896295, -1.063653, 0.678170, -0.553564, -2.208820}, {0.897298, -0.133685, -0.128142, -0.148433, -0.085019, 1.801986, 1.072220}, {0.597457, -0.805562, 0.035261, -0.879307, 1.427874, -0.147667, -0.329570}, {-0.366649, -0.895847, 2.338340, 1.050407, -0.086942, -0.121587, 1.137328}, {-0.195762, 0.072437, 1.148306, -0.231492, 0.772345, 0.324853, -0.260309}, {1.088724, 0.455597, 0.709724, -0.971133, -0.112550, 0.325859, -0.954812}, {0.652798, -0.007621, -0.162059, 0.024746, -1.068411, 0.297974, 1.390609}, {1.681045, -1.402082, -0.252862, -1.308243, -1.422212, 1.988710, 0.639325}, {-0.945204, 0.614140, -0.001997, 1.013244, -0.408911, 0.088733, 0.406346}, {0.698494, -0.622906, 1.580223, 1.810702, 1.506771, -0.631858, 1.706450}, {-1.321365, 1.540004, -0.976053, -1.695880, -0.985273, -0.761974, -0.620609} };
    }
    else{
      throw art::Exception(art::errors::StdException)
	<< "Feynman Scaling MiniBooNE random numbers are only configured for positively charged kaons";	        
    }
  }
  REGISTER_WEIGHTCALC(PrimaryHadronFeynmanScalingWeightCalc)
}

